#include <dirent.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "spasm.h"

struct spasmConfigStruct *spasm_config;

void
spasm_error(int syslevel, char *errstr, ...)
{   
    va_list arglist;

    va_start(arglist, errstr);
    vfprintf(stderr, errstr, arglist);
    fprintf(stderr, "\n");      /* spasm_error args aren't terminated by \n */
    va_end(arglist);
}

int
main(int argc, char **argv)
{
    DIR *ldir;
    struct dirent *dent;
    char lfilepath[FILENAME_MAX+1];
    time_t curtime;

    if((spasm_config = spasm_cfg()) == NULL) {
	exit(1);
    }

    /* Try not to bog down the system */
    nice(5);

    curtime = time(NULL);

    if(spasm_config->logsdb == TEXT_DB) {
	/* Get current time */

        if((ldir = opendir(spasm_config->logdir)) == NULL) {
	    fprintf(stderr, "Error opening log directory: %m\n");
	    exit(1);
        }

        /* Loop through log directory */
        while((dent = readdir(ldir)) != NULL) {
	    struct stat filestat;
	    DIR *userldir;
	    char userldirpath[FILENAME_MAX+1];
	    struct dirent *userdent;
	    int lexpire;
	    char buf[SPASM_DBT_MAX+1];

	    snprintf(userldirpath, sizeof(userldirpath), "%s/%s", spasm_config->logdir, dent->d_name);
	    if(stat(userldirpath, &filestat) < 0) {
	        fprintf(stderr, "Error getting file status (%s): %m\n", userldirpath);
	        continue;
	    }
	    /* Skip non-directories and "." and ".." */
	    if(!S_ISDIR(filestat.st_mode)) continue;
	    if(strcmp(dent->d_name, ".") == 0 || strcmp(dent->d_name, "..") == 0) continue;
	    if((userldir = opendir(userldirpath)) == NULL) {
	        fprintf(stderr, "Error opening user log directory (%s): %m\n", userldirpath);
	        continue;
	    }
	    memset(buf, 0, sizeof(buf));
	    /* Get log expiration time (in days) */
	    if((spasm_lookup_prefs(dent->d_name, "log_expire", buf)) == 0) {
	        lexpire = spasm_config->defaultlexpire;
	    } else {
	        lexpire = atoi(buf);
	    }
	    if(lexpire < 1) lexpire = spasm_config->defaultlexpire;
	    /* Convert to seconds */
	    lexpire *= 86400;
	    /* Loop through user's log directory */
	    while((userdent = readdir(userldir)) != NULL) {
	        snprintf(lfilepath, sizeof(lfilepath), "%s/%s/%s", spasm_config->logdir, dent->d_name, userdent->d_name);
	        if(stat(lfilepath, &filestat) < 0) {
		    fprintf(stderr, "Error getting file status (%s): %m\n", lfilepath);
		    continue;
	        }
	        /* Skip directories */
	        if(S_ISDIR(filestat.st_mode)) continue;
	        /* Compare time, and unlink if necessary */
	        if(curtime - filestat.st_mtime > lexpire) unlink(lfilepath);
	    }
	    closedir(userldir);

            /* Remove directory if it's empty */
            if((userldir = opendir(userldirpath)) != NULL) {
                int isempty = 1;
                while((userdent = readdir(userldir)) != NULL) {
                    if(strcmp(userdent->d_name, ".") == 0 || strcmp(userdent->d_name, "..") == 0) continue;
                    isempty = 0;
                    break;
                }
                closedir(userldir);
                if(isempty) rmdir(userldirpath);
            }
        }
        closedir(ldir);
    }

#ifdef MYSQL_DB
    if(spasm_config->logsdb == MYSQL_DB) {
	MYSQL *mysql;
	MYSQL *myres;
	char query[1024];
	struct spasm_list *cur = NULL;
	struct spasm_list *entries = NULL;
	int lexpire;
	char buf[SPASM_DBT_MAX+1];

	snprintf(query, sizeof(query), "SELECT DISTINCT username FROM logs");
	entries = mysql_dump_entries(query, 0);
	if(entries == NULL) {
	    fprintf(stderr, "Error retrieving usernames\n");
	    exit(1);
	}
	cur = entries;

	if((mysql = mysql_init(NULL)) == NULL) {
	    fprintf(stderr, "Error initialising MySQL structure\n");
	    exit(1);
	}
	if((myres = mysql_real_connect(mysql, spasm_config->sqlserver, spasm_config->sqluser, spasm_config->sqlpass,
	    spasm_config->sqldb, spasm_config->sqlport, spasm_config->sqlsock, 0)) == NULL) {
	    fprintf(stderr, "Error connecting to MySQL server: %s\n", mysql_error(mysql));
	    mysql_close(mysql);
	    exit(1);
	}

	while(cur != NULL) {
            /* Get log expiration time (in days) */
	    memset(buf, 0, sizeof(buf));
            if((spasm_lookup_prefs(cur->entry, "log_expire", buf)) == 0) {
                lexpire = spasm_config->defaultlexpire;
            } else {
                lexpire = atoi(buf);
            }
            if(lexpire < 1) lexpire = spasm_config->defaultlexpire;
            /* Convert to seconds */
            lexpire *= 86400;
	    snprintf(query, sizeof(query), "DELETE FROM logs WHERE username = '%s' AND timestamp < '%li0000000000'", cur->entry, curtime-lexpire);
	    if(mysql_query(mysql, query) != 0) {
		fprintf(stderr, "Error executing SQL query: %s\n", mysql_error(mysql));
	    }
	    entries = cur->next;
	    free(cur);
	    cur = entries;
	}
	mysql_close(mysql);
    }
#endif

    exit(0);
}
