#include <dirent.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "spasm.h"

struct spasmConfigStruct *spasm_config;

void
spasm_error(int syslevel, char *errstr, ...)
{   
    va_list arglist;

    va_start(arglist, errstr);
    vfprintf(stderr, errstr, arglist);
    fprintf(stderr, "\n");      /* spasm_error args aren't terminated by \n */
    va_end(arglist);
}

int
main(int argc, char **argv)
{
    DIR *qdir;
    struct dirent *dent;
    char qfilepath[FILENAME_MAX+1];
    time_t curtime;
    char *oneuser;

    if((spasm_config = spasm_cfg()) == NULL) {
	exit(1);
    }

    if(argv[1]) oneuser = argv[1];

    /* Try not to bog down the system */
    nice(5);

    /* Get current time */
    curtime = time(NULL);

    if((qdir = opendir(spasm_config->quarantinedir)) == NULL) {
	fprintf(stderr, "Error opening quarantine directory: %m\n");
	exit(1);
    }

    /* Loop through quarantine directory */
    while((dent = readdir(qdir)) != NULL) {
	struct stat filestat;
	DIR *userqdir;
	char userqdirpath[FILENAME_MAX+1];
	struct dirent *userdent;
	char buf[SPASM_DBT_MAX+1];
	int qexpire;

	if(oneuser != NULL) {
	    if(strcasecmp(oneuser, dent->d_name) != 0) continue;
	}

	snprintf(userqdirpath, sizeof(userqdirpath), "%s/%s", spasm_config->quarantinedir, dent->d_name);
	if(stat(userqdirpath, &filestat) < 0) {
	    fprintf(stderr, "Error getting file status (%s): %m\n", userqdirpath);
	    continue;
	}
	/* Skip non-directories and "." and ".." */
	if(!S_ISDIR(filestat.st_mode)) continue;
	if(strcmp(dent->d_name, ".") == 0 || strcmp(dent->d_name, "..") == 0) continue;
	if((userqdir = opendir(userqdirpath)) == NULL) {
	    fprintf(stderr, "Error opening user quarantine directory (%s): %m\n", userqdirpath);
	    continue;
	}
	memset(buf, 0, sizeof(buf));
	/* Get quarantine expiration time (in days) */
	if(spasm_lookup_prefs(dent->d_name, "quarantine_expire", buf))
	    qexpire = atoi(buf);
	else qexpire = spasm_config->defaultqexpire;
	if(qexpire < 1) qexpire = spasm_config->defaultqexpire;
	/* Convert to seconds */
	qexpire *= 86400;
	/* Loop through user's quarantine directory */
	while((userdent = readdir(userqdir)) != NULL) {
	    snprintf(qfilepath, sizeof(qfilepath), "%s/%s/%s", spasm_config->quarantinedir, dent->d_name, userdent->d_name);
	    if(stat(qfilepath, &filestat) < 0) {
		fprintf(stderr, "Error getting file status (%s): %m\n", qfilepath);
		continue;
	    }
	    /* Skip directories */
	    if(S_ISDIR(filestat.st_mode)) continue;
	    /* Compare time, and unlink if necessary */
	    if(curtime - filestat.st_mtime > qexpire) unlink(qfilepath);
	}
	closedir(userqdir);

	/* Remove directory if it's empty */
	if((userqdir = opendir(userqdirpath)) != NULL) {
	    int isempty = 1;
	    while((userdent = readdir(userqdir)) != NULL) {
		if(strcmp(userdent->d_name, ".") == 0 || strcmp(userdent->d_name, "..") == 0) continue;
		isempty = 0;
		break;
	    }
	    closedir(userqdir);
	    if(isempty) rmdir(userqdirpath);
	}
    }
    closedir(qdir);

    exit(0);
}
