/*
 * **    spasm - The spasm anti-spam milter
 * **    Copyright (C) 2002-2005 Sean Hanson <aradia@theasylum.org>
 * **
 * **    This program is free software; you can redistribute it and/or modify
 * **    it under the terms of the GNU General Public License as published by
 * **    the Free Software Foundation; either version 2 of the License, or
 * **    (at your option) any later version.
 * **
 * **    This program is distributed in the hope that it will be useful,
 * **    but WITHOUT ANY WARRANTY; without even the implied warranty of
 * **    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * **    GNU General Public License for more details.
 * **
 * **    You should have received a copy of the GNU General Public License
 * **    along with this program; if not, write to the Free Software
 * **    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * */

#ifndef _EXTERN_H
#define _EXTERN_H

#ifdef BERKELEY_DB
  #include <db.h>
#endif

#ifdef MYSQL_DB
  #include <mysql/mysql.h>
#endif

#ifdef QDBM_DB
  #include "depot.h"
#endif

#define SPASM_DBT_MAX 1024

struct spasm_list {
    char entry[SPASM_DBT_MAX+1];
    char strval[SPASM_DBT_MAX+1];
    int value;
    struct spasm_list *next;
};

#define SP_STRCAT(buf, a, b) snprintf(buf, sizeof(buf), "%s/%s", a, b)

#ifdef BERKELEY_DB
int sp_db_open(DB *db, DB_TXN *txnid, const char *file, const char *database, DBTYPE type, u_int32_t flags, int mode);
int db_lookup_entry(char *dbfile, char *key, char *value);
extern void db_insert_entry(char *dbfile, char *key, char *value);
void db_delete_entry(char *dbfile, char *key);
struct spasm_list* db_dump_entries(char *dbfile, int vals);
#endif

#ifdef MYSQL_DB
int mysql_lookup_entry(char *query, char *value);
extern void mysql_insert_entry(char *query);
void mysql_delete_entry(char *query);
struct spasm_list* mysql_dump_entries(char *query, int vals);
#endif

#ifdef QDBM_DB
int qdbm_lookup_entry(char *dbfile, char *key, char *value);
void qdbm_insert_entry(char *dbfile, char *key, char *value);
void qdbm_delete_entry(char *dbfile, char *key);
struct spasm_list* qdbm_dump_entries(char *dbfile, int vals);
#endif

/* Database Independent Functions */
int spasm_verify_filters(char *username);
int spasm_lookup_filter(char *username, char *entry);
struct spasm_list* spasm_get_filters(char *username);
void spasm_set_filter(char *username, char *entry, int value);

int spasm_lookup_bl(char *blacklist, char *entry);
struct spasm_list* spasm_get_bl(char *blacklist);
void spasm_insert_bl(char *blacklist, char *entry, char *val);
void spasm_delete_bl(char *blacklist, char *entry);
int spasm_set_bl(char *blacklist, struct spasm_list *entries);

int spasm_lookup_wl(char *whitelist, char *entry);
struct spasm_list* spasm_get_wl(char *whitelist);
void spasm_insert_wl(char *whitelist, char *entry, char *val);
void spasm_delete_wl(char *whitelist, char *entry);
int spasm_set_wl(char *whitelist, struct spasm_list *entries);

struct spasm_list* spasm_get_re(char *dbname);
void spasm_insert_re(char *dbname, char *entry);
void spasm_delete_re(char *dbname, char *entry);

int spasm_lookup_gl(char *entry, char *value);
void spasm_insert_gl(char *entry, char *val);

int verify_spasm_user(char *username, char *domain);

int spasm_lookup_prefs(char *username, char *entry, char *value);
void spasm_set_prefs(char *username, char *entry, char *value);
void spasm_del_prefs(char *username, char *entry);

struct spasm_list* spasm_get_log(char *username, char *key, int all);
void spasm_delete_log(char *username, char *key, int all);
void spasm_insert_log(char *username, char *entry);

int spasm_lookup_mx(char *entry, char *value);
void spasm_insert_mx(char *entry, char *value);

void spasm_update_stats(char *key);
/* End DB Independent Functions */

/* Internal Error Function */
#define USE_PRINTF 0
#define USE_SYSLOG 1
#define USE_CURSES 2
void spasm_error(int syslevel, char *errstr, ...);

#endif /* _EXTERN_H */
