/*
 * **    spasm - The spasm anti-spam milter
 * **    Copyright (C) 2002-2005 Sean Hanson <aradia@theasylum.org>
 * **
 * **    This program is free software; you can redistribute it and/or modify
 * **    it under the terms of the GNU General Public License as published by
 * **    the Free Software Foundation; either version 2 of the License, or
 * **    (at your option) any later version.
 * **
 * **    This program is distributed in the hope that it will be useful,
 * **    but WITHOUT ANY WARRANTY; without even the implied warranty of
 * **    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * **    GNU General Public License for more details.
 * **
 * **    You should have received a copy of the GNU General Public License
 * **    along with this program; if not, write to the Free Software
 * **    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * */

#include <arpa/inet.h>
#include <curses.h>
#include <dirent.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>

#include "spasm.h"
#include "dbfunc.h"

extern struct spasmConfigStruct *spasm_config;

#ifdef BERKELEY_DB
  #include "spasm_bdb.c"
#endif /* BERKELEY_DB */

#ifdef MYSQL_DB
  #include "spasm_mysql.c"
#endif /* MYSQL_DB */

#ifdef QDBM_DB
  #include "spasm_qdbm.c"
#endif /* QDBM_DB */

/* Database Independent Functions */

/* spasm_verify_filters() - verify that a user has filters set */
int
spasm_verify_filters(char *username)
{
    int rv = 0;

    if(username == NULL) return 0;

#ifdef BERKELEY_DB
    if(spasm_config->filterdb == BERKELEY_DB) {
	DB *db;
	char dbfile[FILENAME_MAX+1];

	SP_STRCAT(dbfile, spasm_config->filters, username);
	if((db_create(&db, NULL, 0) != 0)) return 0;
	if((sp_db_open(db, NULL, dbfile, NULL, DB_HASH, DB_RDONLY, 0) == 0)) rv = 1;
	else rv = 0;
	db->close(db, 0);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->filterdb == MYSQL_DB) {
	char query[1024];
	snprintf(query, sizeof(query), "SELECT filter FROM filters WHERE username = '%s'", username);
	if(mysql_lookup_entry(query, NULL)) rv = 1;
    }
#endif

#ifdef QDBM_DB
    if(spasm_config->filterdb == QDBM_DB) {
	DEPOT *db;
	char dbfile[FILENAME_MAX+1];

	SP_STRCAT(dbfile, spasm_config->filters, username);
	if((db = dpopen(dbfile, DP_OREADER, -1)) != NULL) {
	    dpclose(db);
	    rv = 1;
	} else rv = 0;
    }
#endif

    return rv;
}

/* spasm_lookup_filter() - look up a filter */
int
spasm_lookup_filter(char *username, char *entry)
{
    char result[SPASM_DBT_MAX+1];
    char *domain;
    int rv = 0;

    if(username == NULL || entry == NULL) return 0;
    memset(result, 0, sizeof(result));

#ifdef BERKELEY_DB
    if(spasm_config->filterdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];
	if(spasm_verify_filters(username)) {
	    SP_STRCAT(dbfile, spasm_config->filters, username);
	    rv = db_lookup_entry(dbfile, entry, result);
	    if(rv != 0) rv = atoi(result);
	} else if((domain = strchr(username, '@')) != NULL) {
	    domain++;
	    SP_STRCAT(dbfile, spasm_config->filters, domain);
	    rv = db_lookup_entry(dbfile, entry, result);
	    if(rv != 0) rv = atoi(result);
	}
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->filterdb == MYSQL_DB) {
	char query[1024];
	if(spasm_verify_filters(username)) {
	    snprintf(query, sizeof(query), "SELECT value FROM filters WHERE username = '%s' AND filter = '%s'", username, entry);
	    if(mysql_lookup_entry(query, result)) rv = atoi(result);
	} else if((domain = strchr(username, '@')) != NULL) {
	    domain++;
	    snprintf(query, sizeof(query), "SELECT value FROM filters WHERE username = '%s' AND filter = '%s'", domain, entry);
	    if(mysql_lookup_entry(query, result)) rv = atoi(result);
	}
    }
#endif

#ifdef QDBM_DB
    if(spasm_config->filterdb == QDBM_DB) {
	char dbfile[FILENAME_MAX+1];

	if(spasm_verify_filters(username)) {
	    SP_STRCAT(dbfile, spasm_config->filters, username);
	    rv = qdbm_lookup_entry(dbfile, entry, result);
	    if(rv != 0) rv = atoi(result);
	} else if((domain = strchr(username, '@')) != NULL) {
	    domain++;
	    SP_STRCAT(dbfile, spasm_config->filters, domain);
	    rv = db_lookup_entry(dbfile, entry, result);
	    if(rv != 0) rv = atoi(result);
	}
    }
#endif

    return rv;
}

/* spasm_get_filters() - get filters */
struct spasm_list*
spasm_get_filters(char *username)
{   
    struct spasm_list *first = NULL;
    char *domain;
    
    if(username == NULL) return NULL;

#ifdef BERKELEY_DB
    if(spasm_config->filterdb == BERKELEY_DB) {
        char dbfile[FILENAME_MAX+1];

	if(spasm_verify_filters(username)) {
	    SP_STRCAT(dbfile, spasm_config->filters, username);
	    first = db_dump_entries(dbfile, 1);
	} else if((domain = strchr(username, '@')) != NULL) {
	    domain++;
	    SP_STRCAT(dbfile, spasm_config->filters, domain);
	    first = db_dump_entries(dbfile, 1);
	}
    }
#endif
    
#ifdef MYSQL_DB
    if(spasm_config->filterdb == MYSQL_DB) {
        char query[1024]; 

	if(spasm_verify_filters(username)) {
	    snprintf(query, sizeof(query), "SELECT filter, value FROM filters WHERE username = '%s'", username);
	    first = mysql_dump_entries(query, 1);
	} else if((domain = strchr(username, '@')) != NULL) {
	    domain++;
	    snprintf(query, sizeof(query), "SELECT filter, value FROM filters WHERE username = '%s'", domain);
	    first = mysql_dump_entries(query, 1);
	}
    }
#endif

#ifdef QDBM_DB
    if(spasm_config->filterdb == QDBM_DB) {
        char dbfile[FILENAME_MAX+1];

        if(spasm_verify_filters(username)) {
            SP_STRCAT(dbfile, spasm_config->filters, username);
            first = qdbm_dump_entries(dbfile, 1);
        } else if((domain = strchr(username, '@')) != NULL) {
            domain++;
            SP_STRCAT(dbfile, spasm_config->filters, domain);
            first = qdbm_dump_entries(dbfile, 1);
        }
    }
#endif

    return first;
}

/* spasm_set_filter() - set a filter */
void
spasm_set_filter(char *username, char *entry, int value)
{
     if(username == NULL || entry == NULL) return;

#ifdef BERKELEY_DB
    if(spasm_config->filterdb == BERKELEY_DB) {
	char buf[128];
	char dbfile[FILENAME_MAX+1];

	SP_STRCAT(dbfile, spasm_config->filters, username);
	snprintf(buf, sizeof(buf), "%d", value);
	db_insert_entry(dbfile, entry, buf);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->filterdb == MYSQL_DB) {
	char query[1024];
	/* Update if filter is already set, otherwise insert */
	snprintf(query, sizeof(query), "SELECT value FROM filters WHERE username = '%s' AND filter = '%s'", username, entry);
	if(mysql_lookup_entry(query, NULL))
	    snprintf(query, sizeof(query), "UPDATE filters SET value = '%d' WHERE username = '%s' AND filter = '%s'", value, username, entry);
	else
	    snprintf(query, sizeof(query), "INSERT INTO filters VALUES ('%s', '%s', '%d')", username, entry, value);
	mysql_insert_entry(query);
    }
#endif

#ifdef QDBM_DB
    if(spasm_config->filterdb == QDBM_DB) {
        char buf[128];
        char dbfile[FILENAME_MAX+1];

        SP_STRCAT(dbfile, spasm_config->filters, username);
        snprintf(buf, sizeof(buf), "%d", value);
        qdbm_insert_entry(dbfile, entry, buf);
    }
#endif

    return;
}

/* spasm_lookup_bl() - look up a blacklist entry */
int
spasm_lookup_bl(char *blacklist, char *entry)
{
    int rv = 0;

    if(blacklist == NULL || entry == NULL) return 0;

#ifdef BERKELEY_DB
    if(spasm_config->blacklistdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];

	SP_STRCAT(dbfile, spasm_config->blacklists, blacklist);
	rv = db_lookup_entry(dbfile, entry, NULL);
	if(rv != 0) rv = 1;
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->blacklistdb == MYSQL_DB) {
	char query[SPASM_DBT_MAX*2+SMTP_PATH_MAX+65];
	char quoted_entry[SPASM_DBT_MAX*2+1];

	mysql_quote_string(entry, quoted_entry);
	snprintf(query, sizeof(query), "SELECT value FROM blacklists WHERE blacklist = '%s' AND entry = '%s'", blacklist, quoted_entry);
	if(mysql_lookup_entry(query, NULL)) rv = 1;
    }
#endif

#ifdef QDBM_DB
    if(spasm_config->blacklistdb == QDBM_DB) {
        char dbfile[FILENAME_MAX+1];
        
        SP_STRCAT(dbfile, spasm_config->blacklists, blacklist);
        rv = qdbm_lookup_entry(dbfile, entry, NULL);
        if(rv != 0) rv = 1;
    }
#endif

    return rv;
}

/* spasm_get_bl() - get blacklist entries */
struct spasm_list*
spasm_get_bl(char *blacklist)
{
    struct spasm_list *first = NULL;

    if(blacklist == NULL) return NULL;

#ifdef BERKELEY_DB
    if(spasm_config->blacklistdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];

	SP_STRCAT(dbfile, spasm_config->blacklists, blacklist);
	first = db_dump_entries(dbfile, 1);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->blacklistdb == MYSQL_DB) {
	char query[1024];
	snprintf(query, sizeof(query), "SELECT entry, value FROM blacklists WHERE blacklist = '%s'", blacklist);
	first = mysql_dump_entries(query, 1);
    }
#endif

#ifdef QDBM_DB
    if(spasm_config->blacklistdb == QDBM_DB) {
        char dbfile[FILENAME_MAX+1];

        SP_STRCAT(dbfile, spasm_config->blacklists, blacklist);
        first = qdbm_dump_entries(dbfile, 1);
    }
#endif

    return first;
}

/* spasm_insert_bl() - add a blacklist entry */
void
spasm_insert_bl(char *blacklist, char *entry, char *val)
{
    if(blacklist == NULL || entry == NULL || val == NULL) return;

#ifdef BERKELEY_DB
    if(spasm_config->blacklistdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];

	SP_STRCAT(dbfile, spasm_config->blacklists, blacklist);
	db_insert_entry(dbfile, entry, val);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->blacklistdb == MYSQL_DB) {
	char query[SPASM_DBT_MAX*4+SMTP_PATH_MAX+69];
	char quoted_entry[SPASM_DBT_MAX*2+1];
	char quoted_val[SPASM_DBT_MAX*2+1];

	mysql_quote_string(entry, quoted_entry);
	mysql_quote_string(val, quoted_val);
        /* Update if entry exists, otherwise insert */
        snprintf(query, sizeof(query), "SELECT value FROM blacklists WHERE blacklist = '%s' AND entry = '%s'", blacklist, quoted_entry);
        if(mysql_lookup_entry(query, NULL))
            snprintf(query, sizeof(query), "UPDATE blacklists SET value = '%s' WHERE blacklist = '%s' AND entry = '%s'", quoted_val, blacklist, quoted_entry);
        else
            snprintf(query, sizeof(query), "INSERT INTO blacklists VALUES ('%s', '%s', '%s')", blacklist, quoted_entry, quoted_val);
        mysql_insert_entry(query);
    }
#endif

#ifdef QDBM_DB
    if(spasm_config->blacklistdb == QDBM_DB) {
        char dbfile[FILENAME_MAX+1];

        SP_STRCAT(dbfile, spasm_config->blacklists, blacklist);
        qdbm_insert_entry(dbfile, entry, val);
    }
#endif

    return;
}

/* spasm_delete_bl() - delete a blacklist entry */
void
spasm_delete_bl(char *blacklist, char *entry)
{
    if(blacklist == NULL || entry == NULL) return;

#ifdef BERKELEY_DB
    if(spasm_config->blacklistdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];

	SP_STRCAT(dbfile, spasm_config->blacklists, blacklist);
	db_delete_entry(dbfile, entry);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->blacklistdb == MYSQL_DB) {
	char query[SPASM_DBT_MAX*2+SMTP_PATH_MAX+59];
	char quoted_entry[SPASM_DBT_MAX*2+1];

	mysql_quote_string(entry, quoted_entry);
	snprintf(query, sizeof(query), "DELETE FROM blacklists WHERE blacklist = '%s' AND entry = '%s'", blacklist, quoted_entry);
	mysql_delete_entry(query);
    }
#endif

#ifdef QDBM_DB
    if(spasm_config->blacklistdb == QDBM_DB) {
        char dbfile[FILENAME_MAX+1];
        
        SP_STRCAT(dbfile, spasm_config->blacklists, blacklist);
        qdbm_delete_entry(dbfile, entry);
    }
#endif

    return;
}

/* returns:  1 - all is well, 0 - failure */
int
spasm_set_bl(char *blacklist, struct spasm_list *entries)
{   
    int rc = 0;
    struct spasm_list *first;

    if(blacklist == NULL || entries == NULL) return 0;

#ifdef BERKELEY_DB
    if(spasm_config->blacklistdb == BERKELEY_DB) {
	DB *db;
	DBT dbk, dbv;
	int dbfd, discarded;
	struct flock lockinfo;
	char dbfile[FILENAME_MAX+1];
    
	SP_STRCAT(dbfile, spasm_config->blacklists, blacklist);
	memset(&dbk, '\0', sizeof(DBT));
	memset(&dbv, '\0', sizeof(DBT));
	memset(&lockinfo, '\0', sizeof(lockinfo));
	lockinfo.l_type = F_WRLCK;
	if((rc = db_create(&db, NULL, 0)) != 0) {
	    spasm_error(LOG_ERR, "(spasm_set_bl) Error creating db: %s", db_strerror(rc));
	    return 0;
	}
	if((rc = sp_db_open(db, NULL, dbfile, NULL, DB_HASH, DB_CREATE, 0644)) != 0) {
	    spasm_error(LOG_ERR, "(spasm_set_bl) Error opening database: %s", db_strerror(rc));
	    db->close(db, 0);
	    return 0;
	}
	if((rc = db->fd(db, &dbfd)) != 0) {
	    spasm_error(LOG_ERR, "(spasm_set_bl) Error reading file descriptor for database: %s", db_strerror(rc));
	    db->close(db, 0);
	    return 0;
	}
	if(fcntl(dbfd, F_SETLKW, &lockinfo) == -1) {
	    spasm_error(LOG_ERR, "(spasm_set_bl) Error locking database: %m");
	    db->close(db, 0);
	    return 0;
	}
	if((rc = db->truncate(db, NULL, &discarded, 0)) != 0) {
	    spasm_error(LOG_ERR, "(spasm_set_bl) Error truncating database: %s", db_strerror(rc));
	    db->close(db, 0);
	    return 0;
	}
	while(entries != NULL) {
	    dbk.data = entries->entry;
	    dbk.size = strlen(entries->entry);
	    if(*entries->strval != '\0') {
		dbv.data = entries->strval;
		dbv.size = strlen(dbv.data);
	    } else {
		dbv.data = "1";
		dbv.size = 1;
	    }
	    if((rc = db->put(db, NULL, &dbk, &dbv, 0)) != 0) {
		spasm_error(LOG_ERR, "(spasm_set_bl) Error adding %s to %s: %s", entries->entry, tmpfile, db_strerror(rc));
	    }
	    first = entries->next;
	    free(entries);
	    entries = first;
	}
	if((rc = db->close(db, 0)) != 0) {
	    spasm_error(LOG_ERR, "(spasm_set_bl) Error closing database: %s", db_strerror(rc));
	    if(close(dbfd) != 0) spasm_error(LOG_ERR, "(spasm_set_bl) Error closing dbfd: %m");
	}
	rc = 1;
    }
#endif

#ifdef MYSQL_DB
    /* This should be made into a transaction */
    if(spasm_config->blacklistdb == MYSQL_DB) {
	char query[SPASM_DBT_MAX*4+SMTP_PATH_MAX+44];
	char quoted_entry[SPASM_DBT_MAX*2+1];
	char quoted_val[SPASM_DBT_MAX*2+1];
	MYSQL *mysql;
	MYSQL *myres;
        
	if((mysql = mysql_init(NULL)) == NULL) {
	    spasm_error(LOG_ERR, "(spasm_set_bl) Error initialising MySQL structure");
            return 0;
        }           
        if((myres = mysql_real_connect(mysql, spasm_config->sqlserver, spasm_config->sqluser, spasm_config->sqlpass,
            spasm_config->sqldb, spasm_config->sqlport, spasm_config->sqlsock, 0)) == NULL) {
            spasm_error(LOG_ERR, "(spasm_set_bl) Error connecting to MySQL server: %s", mysql_error(mysql));
            mysql_close(mysql);
            return 0;
        }       
	/* First clear the blacklist */
	snprintf(query, sizeof(query), "DELETE FROM blacklists WHERE blacklist = '%s'", blacklist);
	mysql_delete_entry(query);
	/* Next, add the new entries */
	while(entries != NULL) {
	    mysql_quote_string(entries->entry, quoted_entry);
	    mysql_quote_string(entries->strval, quoted_val);
	    if(*entries->strval != '\0')
		snprintf(query, sizeof(query), "INSERT INTO blacklists VALUES ('%s', '%s', '%s')", blacklist, quoted_entry, quoted_val);
	    else
		snprintf(query, sizeof(query), "INSERT INTO blacklists VALUES ('%s', '%s', '1')", blacklist, quoted_entry);
	    if(mysql_query(mysql, query) != 0)
		spasm_error(LOG_ERR, "(spasm_set_bl) Error executing SQL query: %s", mysql_error(mysql));
	    first = entries->next;
	    free(entries);
	    entries = first;
	}
	rc = 1;
	mysql_close(mysql);
    }
#endif
    return rc;
}

/* spasm_lookup_wl() - look up whitelist entries */
int
spasm_lookup_wl(char *whitelist, char *entry)
{
    int rv = 0;
    char *domain;

    if(whitelist == NULL || entry == NULL) return 0;

#ifdef BERKELEY_DB
    if(spasm_config->whitelistdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];

	SP_STRCAT(dbfile, spasm_config->whitelists, whitelist);
	rv = db_lookup_entry(dbfile, entry, NULL);
	if(rv != 0) rv = 1;
	else if((domain = strchr(whitelist, '@')) != NULL) {
	    domain++;
	    SP_STRCAT(dbfile, spasm_config->whitelists, domain);
	    rv = db_lookup_entry(dbfile, entry, NULL);
	    if(rv != 0) rv = 1;
	}
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->whitelistdb == MYSQL_DB) {
        char query[SPASM_DBT_MAX*2+SMTP_PATH_MAX+65];
	char quoted_entry[SPASM_DBT_MAX*2+1];

	mysql_quote_string(entry, quoted_entry);
        snprintf(query, sizeof(query), "SELECT value FROM whitelists WHERE whitelist = '%s' AND entry = '%s'", whitelist, quoted_entry);
        if(mysql_lookup_entry(query, NULL)) rv = 1;
	else if((domain = strchr(whitelist, '@')) != NULL) {
	    domain++;
	    snprintf(query, sizeof(query), "SELECT value FROM whitelists WHERE whitelist = '%s' AND entry = '%s'", domain, quoted_entry);
	    if(mysql_lookup_entry(query, NULL)) rv = 1;
	}
    }
#endif

    return rv;
}

/* spasm_get_wl() - get whitelist entries */
struct spasm_list*
spasm_get_wl(char *whitelist)
{
    struct spasm_list *first = NULL;

    if(whitelist == NULL) return NULL;

#ifdef BERKELEY_DB
    if(spasm_config->whitelistdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];
	SP_STRCAT(dbfile, spasm_config->whitelists, whitelist);
	first = db_dump_entries(dbfile, 1);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->whitelistdb == MYSQL_DB) {
        char query[1024];
        snprintf(query, sizeof(query), "SELECT entry, value FROM whitelists WHERE whitelist = '%s'", whitelist);
        first = mysql_dump_entries(query, 1);
    }
#endif

    return first;
}

/* spasm_insert_wl() - add a whitelist entry */
void
spasm_insert_wl(char *whitelist, char *entry, char *val)
{
    if(whitelist == NULL || entry == NULL || val == NULL) return;

#ifdef BERKELEY_DB
    if(spasm_config->whitelistdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];
	SP_STRCAT(dbfile, spasm_config->whitelists, whitelist);
	db_insert_entry(dbfile, entry, val);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->whitelistdb == MYSQL_DB) {
	char query[SPASM_DBT_MAX*4+SMTP_PATH_MAX+69];
	char quoted_entry[SPASM_DBT_MAX*2+1];
	char quoted_val[SPASM_DBT_MAX*2+1];

	mysql_quote_string(entry, quoted_entry);
	mysql_quote_string(val, quoted_val);

        /* Update if entry exists, otherwise insert */
        snprintf(query, sizeof(query), "SELECT value FROM whitelists WHERE whitelist = '%s' AND entry = '%s'", whitelist, quoted_entry);
        if(mysql_lookup_entry(query, NULL))
            snprintf(query, sizeof(query), "UPDATE whitelists SET value = '%s' WHERE whitelist = '%s' AND entry = '%s'", quoted_val, whitelist, quoted_entry);
        else
            snprintf(query, sizeof(query), "INSERT INTO whitelists VALUES ('%s', '%s', '%s')", whitelist, quoted_entry, quoted_val);
        mysql_insert_entry(query);
    }
#endif

    return;
}

/* spasm_delete_wl() - delete a whitelist entry */
void
spasm_delete_wl(char *whitelist, char *entry)
{
    if(whitelist == NULL || entry == NULL) return;

#ifdef BERKELEY_DB
    if(spasm_config->whitelistdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];
	SP_STRCAT(dbfile, spasm_config->whitelists, whitelist);
	db_delete_entry(dbfile, entry);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->whitelistdb == MYSQL_DB) {
        char query[SPASM_DBT_MAX*2+SMTP_PATH_MAX+59];
	char quoted_entry[SPASM_DBT_MAX*2+1];

	mysql_quote_string(entry, quoted_entry);
        snprintf(query, sizeof(query), "DELETE FROM whitelists WHERE whitelist = '%s' AND entry = '%s'", whitelist, quoted_entry);
        mysql_delete_entry(query);
    }
#endif

    return;
}

/* returns:  1 - all is well, 0 - failure */
int
spasm_set_wl(char *whitelist, struct spasm_list *entries)
{
    struct spasm_list *first;
    int rc = 0;

    if(whitelist == NULL || entries == NULL) return 0;

#ifdef BERKELEY_DB
    if(spasm_config->whitelistdb == BERKELEY_DB) {
        DB *db;
        DBT dbk, dbv;
        int dbfd, discarded;
        struct flock lockinfo;
        char dbfile[FILENAME_MAX+1];

        SP_STRCAT(dbfile, spasm_config->whitelists, whitelist);
        memset(&dbk, '\0', sizeof(DBT));
        memset(&dbv, '\0', sizeof(DBT));
        memset(&lockinfo, '\0', sizeof(lockinfo));
        lockinfo.l_type = F_WRLCK;
        if((rc = db_create(&db, NULL, 0)) != 0) {
            spasm_error(LOG_ERR, "(spasm_set_wl) Error creating db: %s", db_strerror(rc));
            return 0;
        }
        if((rc = sp_db_open(db, NULL, dbfile, NULL, DB_HASH, DB_CREATE, 0644)) != 0) {
            spasm_error(LOG_ERR, "(spasm_set_wl) Error opening database: %s", db_strerror(rc));
            db->close(db, 0);
            return 0;
        }
        if((rc = db->fd(db, &dbfd)) != 0) {
            spasm_error(LOG_ERR, "(spasm_set_wl) Error reading file descriptor for database: %s", db_strerror(rc));
            db->close(db, 0);
            return 0;
        }
        if(fcntl(dbfd, F_SETLKW, &lockinfo) == -1) {                                                                                                 
            spasm_error(LOG_ERR, "(spasm_set_wl) Error locking database: %m");
            db->close(db, 0);
            return 0;
        }
        if((rc = db->truncate(db, NULL, &discarded, 0)) != 0) {
            spasm_error(LOG_ERR, "(spasm_set_wl) Error truncating database: %s", db_strerror(rc));
            db->close(db, 0);
            return 0;
        }
        while(entries != NULL) {
            dbk.data = entries->entry;
            dbk.size = strlen(entries->entry);
            if(*entries->strval != '\0') { 
                dbv.data = entries->strval;
                dbv.size = strlen(dbv.data);
            } else {
                dbv.data = "1";
                dbv.size = 1;
            }
            if((rc = db->put(db, NULL, &dbk, &dbv, 0)) != 0) {
                spasm_error(LOG_ERR, "(spasm_set_wl) Error adding %s to %s: %s", entries->entry, tmpfile, db_strerror(rc));
            }
	    first = entries->next;
	    free(entries);
	    entries = first;
        }
        if((rc = db->close(db, 0)) != 0) {
            spasm_error(LOG_ERR, "(spasm_set_wl) Error closing database: %s", db_strerror(rc));
	    if(close(dbfd) != 0) spasm_error(LOG_ERR, "(spasm_set_wl) Error closing dbfd: %m");
        }
        rc = 1;
    }
#endif

#ifdef MYSQL_DB
    /* This should be made into a transaction */
    if(spasm_config->whitelistdb == MYSQL_DB) {
        char query[SPASM_DBT_MAX*4+SMTP_PATH_MAX+44];
        char quoted_entry[SPASM_DBT_MAX*2+1];
        char quoted_val[SPASM_DBT_MAX*2+1];

        /* First clear the whitelist */
        snprintf(query, sizeof(query), "DELETE FROM whitelists WHERE whitelist = '%s'", whitelist);
        mysql_delete_entry(query);
        /* Next, add the new entries */
        while(entries != NULL) {
            mysql_quote_string(entries->entry, quoted_entry);
            mysql_quote_string(entries->strval, quoted_val);
            if(*entries->strval != '\0')
                snprintf(query, sizeof(query), "INSERT INTO whitelists VALUES ('%s', '%s', '%s')", whitelist, quoted_entry, quoted_val);
            else
                snprintf(query, sizeof(query), "INSERT INTO whitelists VALUES ('%s', '%s', '1')", whitelist, quoted_entry);
            mysql_insert_entry(query);
	    first = entries->next;
	    free(entries);
	    entries = first;
        }
        rc = 1;
    }
#endif
    return rc;
}

/* spasm_get_re() - get regex db entries */
struct spasm_list*
spasm_get_re(char *dbname)
{
    struct spasm_list *first = NULL;

    if(dbname == NULL) return NULL;

#ifdef BERKELEY_DB
    if(spasm_config->regexdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];
	SP_STRCAT(dbfile, spasm_config->regex, dbname);
	first = db_dump_entries(dbfile, 0);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->regexdb == MYSQL_DB) {
        char query[1024];
        snprintf(query, sizeof(query), "SELECT entry FROM regex WHERE regex = '%s'", dbname);
        first = mysql_dump_entries(query, 0);
    }
#endif

    return first;
}

/* spasm_insert_re() - insert a regex db entry */
void
spasm_insert_re(char *dbname, char *entry)
{
    if(dbname == NULL || entry == NULL) return;

#ifdef BERKELEY_DB
    if(spasm_config->regexdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];
	SP_STRCAT(dbfile, spasm_config->regex, dbname);
	db_insert_entry(dbfile, entry, "1");
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->regexdb == MYSQL_DB) {
        char query[SPASM_DBT_MAX*2+SMTP_PATH_MAX+56];
	char quoted_entry[SPASM_DBT_MAX*2+1];

	mysql_quote_string(entry, quoted_entry);
        /* Insert only if entry doesn't exist */
        snprintf(query, sizeof(query), "SELECT value FROM regex WHERE regex = '%s' AND entry = '%s'", dbname, quoted_entry);
        if(mysql_lookup_entry(query, NULL) == 0) {
            snprintf(query, sizeof(query), "INSERT INTO regex VALUES ('%s', '%s', '1')", dbname, quoted_entry);
	    mysql_insert_entry(query);
	}
    }
#endif

    return;
}

/* spasm_delete_re() - delete a regex db entry */
void
spasm_delete_re(char *dbname, char *entry)
{
    if(dbname == NULL || entry == NULL) return;

#ifdef BERKELEY_DB
    if(spasm_config->regexdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];
	SP_STRCAT(dbfile, spasm_config->regex, dbname);
	db_delete_entry(dbfile, entry);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->regexdb == MYSQL_DB) {
        char query[SPASM_DBT_MAX*2+SMTP_PATH_MAX+50];
	char quoted_entry[SPASM_DBT_MAX*2+1];

	mysql_quote_string(entry, quoted_entry);
        snprintf(query, sizeof(query), "DELETE FROM regex WHERE regex = '%s' AND entry = '%s'", dbname, quoted_entry);
        mysql_delete_entry(query);
    }
#endif

    return;
}

/* spasm_lookup_gl() - look up a greylist entry */
int
spasm_lookup_gl(char *entry, char *value)
{
    if(entry == NULL) return 0;

#ifdef BERKELEY_DB
    return db_lookup_entry(spasm_config->greylist, entry, value);
#endif

    /* TODO: MySQL? */

    return 0;
}

/* spasm_insert_gl() - insert a greylist entry */
void
spasm_insert_gl(char *entry, char *val)
{
    if(entry == NULL) return;
#ifdef BERKELEY_DB
    db_insert_entry(spasm_config->greylist, entry, val);
#endif

    /* TODO: MySQL? */

    return;
}

/* verify_spasm_user() - verify that a user has spasm settings */
int
verify_spasm_user(char *username, char *domain)
{
    int rc = 0;
#ifdef BERKELEY_DB
    if(spasm_config->filterdb == BERKELEY_DB) {
	char email[SMTP_PATH_MAX+1];
	char dbfile[FILENAME_MAX+1];
	struct stat st;
	snprintf(email, sizeof(email), "%s@%s", username, domain);
	SP_STRCAT(dbfile, spasm_config->filters, email);
	rc = !(stat(dbfile, &st));
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->filterdb == MYSQL_DB) {
	char email[SMTP_PATH_MAX+1];
	snprintf(email, sizeof(email), "%s@%s", username, domain);
	if(spasm_verify_filters(email)) rc = 1;
    }
#endif

    return rc;
}

/* spasm_lookup_prefs() - look up a user preference */
int
spasm_lookup_prefs(char *username, char *entry, char *value)
{
    int rc = 0;
    char *domain;

    if(username == NULL || entry == NULL) return 0;

#ifdef BERKELEY_DB
    if(spasm_config->prefsdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];

	SP_STRCAT(dbfile, spasm_config->prefs, username);
	rc = db_lookup_entry(dbfile, entry, value);
	if(!rc && (domain = strchr(username, '@')) != NULL) {
	    domain++;
	    SP_STRCAT(dbfile, spasm_config->prefs, domain);
	    rc = db_lookup_entry(dbfile, entry, value);
	}
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->prefsdb == MYSQL_DB) {
        char query[1024];
        snprintf(query, sizeof(query), "SELECT value FROM prefs WHERE username = '%s' AND pref = '%s'", username, entry);
        rc = mysql_lookup_entry(query, value);
	if(!rc && (domain = strchr(username, '@')) != NULL) {
	    domain++;
	    snprintf(query, sizeof(query), "SELECT value FROM prefs WHERE username = '%s' AND pref = '%s'", domain, entry);
	    rc = mysql_lookup_entry(query, value);
	}
    }
#endif

    return rc;
}

/* spasm_set_prefs() - insert a user preference */
void
spasm_set_prefs(char *username, char *entry, char *value)
{
    if(username == NULL || entry == NULL || value == NULL) return;

#ifdef BERKELEY_DB
    if(spasm_config->prefsdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];

	SP_STRCAT(dbfile, spasm_config->prefs, username);
	db_insert_entry(dbfile, entry, value);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->prefsdb == MYSQL_DB) {
        char query[SPASM_DBT_MAX*4+SMTP_PATH_MAX+62];
	char quoted_entry[SPASM_DBT_MAX*2+1];
	char quoted_value[SPASM_DBT_MAX*2+1];

	mysql_quote_string(entry, quoted_entry);
	mysql_quote_string(value, quoted_value);

        /* Update if entry exists, otherwise insert */
        snprintf(query, sizeof(query), "SELECT value FROM prefs WHERE username = '%s' AND pref = '%s'", username, quoted_entry);
        if(mysql_lookup_entry(query, NULL))
            snprintf(query, sizeof(query), "UPDATE prefs SET value = '%s' WHERE username = '%s' AND pref = '%s'", quoted_value, username, quoted_entry);
        else
            snprintf(query, sizeof(query), "INSERT INTO prefs VALUES ('%s', '%s', '%s')", username, quoted_entry, quoted_value);
        mysql_insert_entry(query);
    }
#endif

    return;
}

/* spasm_del_prefs() - delete a user preference */
void
spasm_del_prefs(char *username, char *entry)
{
    if(username == NULL || entry == NULL) return;

#ifdef BERKELEY_DB
    if(spasm_config->prefsdb == BERKELEY_DB) {
        char dbfile[FILENAME_MAX+1];
                                                                                                                                                       
        SP_STRCAT(dbfile, spasm_config->prefs, username);
        db_delete_entry(dbfile, entry);
    }
#endif

#ifdef MYSQL_DB
    if(spasm_config->prefsdb == MYSQL_DB) {
        char query[SPASM_DBT_MAX*2+SMTP_PATH_MAX+52];
	char quoted_entry[SPASM_DBT_MAX*2+1];

	mysql_quote_string(entry, quoted_entry);
        snprintf(query, sizeof(query), "DELETE FROM prefs WHERE username = '%s' AND pref = '%s'", username, quoted_entry);
        mysql_delete_entry(query);
    }
#endif

    return;
}

/* spasm_get_log() - get a log entry */
struct spasm_list*
spasm_get_log(char *username, char *key, int all)
{
    struct spasm_list *entries = NULL;
    struct spasm_list *first = NULL;

    if(username == NULL || (key == NULL && all == 0)) return NULL;

    if(spasm_config->logsdb == TEXT_DB) {
	char lfilepath[FILENAME_MAX+1];
	DIR *ldir;
	struct dirent *dent;
	int fd;
	char *ptr;
	char buf[SPASM_DBT_MAX+1];

	if(key != NULL) {	/* Get one log entry */
	    snprintf(lfilepath, sizeof(lfilepath), "%s/%s/%s", spasm_config->logdir, username, key);
	    if((fd = open(lfilepath, O_RDONLY)) < 0) {
		spasm_error(LOG_ERR, "(spasm_get_log) Error opening log file (%s): %m", lfilepath);
		return NULL;
	    }
	    if((first = calloc(sizeof(struct spasm_list), 1)) == NULL) {
		spasm_error(LOG_ERR, "(spasm_get_log) Error allocating memory for first: %m");
		close(fd);
	    }
	    strncpy(first->entry, key, sizeof(first->entry));
	    read(fd, first->strval, sizeof(first->strval)-1);
	    close(fd);
	    if((ptr = strchr(first->strval, '\n')) != NULL) *ptr = '\0';
	} else if(all) {	/* Get all log entries */
	    snprintf(lfilepath, sizeof(lfilepath), "%s/%s", spasm_config->logdir, username);
	    if((ldir = opendir(lfilepath)) == NULL) {
		spasm_error(LOG_ERR, "(spasm_get_log) Error opening log directory (%s): %m", lfilepath);
		return NULL;
	    }
	    while((dent = readdir(ldir)) != NULL) {
		struct stat filestat;

		snprintf(lfilepath, sizeof(lfilepath), "%s/%s/%s", spasm_config->logdir, username, dent->d_name);
		if(stat(lfilepath, &filestat) < 0) {
		    spasm_error(LOG_ERR, "(spasm_get_log) Error getting file status (%s): %m", lfilepath);
		    continue;
		}
		if(S_ISDIR(filestat.st_mode)) continue;
		memset(buf, 0, sizeof(buf));
		if((fd = open(lfilepath, O_RDONLY)) < 0) {
		    spasm_error(LOG_ERR, "(spasm_get_log) Error opening log file (%s): %m", lfilepath);
		    continue;
		}
		if(read(fd, buf, sizeof(buf)-1) < 0) {
		    spasm_error(LOG_ERR, "(spasm_get_log) Error reading log file (%s): %m", lfilepath);
		    close(fd);
		    continue;
		}
		close(fd);
		if((ptr = strchr(buf, '\n')) != NULL) *ptr = '\0';
		if(first == NULL) {
		    if((first = calloc(sizeof(struct spasm_list), 1)) != NULL) {
			entries = first;
			strncpy(first->entry, dent->d_name, sizeof(first->entry));
			strncpy(first->strval, buf, sizeof(first->strval));
		    } else spasm_error(LOG_ERR, "(spasm_get_log) Error allocating memory for first: %m");
		} else {
		    if((entries->next = calloc(sizeof(struct spasm_list), 1)) != NULL) {
			entries = entries->next;
			strncpy(entries->entry, dent->d_name, sizeof(entries->entry));
			strncpy(entries->strval, buf, sizeof(entries->strval));
		    } else spasm_error(LOG_ERR, "(spasm_get_log) Error allocating memory for entries->next: %m");
		}
		/*free(dent);*/
	    }
	    closedir(ldir);
	}
    }

#ifdef MYSQL_DB
    if(spasm_config->logsdb == MYSQL_DB) {
	char query[1024];

	if(key != NULL) {	/* Get one log entry */
	    snprintf(query, sizeof(query), "SELECT timestamp, entry FROM logs WHERE username = '%s' AND timestamp = '%s'", username, key);
	} else if(all) {	/* Get all log entries */
	    snprintf(query, sizeof(query), "SELECT timestamp, entry FROM logs WHERE username = '%s'", username);
	}
	first = mysql_dump_entries(query, 1);
    }
#endif

    return first;
}

/* spasm_delete_log() - delete a log entry */
void
spasm_delete_log(char *username, char *key, int all)
{
    if(username == NULL || (key == NULL && all == 0)) return;

    if(spasm_config->logsdb == TEXT_DB) {
	if(key != NULL) {	/* Delete one log entry */
	    char lfilepath[FILENAME_MAX+1];
	    snprintf(lfilepath, sizeof(lfilepath), "%s/%s/%s", spasm_config->logdir, username, key);
	    unlink(lfilepath);
	} else if(all) {	/* Delete all log entries */
	    char lfilepath[FILENAME_MAX+1];
	    DIR *ldir;
	    struct dirent *dent;

	    snprintf(lfilepath, sizeof(lfilepath), "%s/%s", spasm_config->logdir, username);
	    if((ldir = opendir(lfilepath)) == NULL) {
		spasm_error(LOG_ERR, "(spasm_delete_log) Error opening log directory (%s): %m", lfilepath);
		return;
	    }
	    while((dent = readdir(ldir)) != NULL) {
		struct stat filestat;

		snprintf(lfilepath, sizeof(lfilepath), "%s/%s/%s", spasm_config->logdir, username, dent->d_name);
		if(stat(lfilepath, &filestat) < 0) {
		    spasm_error(LOG_ERR, "(spasm_delete_log) Error getting file status (%s): %m", lfilepath);
		    continue;
		}
		if(S_ISDIR(filestat.st_mode)) continue;
		unlink(lfilepath);
		/*free(dent);*/
	    }
	    closedir(ldir);
	}
    }

#ifdef MYSQL_DB
    if(spasm_config->logsdb == MYSQL_DB) {
	char query[1024];

	if(key != NULL) {	/* Delete one log entry */
	    snprintf(query, sizeof(query), "DELETE FROM logs WHERE username = '%s' AND timestamp = '%s'", username, key);
	} else if(all) {	/* Delete all log entries */
	    snprintf(query, sizeof(query), "DELETE FROM logs WHERE username = '%s'", username);
	}
	mysql_delete_entry(query);
    }
#endif

    return;
}

/* spasm_insert_log() - add a log entry */
void
spasm_insert_log(char *username, char *entry)
{
    if(username == NULL || entry == NULL) return;

    if(spasm_config->logsdb == TEXT_DB) {
        char logfile[FILENAME_MAX+1];
        char ldir[FILENAME_MAX+1];
        int fd;
	char *ptr;
        
        SP_STRCAT(ldir, spasm_config->logdir, username);
        mkdir(ldir, 0755);
	if((ptr = strchr(entry, ':')) == NULL) {
	    spasm_error(LOG_ERR, "(spasm_insert_log) Error parsing timestamp");
	    return;
	}

	/* Take the timestamp for filename generation directly from the log entry */
	*ptr = '\0';
        snprintf(logfile, sizeof(logfile), "%s/%s/%s.XXXXXX", spasm_config->logdir, username, entry);
	*ptr = ':';

        if((fd = mkstemp(logfile)) < 0) {
            spasm_error(LOG_ERR, "(log_rejection), Error opening log file %s: %m", logfile);
            return;
        }
        chmod(logfile, 0644);
        write(fd, entry, strlen(entry));
        write(fd, "\n", 1);
        close(fd);
    }

#ifdef MYSQL_DB
    if(spasm_config->logsdb == MYSQL_DB) {
	/* The '4096' is actually REASON_MAX defined in filters.h, but I'm having issues
	   with compilation if I use references to REASON_MAX. *sigh* */
        char query[(70+(2*SMTP_PATH_MAX)+(2*SMTP_DOMAIN_MAX)+4096)*2+SMTP_PATH_MAX+60];
	char *ptr;
	char buf[21];
	char quoted_entry[(70+(2*SMTP_PATH_MAX)+(2*SMTP_DOMAIN_MAX)+4096)*2+1];

	if((ptr = strchr(entry, ':')) == NULL) {
	    spasm_error(LOG_ERR, "(spasm_insert_log) Error parsing timestamp");
	    return;
	}
	/* Take the timestamp for the key directly from the log entry */
	*ptr = '\0';
	strncpy(buf, entry, sizeof(buf));
	*ptr = ':';
	mysql_quote_string(entry, quoted_entry);
        snprintf(query, sizeof(query), "INSERT into logs VALUES ('%s', '%s', '%s')", username, buf, quoted_entry);
        mysql_insert_entry(query);
    }
#endif
    return;
}

/* spasm_lookup_mx() - look up an mx verification result */
int
spasm_lookup_mx(char *entry, char *value)
{
    if(entry == NULL) return 0;

#ifdef BERKELEY_DB
    return db_lookup_entry(spasm_config->verifymx, entry, value);
#endif

    return 0;
}

/* spasm_insert_mx() - insert an mx verification result */
void
spasm_insert_mx(char *entry, char *value)
{
    if(entry == NULL || value == NULL) return;

#ifdef BERKELEY_DB
    db_insert_entry(spasm_config->verifymx, entry, value);
#endif

    return;
}

/* spasm_update_stats() - update statistics */
void
spasm_update_stats(char *key)
{
    if(key == NULL) return;

#ifdef BERKELEY_DB
    if(spasm_config->statsdb == BERKELEY_DB) {
	char dbfile[FILENAME_MAX+1];
	char datestamp[14];
	char value[SPASM_DBT_MAX+1];
	struct tm tms;
	time_t timep;
	int i = 0;

	timep = time(NULL);
	if(localtime_r(&timep, &tms) == NULL) {
	    spasm_error(LOG_ERR, "(spasm_update_stats) Error reading localtime");
	    return;
	}
	snprintf(datestamp, sizeof(datestamp), "%4d-%02d-%02d.db", tms.tm_year+1900, tms.tm_mon+1, tms.tm_mday);
	SP_STRCAT(dbfile, spasm_config->stats, datestamp);
	memset(value, 0, sizeof(value));
	if(db_lookup_entry(dbfile, key, value) == 1) i = atoi(value);
	i++;
	snprintf(value, sizeof(value), "%d", i);
	db_insert_entry(dbfile, key, value);
    }
#endif

    return;
}

/* END DB INDEPENDENT FUNCTIONS */
