#include <arpa/inet.h>
#include <netinet/in.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include "spasm.h"

void spasm_error(int syslevel, char *errstr, ...);

/* Cleanup memory used by config struct */
void
free_spasm_cfg(struct spasmConfigStruct *config)
{
    if(config == NULL) return;

    if(config->spasmuser != NULL) free(config->spasmuser);
    if(config->defaultdomain != NULL) free(config->defaultdomain);
    if(config->domlogexceptions != NULL) free(config->domlogexceptions);
    if(config->spasmroot != NULL) free(config->spasmroot);
    if(config->filters != NULL) free(config->filters);
    if(config->blacklists != NULL) free(config->blacklists);
    if(config->whitelists != NULL) free(config->whitelists);
    if(config->regex != NULL) free(config->regex);
    if(config->prefs != NULL) free(config->prefs);
    if(config->logdir != NULL) free(config->logdir);
    if(config->stats != NULL) free(config->stats);
    if(config->quarantinedir != NULL) free(config->quarantinedir);
    if(config->tempdir != NULL) free(config->tempdir);
    if(config->pidfile != NULL) free(config->pidfile);
    if(config->socket != NULL) free(config->socket);
    if(config->sahost != NULL) free(config->sahost);
    if(config->dspam != NULL) free(config->dspam);
    if(config->localbl != NULL) free(config->localbl);
    if(config->spamtrapdb != NULL) free(config->spamtrapdb);
    if(config->spamtrapbl != NULL) free(config->spamtrapbl);
    if(config->globalipwl != NULL) free(config->globalipwl);
    if(config->globalipbl != NULL) free(config->globalipbl);
    if(config->aliasesfile != NULL) free(config->aliasesfile);
    if(config->sendmail != NULL) free(config->sendmail);
    if(config->greylist != NULL) free(config->greylist);
    if(config->verifymx != NULL) free(config->verifymx);
    if(config->sqlserver != NULL) free(config->sqlserver);
    if(config->sqluser != NULL) free(config->sqluser);
    if(config->sqlpass != NULL) free(config->sqlpass);
    if(config->sqldb != NULL) free(config->sqldb);
    if(config->sqlsock != NULL) free(config->sqlsock);

    free(config);
}

/* Read and return spasm.conf data */
struct spasmConfigStruct* spasm_cfg()
{
    FILE *fp;
    char buf[128];
    char key[128];
    char val[128];
    int seenthresh = 0;
    struct stat filestat;
    struct spasmConfigStruct *spasm_conf;

    /* stat file to get mtime */
    if(stat(SPASM_CONF, &filestat) < 0) {
	spasm_error(LOG_ERR, "(spasm_cfg) Unable to stat config file: %m");
	return NULL;
    }

    /* Open file for reading */
    if((fp = fopen(SPASM_CONF, "r")) == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) Error opening config file: %m");
	return NULL;
    }
    /* Allocate memory... */
    if((spasm_conf = calloc(sizeof(struct spasmConfigStruct), 1)) == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for spasm_conf: %m");
	return NULL;
    }
    /* And load the options! */
    while(fgets(buf, sizeof(buf), fp) != NULL) {
	/* Skip comments */
	if(buf[0] == '#') continue;

	/* Split key/value pair */
	if(sscanf(buf, "%127s %127s", key, val) != 2) continue;

	/* Check for key validity, add value to struct if valid */
	if(strcasecmp(key, "debug") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->debug = 1;
	    else spasm_conf->debug = 0;
	} else if(strcasecmp(key, "spasmuser") == 0) {
            spasm_conf->spasmuser = strdup(val);
            if(spasm_conf->spasmuser == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SpasmUser: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "defaultdomain") == 0) {
	    spasm_conf->defaultdomain = strdup(val);
	    if(spasm_conf->defaultdomain == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for DefaultDomain: %m");
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "domainlogging") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->domainlogging = 1;
	    else spasm_conf->domainlogging = 0;
	} else if(strcasecmp(key, "domainloggingexceptions") == 0) {
	    spasm_conf->domlogexceptions = strdup(val);
	    if(spasm_conf->domlogexceptions == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for DomainLoggingExceptions: %m");
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "spasmroot") == 0) {
            spasm_conf->spasmroot = strdup(val);
            if(spasm_conf->spasmroot == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SpasmRoot: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "filters") == 0) {
            spasm_conf->filters = strdup(val);
            if(spasm_conf->filters == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Filters: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "blacklists") == 0) {
            spasm_conf->blacklists = strdup(val);
            if(spasm_conf->blacklists == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Blacklists: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "whitelists") == 0) {
            spasm_conf->whitelists = strdup(val);
            if(spasm_conf->whitelists == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Whitelists: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "regex") == 0) {
            spasm_conf->regex = strdup(val);
            if(spasm_conf->regex == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for RegEx: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "prefs") == 0) {
	    spasm_conf->prefs = strdup(val);
	    if(spasm_conf->prefs == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Prefs: %m");
		fclose(fp);
		return NULL;
	    }
        } else if(strcasecmp(key, "logdir") == 0) {
            spasm_conf->logdir = strdup(val);
            if(spasm_conf->logdir == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for LogDir: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "stats") == 0) {
	    spasm_conf->stats = strdup(val);
	    if(spasm_conf->stats == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Stats: %m");
		fclose(fp);
		return NULL;
	    }
        } else if(strcasecmp(key, "quarantinedir") == 0) {
            spasm_conf->quarantinedir = strdup(val);
            if(spasm_conf->quarantinedir == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for QuarantineDir: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "tempdir") == 0) {
            spasm_conf->tempdir = strdup(val);
            if(spasm_conf->tempdir == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for TempDir: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "pidfile") == 0) {
            spasm_conf->pidfile = strdup(val);
            if(spasm_conf->pidfile == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for PidFile: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "socket") == 0) {
            spasm_conf->socket = strdup(val);
            if(spasm_conf->socket == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Socket: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "clamav") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->clamav = 1;
	    else spasm_conf->clamav = 0;
	} else if(strcasecmp(key, "spamassassinhost") == 0) {
	    if((spasm_conf->sahost = calloc(sizeof(struct in_addr), 1)) == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SpamAssassinHost: %m");
		fclose(fp);
		return NULL;
	    }
	    if(inet_aton(val, spasm_conf->sahost) == 0) {
		spasm_error(LOG_ERR, "(spasm_cfg) Invalid SpamAssassin Host: %s", val);
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "spamassassinport") == 0) {
	    spasm_conf->saport = htons(atoi(val));
	} else if(strcasecmp(key, "localblacklistdb") == 0) {
	    spasm_conf->localbl = strdup(val);
	    if(spasm_conf->localbl == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for LocalBlacklistDB: %m");
		fclose(fp);
		return NULL;
	    }
        } else if(strcasecmp(key, "spamtrapdb") == 0) {
            spasm_conf->spamtrapdb = strdup(val);
            if(spasm_conf->spamtrapdb == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SpamtrapDB: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "spamtrapblacklistdb") == 0) {
            spasm_conf->spamtrapbl = strdup(val);
            if(spasm_conf->spamtrapbl == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SpamtrapBlacklistDB: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "globalipwhitelist") == 0) {
            spasm_conf->globalipwl = strdup(val);
            if(spasm_conf->globalipwl == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for GLobalIPWhitelist: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "globalipblacklist") == 0) {
            spasm_conf->globalipbl = strdup(val);
            if(spasm_conf->globalipbl == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for GlobalIPBlacklist: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "filtertype") == 0) {
	    if(strcasecmp(val, "berkeleydb") == 0) spasm_conf->filterdb = BERKELEY_DB;
#ifdef MYSQL_DB
	    else if(strcasecmp(val, "mysql") == 0) spasm_conf->filterdb = MYSQL_DB;
#endif
	    else {
		spasm_error(LOG_ERR, "(spasm_cfg) Invalid FilterType '%s'", val);
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "blacklisttype") == 0) {
            if(strcasecmp(val, "berkeleydb") == 0) spasm_conf->blacklistdb = BERKELEY_DB;
#ifdef MYSQL_DB
	    else if(strcasecmp(val, "mysql") == 0) spasm_conf->blacklistdb = MYSQL_DB;
#endif
            else {
		spasm_error(LOG_ERR, "(spasm_cfg) Invalid BlacklistType '%s'", val);
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "whitelisttype") == 0) {
            if(strcasecmp(val, "berkeleydb") == 0) spasm_conf->whitelistdb = BERKELEY_DB;
#ifdef MYSQL_DB
	    else if(strcasecmp(val, "mysql") == 0) spasm_conf->whitelistdb = MYSQL_DB;
#endif
            else {
		spasm_error(LOG_ERR, "(spasm_cfg) Invalid WhitelistType '%s'", val);
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "regextype") == 0) {
            if(strcasecmp(val, "berkeleydb") == 0) spasm_conf->regexdb = BERKELEY_DB;
#ifdef MYSQL_DB
	    else if(strcasecmp(val, "mysql") == 0) spasm_conf->regexdb = MYSQL_DB;
#endif
            else {
		spasm_error(LOG_ERR, "(spasm_cfg) Invalid RegExType '%s'", val);
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "prefstype") == 0) {
	    if(strcasecmp(val, "berkeleydb") == 0) spasm_conf->prefsdb = BERKELEY_DB;
#ifdef MYSQL_DB
	    else if(strcasecmp(val, "mysql") == 0) spasm_conf->prefsdb = MYSQL_DB;
#endif
	    else {
		spasm_error(LOG_ERR, "(spasm_cfg) Invalid PrefsType '%s'", val);
		fclose(fp);
		return NULL;
	    }
        } else if(strcasecmp(key, "statstype") == 0) {
            if(strcasecmp(val, "berkeleydb") == 0) spasm_conf->statsdb = BERKELEY_DB;
#ifdef MYSQL_DB
            else if(strcasecmp(val, "mysql") == 0) spasm_conf->statsdb = MYSQL_DB;
#endif
            else {
                spasm_error(LOG_ERR, "(spasm_cfg) Invalid StatsType '%s'", val);
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "logstype") == 0) {
            if(strcasecmp(val, "text") == 0) spasm_conf->logsdb = TEXT_DB;
#ifdef MYSQL_DB
            else if(strcasecmp(val, "mysql") == 0) spasm_conf->logsdb = MYSQL_DB;
#endif
            else {
                spasm_error(LOG_ERR, "(spasm_cfg) Invalid LogsType '%s'", val);
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "verifyuserspasswd") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->verifypasswd = 1;
	    else spasm_conf->verifypasswd = 0;
        } else if(strcasecmp(key, "verifyusersaliases") == 0) {
            if(strcasecmp(val, "on") == 0) spasm_conf->verifyaliases = 1;
            else spasm_conf->verifyaliases = 0;
        } else if(strcasecmp(key, "verifyusersspasm") == 0) {
            if(strcasecmp(val, "on") == 0) spasm_conf->verifyspasm = 1;
            else spasm_conf->verifyspasm = 0;
        } else if(strcasecmp(key, "verifyuserscustom") == 0) {
            if(strcasecmp(val, "on") == 0) spasm_conf->verifycustom = 1;
            else spasm_conf->verifycustom = 0;
	} else if(strcasecmp(key, "aliasesfile") == 0) {
            spasm_conf->aliasesfile = strdup(val);
            if(spasm_conf->aliasesfile == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for AliasesFile: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "sendmail") == 0) {
            spasm_conf->sendmail = strdup(val);
            if(spasm_conf->sendmail == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Sendmail: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "unknownusersthreshold") == 0) {
	    spasm_conf->unknownusersthreshold = atoi(val);
	    seenthresh = 1;
	} else if(strcasecmp(key, "syslogrejections") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->syslogrejections = 1;
	    else spasm_conf->syslogrejections = 0;
	} else if(strcasecmp(key, "greylist") == 0) {
	    spasm_conf->greylist = strdup(val);
	    if(spasm_conf->greylist == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Greylist: %m");
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "greylistdelay") == 0) {
	    spasm_conf->greylistdelay = atoi(val) * 60; /* minutes -> seconds */
	} else if(strcasecmp(key, "greylistexpire") == 0) {
	    spasm_conf->greylistexpire = atoi(val) * 3600; /* hours -> seconds */
	} else if(strcasecmp(key, "greylistupdate") == 0) {
	    spasm_conf->greylistupdate = atoi(val) * 86400; /* days -> seconds */
	} else if(strcasecmp(key, "greylistexpireinterval") == 0) {
	    spasm_conf->greylistexpireinterval = atoi(val) * 60; /* minutes -> seconds */
	} else if(strcasecmp(key, "greylistforcebl") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->greylistforcebl = 1;
        } else if(strcasecmp(key, "dspam") == 0) {
            spasm_conf->dspam = strdup(val);
            if(spasm_conf->dspam == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for DSPAM: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "sqlserver") == 0) {
	    spasm_conf->sqlserver = strdup(val);
	    if(spasm_conf->sqlserver == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SQLServer: %m");
		fclose(fp);
		return NULL;
	    }
        } else if(strcasecmp(key, "sqluser") == 0) {
            spasm_conf->sqluser = strdup(val);
            if(spasm_conf->sqluser == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SQLUser: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "sqlpass") == 0) {
            spasm_conf->sqlpass = strdup(val);
            if(spasm_conf->sqlpass == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SQLPass: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "sqldb") == 0) {
            spasm_conf->sqldb = strdup(val);
            if(spasm_conf->sqldb == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SQLDB: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "sqlsock") == 0) {
            spasm_conf->sqlsock = strdup(val);
            if(spasm_conf->sqlsock == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SQLSock: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "sqlport") == 0) {
	    spasm_conf->sqlport = atoi(val);
	} else if(strcasecmp(key, "defaultquarantineexpire") == 0) {
	    spasm_conf->defaultqexpire = atoi(val);
	} else if(strcasecmp(key, "defaultlogexpire") == 0) {
	    spasm_conf->defaultlexpire = atoi(val);
	} else if(strcasecmp(key, "verifymx") == 0) {
	    spasm_conf->verifymx = strdup(val);
	    if(spasm_conf->verifymx == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for VerifyMX: %m");
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "verifymxexpire") == 0) {
	    spasm_conf->verifymxexpire = atoi(val) * 60; /* minutes -> seconds */
	} else if(strcasecmp(key, "verifymxexpireinterval") == 0) {
	    spasm_conf->verifymxexpireinterval = atoi(val) * 60; /* minutes -> seconds */
	} else if(strcasecmp(key, "verifymxmaxattempts") == 0) {
	    spasm_conf->verifymxmaxattempts = atoi(val);
	} else {
	    spasm_error(LOG_WARNING, "(spasm_cfg) Unknown configuration option: %s", key);
	}
    }
    fclose(fp);

    /* Now add mtime to spasm_conf */
    spasm_conf->mtime = filestat.st_mtime;

    /* Now verify that we have all the needed options */
    if(spasm_conf->spasmuser == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) SpasmUser not found");
        return NULL;
    }
    if(spasm_conf->defaultdomain == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) DefaultDomain not found");
	return NULL;
    }
    if(spasm_conf->spasmroot == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) SpasmRoot not found");
        return NULL;
    }
    if(spasm_conf->filters == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) Filters not found");
        return NULL;
    }
    if(spasm_conf->blacklists == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) Blacklists not found");
        return NULL;
    }
    if(spasm_conf->whitelists == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) Whitelists not found");
        return NULL;
    }
    if(spasm_conf->regex == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) RegEx not found");
        return NULL;
    }
    if(spasm_conf->prefs == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) Prefs not found");
	return NULL;
    }
    if(spasm_conf->logdir == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) LogDir not found");
        return NULL;
    }
    if(spasm_conf->quarantinedir == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) QuarantineDir not found");
        return NULL;
    }
    if(spasm_conf->tempdir == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) TempDir not found");
	return NULL;
    }
    if(spasm_conf->pidfile == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) PidFile not found");
        return NULL;
    }
    if(spasm_conf->socket == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) Socket not found");
	return NULL;
    }
    if(spasm_conf->sahost != NULL && spasm_conf->saport == 0) {
	spasm_error(LOG_ERR, "(spasm_cfg) SpamAssassinPort not found");
	return NULL;
    }
    if(spasm_conf->filterdb == 0) {
        spasm_error(LOG_ERR, "(spasm_cfg) FilterType not found");
        return NULL;
    }
    if(spasm_conf->blacklistdb == 0) {
        spasm_error(LOG_ERR, "(spasm_cfg) BlacklistType not found");
        return NULL;
    }
    if(spasm_conf->whitelistdb == 0) {
        spasm_error(LOG_ERR, "(spasm_cfg) WhitelistType not found");
        return NULL;
    }
    if(spasm_conf->regexdb == 0) {
        spasm_error(LOG_ERR, "(spasm_cfg) RegExType not found");
        return NULL;
    }
    if(spasm_conf->prefsdb == 0) {
	spasm_error(LOG_ERR, "(spasm_cfg) PrefsType not found");
	return NULL;
    }
    if(spasm_conf->logsdb == 0) {
	spasm_error(LOG_ERR, "(spasm_cfg) LogsType not found");
	return NULL;
    }
    if(spasm_conf->aliasesfile == NULL && spasm_conf->verifyaliases) {
        spasm_error(LOG_ERR, "(spasm_cfg) AliasesFile not found");
        return NULL;
    }
    if(spasm_conf->sendmail == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) Sendmail not found");
        return NULL;
    }
    if(spasm_conf->sqlserver != NULL) {
	if(spasm_conf->sqlport == 0 && spasm_conf->sqlsock == NULL) {
	    spasm_error(LOG_ERR, "(spasm_cfg) SQLServer set, but SQLPort or SQLSock not found");
	    return NULL;
	}
	if(spasm_conf->sqluser == NULL) {
	    spasm_error(LOG_ERR, "(spasm_cfg) SQLServer set, but SQLUser not found");
	    return NULL;
	}
	if(spasm_conf->sqlpass == NULL) {
	    spasm_error(LOG_ERR, "(spasm_cfg) SQLServer set, but SQLPass not found");
	    return NULL;
	}
	if(spasm_conf->sqldb == NULL) {
	    spasm_error(LOG_ERR, "(spasm_cfg) SQLServer set, but SQLDB not found");
	    return NULL;
	}
    }
    if(spasm_conf->spamtrapdb != NULL) {
	if(spasm_conf->spamtrapbl == NULL) {
	    spasm_error(LOG_ERR, "(spasm_cfg) SpamtrapDB set, but SpamtrapBlacklistDB not found");
	    return NULL;
	}
    }
    if(spasm_conf->greylist != NULL) {
        if(spasm_conf->greylistdelay == 0) spasm_conf->greylistdelay = 180;
        if(spasm_conf->greylistexpire == 0) spasm_conf->greylistexpire = 28800;
        if(spasm_conf->greylistupdate == 0) spasm_conf->greylistupdate = 604800;
	if(spasm_conf->greylistexpireinterval == 0) spasm_conf->greylistexpireinterval = 900;
    }
    if(spasm_conf->verifymx != NULL) {
	if(spasm_conf->verifymxexpire == 0) spasm_conf->verifymxexpire = 3600;
	if(spasm_conf->verifymxexpireinterval == 0) spasm_conf->verifymxexpireinterval = 900;
    }
    if(spasm_conf->defaultqexpire == 0) spasm_conf->defaultqexpire = 14;
    if(spasm_conf->defaultlexpire == 0) spasm_conf->defaultlexpire = 14;
    if(!seenthresh) spasm_conf->unknownusersthreshold = 5;

    return spasm_conf;
}
