/* Low-level MySQL functions */

void
mysql_quote_string(char *raw, char *quoted)
{   
    int i, size=0;
    for(i=0;i<strlen(raw);i++) {
        if(raw[i] == '\\') {
            quoted[size] = '\\';
            quoted[size+1] = '\\';
            size += 2;
        } else if(raw[i] == '\'') {
            quoted[size] = '\\';
            quoted[size+1] = '\'';
            size += 2;
        } else {
            quoted[size] = raw[i];
            size++;
        }
    }
    quoted[size] = '\0';
}

int
mysql_lookup_entry(char *query, char *value)
{   
    MYSQL *mysql;
    MYSQL *myres;
    MYSQL_RES *res;
    MYSQL_ROW row;
    unsigned long *lengths;

    if(query == NULL) return 0;

    if((mysql = mysql_init(NULL)) == NULL) {
        spasm_error(LOG_ERR, "(mysql_lookup_entry) Error initialising MySQL structure");
        return 0;
    }
    if((myres = mysql_real_connect(mysql, spasm_config->sqlserver, spasm_config->sqluser, spasm_config->sqlpass,
        spasm_config->sqldb, spasm_config->sqlport, spasm_config->sqlsock, 0)) == NULL) {
        spasm_error(LOG_ERR, "(mysql_lookup_entry) Error connecting to MySQL server: %s", mysql_error(mysql));
        mysql_close(mysql);
        return 0;
    }
    if(mysql_query(mysql, query) != 0) {
        spasm_error(LOG_ERR, "(mysql_lookup_entry) Error executing SQL query: %s", mysql_error(mysql));
        mysql_close(mysql);
        return 0;
    }
    if((res = mysql_store_result(mysql)) == NULL) {
        if(mysql_errno(mysql)) spasm_error(LOG_ERR, "(mysql_lookup_entry) Error retrieving SQL results: %s", mysql_error(mysql));
        mysql_close(mysql);
        return 0;
    }
    if(mysql_num_rows(res) < 1) {                                                                                                                    
        mysql_free_result(res);
        mysql_close(mysql);
        return 0;
    }
    if(value == NULL) {
        mysql_free_result(res);
        mysql_close(mysql);
        return 1;
    }
    if(mysql_num_rows(res) > 1) {
        spasm_error(LOG_WARNING, "(mysql_lookup_entry) SQL query returned more than one entry");
    }
    if((row = mysql_fetch_row(res)) == NULL) {
        if(mysql_errno(mysql)) spasm_error(LOG_ERR, "(mysql_lookup_entry) Error fetching row: %s", mysql_error(mysql));
        mysql_free_result(res);
        mysql_close(mysql);
        return 0;
    }
    if(mysql_num_fields(res) > 1) {
        spasm_error(LOG_WARNING, "(mysql_lookup_entry) SQL query returned more than one field");
    }                                                                                                                                                
    lengths = mysql_fetch_lengths(res);
    if(lengths[0] <= SPASM_DBT_MAX) strncpy(value, row[0], lengths[0]);
    mysql_free_result(res);
    mysql_close(mysql);
    return 1;
}

extern void
mysql_insert_entry(char *query)
{   
    MYSQL *mysql;
    MYSQL *myres;

    if(query == NULL) return;

    if((mysql = mysql_init(NULL)) == NULL) {
        spasm_error(LOG_ERR, "(mysql_insert_entry) Error initialising MySQL structure");
        return;
    }
    if((myres = mysql_real_connect(mysql, spasm_config->sqlserver, spasm_config->sqluser, spasm_config->sqlpass,
        spasm_config->sqldb, spasm_config->sqlport, spasm_config->sqlsock, 0)) == NULL) {
        spasm_error(LOG_ERR, "(mysql_insert_entry) Error connecting to MySQL server: %s", mysql_error(mysql));
        mysql_close(mysql);
        return;
    }
    if(mysql_query(mysql, query) != 0) {
        spasm_error(LOG_ERR, "(mysql_insert_entry) Error executing SQL query: %s", mysql_error(mysql));
    }
    mysql_close(mysql);
}

void
mysql_delete_entry(char *query)
{   
    MYSQL *mysql;
    MYSQL *myres;

    if(query == NULL) return;

    if((mysql = mysql_init(NULL)) == NULL) {
        spasm_error(LOG_ERR, "(mysql_delete_entry) Error initialising MySQL structure");
        return;
    }
    if((myres = mysql_real_connect(mysql, spasm_config->sqlserver, spasm_config->sqluser, spasm_config->sqlpass,
        spasm_config->sqldb, spasm_config->sqlport, spasm_config->sqlsock, 0)) == NULL) {
        spasm_error(LOG_ERR, "(mysql_delete_entry) Error connecting to MySQL server: %s", mysql_error(mysql));
        mysql_close(mysql);
        return;
    }
    if(mysql_query(mysql, query) != 0) {
        spasm_error(LOG_ERR, "(mysql_delete_entry) Error executing SQL query: %s", mysql_error(mysql));
    }
    mysql_close(mysql);
}

struct spasm_list*
mysql_dump_entries(char *query, int vals)
{   
    MYSQL *mysql;
    MYSQL *myres;
    MYSQL_RES *res;
    MYSQL_ROW row;
    unsigned long *lengths;
    struct spasm_list *entries = NULL;
    struct spasm_list *first = NULL;

    if(query == NULL) return NULL;

    if((mysql = mysql_init(NULL)) == NULL) {
        spasm_error(LOG_ERR, "(mysql_dump_entries) Error initialising MySQL structure");
        return NULL;
    }
    if((myres = mysql_real_connect(mysql, spasm_config->sqlserver, spasm_config->sqluser, spasm_config->sqlpass,
        spasm_config->sqldb, spasm_config->sqlport, spasm_config->sqlsock, 0)) == NULL) {
        spasm_error(LOG_ERR, "(mysql_dump_entries) Error connecting to MySQL server: %s", mysql_error(mysql));
        mysql_close(mysql);
        return NULL;
    }
    if(mysql_query(mysql, query) != 0) {
        spasm_error(LOG_ERR, "(mysql_dump_entries) Error executing SQL query: %s", mysql_error(mysql));
        mysql_close(mysql);
        return NULL;
    }
    if((res = mysql_store_result(mysql)) == NULL) {                                                                                                  
        if(mysql_errno(mysql)) spasm_error(LOG_ERR, "(mysql_dump_entries) Error retrieving SQL results: %s", mysql_error(mysql));
        mysql_close(mysql);
        return NULL;
    }
    if(mysql_num_fields(res) > 2) {
        spasm_error(LOG_WARNING, "(mysql_dump_entries) SQL query returned more than two fields");
    }
    while((row = mysql_fetch_row(res))) {
        lengths = mysql_fetch_lengths(res);
        if(first == NULL) {
            if((first = calloc(sizeof(struct spasm_list), 1)) != NULL) {
                if(lengths[0] <= SPASM_DBT_MAX) {
                    entries = first;
                    strncpy(entries->entry, row[0], lengths[0]);
                    if(vals) {
                        if(lengths[1] <= SPASM_DBT_MAX) strncpy(entries->strval, row[1], lengths[1]);
                        else spasm_error(LOG_ERR, "(mysql_dump_entries) Value is too large (%d)", lengths[1]);
                    }
                } else spasm_error(LOG_ERR, "(mysql_dump_entries) Entry is too large (%d)", lengths[0]);
            } else spasm_error(LOG_ERR, "(mysql_dump_entries) Error allocating memory for first: %m");
        } else {
            if((entries->next = calloc(sizeof(struct spasm_list) + 1, 1)) != NULL) {
                entries = entries->next;
                if(lengths[0] <= SPASM_DBT_MAX) {
                    strncpy(entries->entry, row[0], lengths[0]);
                    if(vals) {                                                                                                                       
                        if(lengths[1] <= SPASM_DBT_MAX) strncpy(entries->strval, row[1], lengths[1]);
                        else spasm_error(LOG_ERR, "(mysql_dump_entries) Value is too large (%d)", lengths[1]);
                    }
                } else spasm_error(LOG_ERR, "(mysql_dump_entries) Entry is too large (%d)", lengths[0]);
            } else spasm_error(LOG_ERR, "(mysql_dump_entries) Error allocating memory for entries->next: %m");
        }
    }
    if(mysql_errno(mysql)) spasm_error(LOG_ERR, "(mysql_dump_entries) Error retrieving entries: %s", mysql_error(mysql));
    mysql_free_result(res);
    mysql_close(mysql);
    return first;
}
