/* Low-level QDBM functions */

/* qdbm_lookup_entry() - lookup a qdbm hash entry, return the value */
int
qdbm_lookup_entry(char *dbfile, char *key, char *value)
{
    DEPOT *db;
    char buf[SPASM_DBT_MAX+1];
    int vsize;

    if(dbfile == NULL || key == NULL) return 0;

    if((db = dpopen(dbfile, DP_OREADER, -1)) == NULL) {
	spasm_error(LOG_ERR, "(qdbm_lookup_entry) Error opening database %s: %s", dbfile, dperrmsg(dpecode));
	return 0;
    }

    memset(buf, 0, sizeof(buf));
    if((vsize = dpgetwb(db, key, strlen(key), 0, SPASM_DBT_MAX, buf)) == -1) {
        spasm_error(LOG_ERR, "(qdbm_lookup_entry) Error getting value from database %s: %s", dbfile, dperrmsg(dpecode));
	dpclose(db);
        return 0;
    }
    if(value != NULL) strncpy(value, buf, vsize);
    dpclose(db);
    return 1;
}

/* qdbm_insert_entry() - insert a qdbm hash entry */
void
qdbm_insert_entry(char *dbfile, char *key, char *value)
{     
    DEPOT *db;

    if(dbfile == NULL || key == NULL || value == NULL) return;

    if((db = dpopen(dbfile, DP_OWRITER|DP_OCREAT, -1)) == NULL) {
        spasm_error(LOG_ERR, "(qdbm_insert_entry) Error opening database %s: %s", dbfile, dperrmsg(dpecode));
        return;
    }
    if(dpput(db, key, strlen(key), value, strlen(value), DP_DOVER) == 0) {
        spasm_error(LOG_ERR, "(qdbm_insert_entry) Error adding %s/%s to %s: %s", key, value, dbfile, dperrmsg(dpecode));
    }
    if(dpclose(db) == 0) {
        spasm_error(LOG_ERR, "(qdbm_insert_entry) Error closing database %s: %s", dbfile, dperrmsg(dpecode));
    }
    return;
}

/* qdbm_delete_entry() - delete a qdbm hash entry */
void
qdbm_delete_entry(char *dbfile, char *key)
{     
    DEPOT *db;

    if(dbfile == NULL || key == NULL) return;

    if((db = dpopen(dbfile, DP_OWRITER|DP_OCREAT, -1)) == NULL) {
        spasm_error(LOG_ERR, "(qdbm_delete_entry) Error opening database %s: %s", dbfile, dperrmsg(dpecode));
        return;
    }
    if(dpout(db, key, strlen(key)) == 0) {
        spasm_error(LOG_ERR, "(qdbm_delete_entry) Error deleting %s from %s: %s", key, dbfile, dperrmsg(dpecode));
    }
    if(dpclose(db) == 0) {
        spasm_error(LOG_ERR, "(qdbm_delete_entry) Error closing database %s: %s", dbfile, dperrmsg(dpecode));
    }
    return;
}

/* qdbm_dump_entries() - dump spasm_list of entries in qdbm database */
struct spasm_list*
qdbm_dump_entries(char *dbfile, int vals)
{   
    struct spasm_list *entries = NULL;
    struct spasm_list *first = NULL;
    DEPOT *db;
    char *key;
    char buf[SPASM_DBT_MAX+1];
    int ksize;
    int vsize;

    if(dbfile == NULL) return NULL;

    if((db = dpopen(dbfile, DP_OREADER, -1)) == NULL) {
        spasm_error(LOG_ERR, "(qdbm_dump_entries) Error opening database %s: %s", dbfile, dperrmsg(dpecode));
        return NULL;
    }
    if(dpiterinit(db) == 0) {
        spasm_error(LOG_ERR, "(qdbm_dump_entries) Error initialising iterator for %s: %s", dbfile, dperrmsg(dpecode));
	dpclose(db);
        return NULL;
    }

    while((key = dpiternext(db, &ksize)) != NULL) {
        if(first == NULL) {
            if((first = calloc(sizeof(struct spasm_list), 1)) != NULL) {
                if(ksize <= SPASM_DBT_MAX) {
                    entries = first;
                    entries->next = NULL;
                    strncpy(entries->entry, key, ksize);
                    if(vals) {
			memset(buf, 0, sizeof(buf));
			if((vsize = dpgetwb(db, key, strlen(key), 0, SPASM_DBT_MAX, buf)) == -1) {
			    spasm_error(LOG_ERR, "(qdbm_dump_entries) Error getting value from database %s: %s", dbfile, dperrmsg(dpecode));
			} else {
			    strncpy(entries->strval, buf, vsize);
			}
                    }
                } else spasm_error(LOG_ERR, "(qdbm_dump_entries) Entry is too large");
            } else spasm_error(LOG_ERR, "(qdbm_dump_entries) Error allocating memory for first: %m");
        } else {
            if((entries->next = calloc(sizeof(struct spasm_list) + 1, 1)) != NULL) {
                entries = entries->next;
                entries->next = NULL;
                if(ksize <= SPASM_DBT_MAX) {
                    strncpy(entries->entry, key, ksize);
                    if(vals) {
                        memset(buf, 0, sizeof(buf));                                             
                        if((vsize = dpgetwb(db, key, strlen(key), 0, SPASM_DBT_MAX, buf)) == -1) {                                         
                            spasm_error(LOG_ERR, "(qdbm_dump_entries) Error getting value from database %s: %s", dbfile, dperrmsg(dpecode));
                        } else {                                
                            strncpy(entries->strval, buf, vsize);
                        }                                                                         
                    }
                } else spasm_error(LOG_ERR, "(qdbm_dump_entries) Entry is too large");
            } else spasm_error(LOG_ERR, "(qdbm_dump_entries) Error allocating memory for entries->next: %m");
        }
    }
    dpclose(db);
    return first;
}
