/*
**    spasm - The spasm anti-spam milter
**    Copyright (C) 2002-2005 Sean Hanson <aradia@theasylum.org>
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <ctype.h>
#include <curses.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <sysexits.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "spasm.h"
#include "getopt-portable.h"

#ifdef SUNOS
  #define getnstr(a,b)		wgetnstr(stdscr,a,b)
  #define vw_printw		vwprintw
#endif /* SUNOS */

#define SPASM_USER_INTERFACE
#include "filters.h"

/* dangerous macros */
#define min(a, b) (((a) < (b)) ? (a) : (b))
#define max(a, b) (((a) > (b)) ? (a) : (b))

#define ITEM_OPT    0     /* filter option    */
#define ITEM_PARAM  1     /* filter parameter */
#define ITEM_FILTER 2     /* filter           */

WINDOW *mw = NULL;
int xmax, ymax;
char *USERNAME = NULL;
static char *progname;

struct spasmConfigStruct *spasm_config;

const static char anykey[]        = "Press any key to continue";
const static char cancel_hlp[]    = "C[ancel] or X[it] -> cancels this prompt";
const static char bh_prefix_hlp[] =
    "This list is meant to contain all known networks assigned to ";
const static char bh_suffix_hlp[] =
    "\n\nNote that a \"blackhole.us\" filter does not imply that the "
    "affected ISP/country is a source of spam or prone to abuse. "
    "This filter simply encompasses the network address space of the "
    "target entity.";

const static char *set_filter_prompt_help[] = 
    {"L[ogging] -> toggles logging for the filter",
     "",
     "D[isable] -> disables the filter",
     "",
     "T[ag] -> enables the filter; when triggered, the affected",
     "         email is Tagged with \"X-spasm\" headers that",
     "         denote which filter(s) caught the apparent spam.",
     "         The modified message is then delivered to the user.",
     "",
     "Q[uarantine] -> enables the filter; when triggered the",
     "                affected email is tagged with \"X-spasm\"",
     "                headers as above and stored for the recipient", 
     "                for later retrieval.",
     "",
     "R[eject] -> enables the filter; when triggered, the affected",
     "            email is rejected (not delivered, not saved).",
     NULL};

const static char *set_filter_option_prompt_help[] = 
    {"D[isable] -> disables the option",
     "",
     "E[nable] -> enables the option",
     NULL};

/*
const static char *set_filter_param_uint_prompt_help[] = 
    {"Enter an unsigned integer value from 0 to N, where N",
     "makes sense for the parameter in question.",
     NULL};
*/

const static char *general_filter_help[] = 
    {"This menu displays the current settings for all supported antispam",
     "filters. The first checkbox shows the filter's log setting, and",
     "the second checkbox shows the filter's status setting. Visual",
     "settings representations are:",
     "",
     "-       : filter/logging disabled",
     "L       : logging enabled.",
     "T, Q, R : filter enabled.  When a filter triggers, the affected",
     "          email is either:",
     "",
     "          T[agged] with X-spasm headers and delivered to",
     "              recipient, or", 
     "          Q[uarantined] for later retrieval (message is",
     "              also tagged as above), or",
     "          R[ejected] -- not delivered, not saved.",
     "",
     "To modify a filter's current setting, enter its display column",
     "number (integer to left of '.') at this prompt.",
     "",
     "To see a short description of a filter's purpose, type 'h' and",
     "then its display column number. For example, to learn more about",
     "filter number 2, type:    h2 <enter>",
     NULL};

const static char *general_prefs_help[] =
    {"This menu displays the current preferences. Visual settings",
     "representations are:",
     "",
     "-       : option disabled",
     "+       : option enabled",
     "",
     "To modify an option's current setting, enter its display column",
     "number (integer to left of '.') at this prompt.",
     "",
     "To see a short description of an option's purpose, type 'h' and",
     "then its display column number. For example, to learn more about",
     "option number 2, type:    h2 <enter>",
     NULL};

static struct spasm_list*
add_log_list(struct spasm_list *cur, char *data)
{
    struct spasm_list *new;

    if ((new = calloc(sizeof(struct spasm_list), 1)) == NULL) {
        spasm_error(LOG_ERR, "Error allocating memory for next log entry: %m");
        return NULL;
    }
    snprintf(new->entry, sizeof(new->entry), "%s", data);
    new->next = NULL;
    if (cur != NULL ) cur->next = new;
    return new;
}

static void
clean_log_list(struct spasm_list *first)
{
    struct spasm_list *cur = first;

    while (cur != NULL) {
        first = cur->next;
        free(cur);
        cur = first;
    }
}

static void
lcase(char *str)
{
    if (str) {
        while (*str) {
            if (isupper(*str))
                *str = tolower(*str);
            str++;
        }
    }
}

static char *
skipwhite(char *buf)
{
    if (buf) {
        while (isspace(*buf))
            buf++;
    }
    return (buf);
}

/* note: calls cbreak() and noecho() */
static void
show_simple_filter_menu_help(const char *hlpstrings[], int add_cancel_help)
{
    int        add_blank_rows, height, x, y, nstr, width, starty, startx,
               maxlen, maxtext_width;
    WINDOW     *helpwdw;
    const char **hptr;
    size_t     tlen;

    nstr   = 0;
    hptr   = hlpstrings;
    maxlen = (add_cancel_help) ? sizeof(cancel_hlp) - 1 : sizeof(anykey) - 1;
    while (*hptr) {
        nstr++;
        tlen = strlen(*hptr);
        if (tlen > maxlen)
            maxlen = tlen;
        hptr++;
    }
    nstr += 4;             /* account for "\n<anykey str>", box borders. */
    if (add_cancel_help)
        nstr += 2;
    if (nstr + 2 <= LINES) {
        nstr += 2;         /* add blank rows after box borders */
        add_blank_rows = 1;
    } else {
        add_blank_rows = 0;
    }
    height = min(nstr, LINES);
    maxtext_width = maxlen + 4;  /* 4 => 2 border cols + 2 indents */
    width  = (COLS > maxtext_width) ? maxtext_width : COLS;
    starty = (LINES - height) / 2 ; 
    startx = (COLS - width) / 2 ; 

    helpwdw = newwin(height, width, starty, startx);
    box(helpwdw, 0, 0);
    x = 2;
    y = (add_blank_rows) ? 2 : 1;
    hptr = hlpstrings;
    while (*hptr) {
        mvwaddstr(helpwdw, y++, x, *hptr);
        hptr++;
    }
    if (add_cancel_help) {
        y++;                                  /* insert blank line */
        mvwaddstr(helpwdw, y++, x, cancel_hlp);
    }
    y++;                                      /* insert blank line */
    wattron(helpwdw, A_REVERSE);
    x = (width - (sizeof(anykey) - 1)) / 2;   /* center string */
    mvwaddstr(helpwdw, y, x, anykey);
    wmove(helpwdw, height - 2, 1);            /* park the cursor */
    wrefresh(helpwdw);
    cbreak();
    noecho();
    getch();
    werase(helpwdw);
    touchwin(stdscr);
    wrefresh(helpwdw);
    refresh();
    delwin(helpwdw);
}

/*
 * Formats a filter/option/parameter help string in a generally pleasing way.
 * Help text may include newline characters to force line breaks, but tabs
 * are _not_ (I say, NOT) supported.
 *
 * It's just a small matter of code....
 *
 * note: calls cbreak() and noecho()
 *
 * Preconditions: "mnuitem" is a valid index to fltr_items[].
 */
static void
show_complex_filter_menu_help(int mnuitem, int itemtype)
{
    int              height, i, isbhfilter, est1, est2, maxx, width, nlines,
                     startx, starty, x, y;
    char             *hlpstr, *nxtnl, *nxtspc, *saved_hlpstr = NULL, *tmp;
    WINDOW           *helpwdw;
    size_t           nchar, nwords, wordwidth;
    const SPASM_FILTER *pfltr;
    const SPASM_FILTER_OPTION *popt;

    isbhfilter = 0;
    if(itemtype == ITEM_OPT) {
	popt = spasm_filter_opts + mnuitem;
	hlpstr = popt->help;
    } else if(itemtype == ITEM_FILTER) {
	pfltr = spasm_filters + mnuitem;
	isbhfilter = pfltr->is_blackholes_filter;
	hlpstr     = pfltr->help;
    }

    tmp = hlpstr;

    if (isbhfilter) {
        /*
         * prepend invariant help text and terminating invariant suffix to
         * remainder of filter's help msg.
         */

        size_t n;

        n = strlen(hlpstr) + strlen(bh_prefix_hlp) + strlen(bh_suffix_hlp) + 2;
        tmp = malloc(n);
        if (tmp == NULL) {
            spasm_error(LOG_ERR, "Error allocating memory for filter help: %m");
            sleep(5);  /* let user see message... */
            return;
        }
        snprintf(tmp, n, "%s%s.%s", bh_prefix_hlp, hlpstr, bh_suffix_hlp);
        saved_hlpstr = hlpstr = tmp;
    }

    /* precondition: tmp == hlpstr */

    nchar = nlines = nwords = 0;
    while (*tmp) {
        if (*tmp == '\n') {
            tmp++;
            nlines++;
        } else {
            nchar++;
            if (*tmp == ' ')
                nwords++;
            tmp++;
        }
    }
    if (COLS - 18 > sizeof(anykey))
        width = COLS - 18;
    else if (COLS - 10 > sizeof(anykey))
        width = COLS - 10;
    else
        width = COLS;
    if (width < sizeof(anykey)) {
        if (isbhfilter)
            free(saved_hlpstr);
        return;     /* hopeless, pointless, etc. */
    }

    nlines += 6; /* box borders + empty line delimiting each border +
                  * "\n<anykey>" message.
                  */
    /* 
     * in code below...
     *   "- 4" accounts for box borders and 2 char indents
     *   "* 5" is an estimate of 5 chars per word
     *   "+ 1" picks up a remainder from integer math
     */
    est1    = ((nwords * 5) / (width - 4)) + 1;
    est2    = (nchar / (width - 4)) + 1;
    nlines += max(est1, est2);
    height  = min(nlines, LINES);
    starty  = (LINES - height) / 2;
    startx  = (COLS - width) / 2;

    helpwdw = newwin(height, width, starty, startx);
    box(helpwdw, 0, 0);
    x = y = 2;
    maxx = width - 2;     /* right box border, 1 char indent */
    while (*hlpstr) {
        if (x >= maxx) {
            x = 2;
            y++;
        }
        if (*hlpstr == '\n') {
            /* trivial case */

            x = 2;
            y++;
            hlpstr++;
            continue;
        }
        if (*hlpstr == ' ') {
            /* trivial case */

            if (x != 2) /* if not at BOL */
                mvwaddch(helpwdw, y, x++, *hlpstr);
            hlpstr++;
            continue;
        }
        nxtnl  = strchr(hlpstr, '\n');
        nxtspc = strchr(hlpstr, ' ');
        if (nxtnl || nxtspc) {
            if (nxtnl && nxtspc) {
                if (nxtnl < nxtspc)
                    wordwidth = nxtnl - hlpstr;
                else
                    wordwidth = nxtspc - hlpstr;
            } else if (nxtnl) {
                wordwidth = nxtnl - hlpstr;
            } else {
                wordwidth = nxtspc - hlpstr;
            }
            if (wordwidth + x >= maxx) {
                /* 
                 * Can't fit current word on this line.  Ck for corner case
                 * where the entire word won't fit in the window.
                 */

                if (x == 2) {
                    break;   /* bail out...administrator must use shorter
                              * words or end user use bigger windows
                              */
                }
                x = 2;
                y++;
                continue;
            } else {
                for (i = 0; i < wordwidth; i++) {
                    mvwaddch(helpwdw, y, x++, *hlpstr);
                    hlpstr++;
                }
            }
        } else {
            /*
             * no more delimiters left to break a line.  Will the last
             * "word" in the string fit, or require a new line?
             */

            wordwidth = strlen(hlpstr);
            if (wordwidth + x >= maxx) {
                /* cut n paste some code directly from above */

                if (x == 2)
                    break;  /* At BOL, word won't fit */
                x = 2;
                y++;
            }
            for (i = 0; i < wordwidth; i++) {
                mvwaddch(helpwdw, y, x++, *hlpstr);
                hlpstr++;
            }
        }
    }
    if (isbhfilter) {
        free(saved_hlpstr);
        saved_hlpstr = NULL;
    }

    y += 2;                                   /* insert blank line */
    wattron(helpwdw, A_REVERSE);
    x = (width - (sizeof(anykey) - 1)) / 2;   /* center string */
    mvwaddstr(helpwdw, y, x, anykey);
    wmove(helpwdw, height - 2, 1);            /* park the cursor */
    wrefresh(helpwdw);
    cbreak();
    noecho();
    getch();
    werase(helpwdw);
    touchwin(stdscr);
    wrefresh(helpwdw);
    refresh();
    delwin(helpwdw);
}

/*
 * Permit the user to specify a filter menu item's value.
 *
 * Preconditions: "numitem" is a valid index to spasm_filters.
 */
static void
set_filter_menu_value(int numitem, int xmax, int ymax)
{
    int              blanks, done;
    char             buf[128], *answer;
    const SPASM_FILTER *pfltr;
    unsigned         val;

    pfltr = spasm_filters + numitem;

    done = 0;
    do {
        move(ymax-2, 1);
        clrtoeol();
        mvvline(ymax - 2, xmax - 1, 0, 1);  /* repair border */
        blanks = (strchr(pfltr->ui_name, ' ') != NULL);
        mvprintw(ymax-2, 1, "Set %s%s%s to [L, D, T, Q, R - ? for help]: ", (blanks) ? "\"" : "", pfltr->ui_name, (blanks) ? "\"" : "");
        crmode();
        echo();
        getnstr(buf, sizeof(buf));
        answer = skipwhite(buf);
        lcase(answer);
        noecho();
        switch (*answer)
        {
            case 'c': case 'x':          /* cancel prompt */
                done = 1;
                break;
            case 'd':                    /* disable filter/option */
                spasm_set_filter(USERNAME, pfltr->db_name, SPASM_OFF);
                done = 1;
                break;
            case '?':
                {
                    const char **hlparray;
                    hlparray = set_filter_prompt_help;
                    show_simple_filter_menu_help(hlparray, 1);
                }
                break;
	    case 'l':          /* toggle logging of filter */
            case 'q':          /* filter operates in quarantine mode */
            case 't':          /* filter operates in "tagged" mode   */
            case 'r':          /* filter operates in "reject" mode   */
		if (*answer == 'l') {
		    val = spasm_lookup_filter(USERNAME, pfltr->db_name);
		    if(val & SPASM_LOG) val ^= SPASM_LOG;
		    else val |= SPASM_LOG;
		} else if (*answer == 'q') {
		    val = spasm_lookup_filter(USERNAME, pfltr->db_name);
		    if(val & SPASM_LOG) val = SPASM_LOG|SPASM_QUARANTINE;
		    else val = SPASM_QUARANTINE;
		} else if (*answer == 't') {
		    val = spasm_lookup_filter(USERNAME, pfltr->db_name);
		    if(val & SPASM_LOG) val = SPASM_LOG|SPASM_TAG;
		    else val = SPASM_TAG;
		} else {
		    val = spasm_lookup_filter(USERNAME, pfltr->db_name);
		    if(val & SPASM_LOG) val = SPASM_LOG|SPASM_REJECT;
		    else val = SPASM_REJECT;
		}
                spasm_set_filter(USERNAME, pfltr->db_name, val);
                done = 1;
                break;
            case '0':  case '1':  case '2':  case '3':  case '4':
            case '5':  case '6':  case '7':  case '8':  case '9':
		if(strcmp("SF_SPAMASSASSIN", pfltr->db_name) == 0) {
		    /* FIXME -- atoi allows overflow... */

		    val = atoi(answer);
		    spasm_set_filter(USERNAME, "SA_REQUIRED_HITS", val);
		    done = 1;
		} else {
                    beep();    /* valid only for UINT parameter option */
		}
                break;
            default:
                beep();
                break;
        }
    } while (! done);
}

/*
 * Permit the user to specify a preference menu item's value.
 *
 * Preconditions: "numitem" is a valid index to spasm_filter_opts[].
 */
static void                                                                                                                                           
set_prefs_menu_value(int numitem, int xmax, int ymax)
{   
    int              blanks, done;
    char             buf[128], *answer;
    const SPASM_FILTER_OPTION *popt;

    popt = spasm_filter_opts + numitem;

    done = 0;
    do {
        move(ymax-2, 1);
        clrtoeol();
        mvvline(ymax - 2, xmax - 1, 0, 1);  /* repair border */
        blanks = (strchr(popt->ui_name, ' ') != NULL);
        mvprintw(ymax-2, 1, "Set %s%s%s to [D, E - ? for help]: ", (blanks) ? "\"" : "", popt->ui_name, (blanks) ? "\"" : "");

	/* UINT parameters
            mvprintw(ymax-2, 1, "Set %s%s%s to [0 - N, ? for help]: ", (blanks) ? "\"" : "", popt->ui_name, (blanks) ? "\"" : "");
	*/

        crmode();
        echo();
        getnstr(buf, sizeof(buf));
        answer = skipwhite(buf);
        lcase(answer);
        noecho();
        switch (*answer)
        {
            case 'c': case 'x':          /* cancel prompt */
                done = 1;
                break;
            case 'd':                    /* disable filter/option */
                spasm_set_prefs(USERNAME, popt->db_name, "0");
                done = 1;
                break;
            case '?':
                {   
                    const char **hlparray;
                    
                    hlparray = set_filter_option_prompt_help;
                    show_simple_filter_menu_help(hlparray, 1);
                }
                break;
            case 'e':          /* enable option */
                spasm_set_prefs(USERNAME, popt->db_name, "1");
                done = 1;
                break;
            case '0':  case '1':  case '2':  case '3':  case '4':
            case '5':  case '6':  case '7':  case '8':  case '9':
                /*if(strcmp("qexpire", popt->db_name) == 0) {*/
		/* FIXME: atoi can overflow */
		/*
                    val = atoi(answer);
                    spasm_set_prefs(USERNAME, "qexpire", val);
                    done = 1;
                } else {
		*/
                    beep();    /* valid only for UINT parameter option */
                /*}*/
                break;
            default:
                beep();
                break;
        }
    } while (! done);
}

void
spasm_error(int syslevel, char *errstr, ...)
{
    va_list arglist;

#ifdef SUNOS
    char *at, message[1024], temp[1024];

    strlcpy(message, errstr, sizeof(message));
    while ((at=strstr(message, "%m")) != NULL) {
        *at = '\0';
        strlcpy(temp, at+2, sizeof(temp));
        strlcat(message, strerror(errno), sizeof(message));
        strlcat(message, temp, sizeof(message));
    }
#endif /* SUNOS */

    if(mw == NULL) {
	va_start(arglist, errstr);
#ifdef SUNOS
	vfprintf(stderr, message, arglist);
#else
	vfprintf(stderr, errstr, arglist);
#endif /* SUNOS */
	fprintf(stderr, "\n");	/* spasm_error args aren't terminated by \n */
	va_end(arglist);
    } else {
        move(ymax-3, 1);
        clrtoeol();
        va_start(arglist, errstr);
#ifdef SUNOS
	vw_printw(mw, message, arglist);
#else
        vw_printw(mw, errstr, arglist);
#endif /* SUNOS */

	/* mvvline's added for everything, or just Solaris? */
	mvvline(ymax - 3, xmax - 1, 0, 1);  /* repair border */
        va_end(arglist);
        move(ymax-2, 1);
        clrtoeol();
        printw("%s", "Press a key to continue...");
	mvvline(ymax - 3, xmax - 1, 0, 1);  /* repair border */
        getch();
        move(ymax-3, 1);
        clrtoeol();
	mvvline(ymax - 3, xmax - 1, 0, 1);  /* repair border */
        move(ymax-2, 1);
        clrtoeol();
	mvvline(ymax - 3, xmax - 1, 0, 1);  /* repair border */
    }
}

static void
cleanup(int sig)
{
    endwin();
    printf("\n");
    if(USERNAME != NULL) free(USERNAME);
    exit(0);
}

static void
changeUsername(void)
{
    char buf[128];
    clear();
    box(mw, ACS_VLINE, ACS_HLINE);
    getmaxyx(stdscr, ymax, xmax);
    mvprintw(1, ((xmax/2)-8), "Change Username");
    mvprintw(ymax-2, 1, "Username: ");
    echo();
    getnstr(buf, sizeof(buf));
    noecho();
    if(strlen(buf) == 0) return;
    free(USERNAME);
    if((USERNAME = calloc(strlen(buf)+1, 1)) == NULL) {
	spasm_error(LOG_ERR, "Error allocating memory for username: %m");
    } else {
	sprintf(USERNAME, "%s", buf);
    }
}

static int
doScreen(char *header, struct spasm_list *first, int start)
{
    struct spasm_list *cur;
    int i, line, center=xmax/2;

    clear();
    box(mw, ACS_VLINE, ACS_HLINE);
    getmaxyx(stdscr, ymax, xmax);
    mvprintw(1, center-strlen(header)/2, header);

    for(cur=first, i=0; cur!=NULL && i<start; cur=cur->next, i++);
    if(cur == NULL) {
	cur = first;
	i = 0;
    }
    for(line=3; cur!=NULL && line<ymax-3; cur=cur->next, i++) {
	if(strcmp(header, "Blacklist Menu") == 0 || strcmp(header, "Whitelist Menu") == 0)
	    mvprintw(line, 2, "%s", cur->entry);
	else
	    mvprintw(line, 2, "%d. %s", i+1, cur->entry);
	line++;
    }
    return(i);
}

static void
viewLogEntry(char *entry)
{
    char *ptr, *ptr2, *ptr3;
    int line = 3, i = 0;

    clear();
    box(mw, ACS_VLINE, ACS_HLINE);
    getmaxyx(stdscr, ymax, xmax);
    mvprintw(1, ((xmax/2)-7), "View Log Entry");

    ptr = entry;
    while((ptr2 = strchr(ptr, ':')) != NULL) {
	*ptr2 = '\0';
	switch(i) {
	    case 0:
		/* Timestamp */
		break;
	    case 1:
		/* Date */
		mvprintw(line, 2, "Date: %s", ptr);
		break;
	    case 2:
		/* Date, cont'd */
		printw(":%s", ptr);
		break;
	    case 3:
		/* Date, cont'd */
		printw(":%s", ptr);
		line++;
		break;
	    case 4:
		/* Recipient */
		mvprintw(line, 2, "Recipient: %s", ptr);
		line++;
		break;
	    case 5:
		/* Sender */
		mvprintw(line, 2, "Sender: %s", ptr);
		line++;
		break;
	    case 6:
		/* Hostname */
		mvprintw(line, 2, "Server Hostname: %s", ptr);
		line++;
		break;
	    case 7:
		/* IP Address */
		mvprintw(line, 2, "Server IP Address: %s", ptr);
		line++;
		break;
	    case 8:
		/* HELO Hostname */
		mvprintw(line, 2, "Server HELO: %s", ptr);
		line++;
		break;
	    case 9:
		/* Filter[s] */
		if((ptr3 = strrchr(ptr, '|')) != NULL) *ptr3 = '\0';
		mvprintw(line, 2, "Filters: %s", ptr);
		if(ptr3 != NULL) *ptr3 = '|';
		line++;
		break;
	    default:
		break;
	}
	*ptr2 = ':';
	ptr = ptr2 + 1;
	i++;
    }
    /* Status */
    i = atoi(ptr);
    if(i == 1) mvprintw(line, 2, "Status: Tagged");
    else if(i == 2) mvprintw(line, 2, "Status: Quarantined");
    else if(i == 4) mvprintw(line, 2, "Status: Rejected");
    else mvprintw(line, 2, "Status: [Unknown Status]");

    mvprintw(ymax-2, 1, "(%s) Press any key to continue...", USERNAME);
    getch();
}

static void
rlMenu(void)
{
    int option, i, start = 0, line, j;
    char buf[128];
    struct spasm_list *first = NULL;
    struct spasm_list *cur = NULL;

    first = spasm_get_log(USERNAME, NULL, 1);
    cur = first;

    do {
	option = 0;

	clear();
	box(mw, ACS_VLINE, ACS_HLINE);
	getmaxyx(stdscr, ymax, xmax);
	mvprintw(1, ((xmax/2)-7), "Rejection Logs");
	for(cur = first, i = 0; cur != NULL && i < start; cur = cur->next, i++);
	if(cur == NULL) {
	    cur = first;
	    i = 0;
	}
	for(line = 3; cur != NULL && line < ymax-3; cur = cur->next, i++) {
	    char *ptr, *ptr2;
	    char datestr[32];
	    char sender[SMTP_PATH_MAX+1];

	    memset(datestr, 0, sizeof(datestr));
	    memset(sender, 0, sizeof(sender));
	    /* Grab date string */
	    if((ptr = strchr(cur->strval, ':')) != NULL) {
		ptr++;
		for(j=0, ptr2=ptr; j<3 && ptr2 != NULL; j++, ptr2++) ptr2 = strchr(ptr2, ':');
		if(ptr2 != NULL) {
		    ptr2--;
		    *ptr2 = '\0';
		    strncpy(datestr, ptr, sizeof(datestr)-1);
		    *ptr2 = ':';
		    ptr2++;
		    /* Grab sender */
		    if((ptr = strchr(ptr2, ':')) != NULL) {
			ptr++;
			if((ptr2 = strchr(ptr, ':')) != NULL) {
			    *ptr2 = '\0';
			    strncpy(sender, ptr, sizeof(sender)-1);
			    *ptr2 = ':';
			}
		    }
		}
	    }
	    mvprintw(line, 2, "%d. %s\t%s", i+1, datestr, sender);
	    line++;
	}

	move(ymax-2, 1);
	clrtoeol();
	mvvline(ymax - 2, xmax - 1, 0, 1);  /* repair border */
	mvprintw(ymax-2, 1, "(%s) Entry to view, [N]ext, [C]lear logs, or e[X]it: ", USERNAME);
	crmode();
	echo();
	getnstr(buf, sizeof(buf));
	noecho();
	if(isdigit(buf[0])) {
	    option = atoi(buf);
	    cur = first;
	    for(i=1; i<option; i++) {
		if(cur != NULL) cur = cur->next;
	    }
	    if(cur != NULL && cur->strval != NULL) {
		viewLogEntry(cur->strval);
	    }
	} else if(buf[0] == 'c' || buf[0] == 'C') {
	    spasm_delete_log(USERNAME, NULL, 1);
        } else if(buf[0] == 'n' || buf[0] == 'N') {
	    start = i;
            option = 1;
	} else if(buf[0] == 'x' || buf[0] == 'X') {
            /* nada -- exit loop and menu */
	} else {
            option = 1;
            beep();
        }
    } while(option);

    clean_log_list(first);
}

static void
qMenu(void)
{
    int option, i, valid_ans, current=0;
    char buf[128], *cp, fbuf[1024];
    DIR *dir;
    struct stat statbuf;
    struct dirent *dent;
    FILE *fp;
    struct spasm_list *first = NULL, *cur = NULL;

    if(chdir(spasm_config->quarantinedir) != 0) {
	spasm_error(LOG_ERR, "Error opening quarantine directory: %m");
	return;
    }
    if (stat(USERNAME, &statbuf) < 0) {
	/* assume no quarantine data for this user */
	spasm_error(LOG_ERR, "No quarantined messages for %s", USERNAME);
	return;
    }
    if(chdir(USERNAME) != 0) {
	spasm_error(LOG_ERR, "Error changing directories: %m");
	return;
    }

    do {
	/* free existing message file list (if any) before rebuilding it */
	clean_log_list(first);
	cur = first = NULL;

	/* now, create the message file list */
        if((dir = opendir(".")) == NULL) {
	    spasm_error(LOG_ERR, "Error reading quarantine list: %m");
	    return;
        }
        while((dent = readdir(dir)) != NULL) {
	    if(dent->d_name[0] == '.') continue;
            if((cur = add_log_list(cur, dent->d_name)) == NULL) {
                clean_log_list(first);
                return;
	    }
	    if(first == NULL) first = cur;
        }
        closedir(dir);

	option = 0;

	doScreen("Quarantine", first, current);
	move(ymax-2, 1);
	clrtoeol();
	mvvline(ymax - 2, xmax - 1, 0, 1);  /* repair border */
	mvprintw(ymax-2, 1, "(%s) Select message, [N]ext, or e[X]it: ", USERNAME);
	crmode();
	echo();
	getnstr(buf, sizeof(buf));
	noecho();
	option = atoi(buf);
	if(option > 0) {
	    int opt2 = 0;
	    clear();
	    box(mw, ACS_VLINE, ACS_HLINE);
	    mvprintw(1, ((xmax/2)-7), "Message Excerpt");
	    cur = first;
	    for(i=1; i<option; i++) {
		if(cur != NULL) cur = cur->next;
	    }
	    if(cur == NULL) {
		spasm_error(LOG_ERR, "Invalid message number");
		continue;
	    }
	    if((fp = fopen(cur->entry, "r")) == NULL) {
		spasm_error(LOG_ERR, "Error reading message: %m");
		continue;
	    }
	    while(fgets(fbuf, sizeof(fbuf), fp) != NULL &&
			strcmp(fbuf, "\r\n") != 0) {
		/* skip headers */
	    }
	    for(i=0;i<ymax-6;i++) {
		if(fgets(fbuf, sizeof(fbuf), fp) != NULL) {
		    char *ptr;
		    if((ptr = strchr(fbuf, '\r')) != NULL) *ptr = '\0';
		    if((ptr = strchr(fbuf, '\n')) != NULL) *ptr = '\0';
		    mvprintw(3+i, 2, "%*s", xmax-10, fbuf);
		}
	    }
	    do {
		valid_ans = 1;
		move(ymax-2, 1);
		clrtoeol();
		mvvline(ymax - 2, xmax - 1, 0, 1);  /* repair border */
		mvprintw(ymax - 2, 1, "(%s) [S]end, [D]elete, [W]hitelist, e[X]it: ", USERNAME);
		opt2 = getch();
		if(opt2 == 's' || opt2 == 'S') {
		    int pfd[2];
		    pid_t pid;
		    move(ymax-2, 1);
		    clrtoeol();
		    mvprintw(ymax-2, 1, "Sending message...");
		    if(pipe(pfd) < 0) {
			spasm_error(LOG_ERR, "Error creating mail pipe: %m");
			continue;
		    }
		    if((pid = fork()) < 0) {
			spasm_error(LOG_ERR, "Error forking mail process: %m");
			close(pfd[0]);
			close(pfd[1]);
			continue;
		    }
		    if(pid == 0) {	/* Child process */
			dup2(pfd[0], 0);
			close(pfd[1]);
			execl(spasm_config->sendmail, "sendmail", USERNAME, 0);
			/* Shouldn't get here */
			spasm_error(LOG_ERR,
			            "Error executing sendmail (%d): %m",
				    errno);
			exit(128);
		    }
		    close(pfd[0]);
		    fseek(fp, 0, SEEK_SET);
		    while(fgets(fbuf, sizeof(fbuf), fp) != NULL) {
			write(pfd[1], fbuf, strlen(fbuf));
		    }
		    close(pfd[1]);
		    waitpid(pid, NULL, 0);
		} else if(opt2 == 'd' || opt2 == 'D') {
		    if((unlink(cur->entry)) != 0)
			spasm_error(LOG_ERR, "Error unlinking quarantine file: %m");
		} else if(opt2 == 'w' || opt2 == 'W') {
		    FILE *qfile;
		    char *buf;
		    char *addr;
		    int foundit = 0;
		    if((buf = calloc(1024, 1)) == NULL) {
			spasm_error(LOG_ERR, "Error allocating memory: %m");
			continue;
		    }
		    if((addr = calloc(1024, 1)) == NULL) {
			spasm_error(LOG_ERR, "Error allocating memory: %m");
			free(buf);
			continue;
		    }
		    if((qfile = fopen(cur->entry, "r")) == NULL) {
			spasm_error(LOG_ERR, "Error reading message: %m");
			free(buf);
			free(addr);
			continue;
		    }
		    while(fgets(buf, 1024, qfile) != NULL) {
		        if(sscanf(buf, "X-spasm-sender: %1023s\r\n", addr)) {
			    spasm_insert_wl(USERNAME, addr, "1");
			    foundit = 1;
			    break;
			}
		    }
		    if(!foundit) {
		        spasm_error(LOG_ERR, "Error reading X-spasm-sender", buf);
		    }
		    free(buf);
		    free(addr);
		    fclose(qfile);
		} else if(! (opt2 == 'x' || opt2 == 'X')) {
		    beep();
		    valid_ans = 0;
		}
	    } while (! valid_ans);
	    fclose(fp);
	} else {
	    /* option <= 0 or maybe 'X' to xit */

	    cp = buf;
	    while (isspace(*cp))
		cp++;
            if (buf[0] == 'n' || buf[0] == 'N') {
                current = doScreen("Quarantine", first, current);
                option = 1;
            } else if (! (*cp == 'x' || *cp == 'X')) {
		beep();
		option = 1;
	    }
	}
    } while(option);

    /* cleanup */
    clean_log_list(first);
}

static void
reMenu(char *dbname)
{
    int option = 0;
    struct spasm_list *entries = NULL;
    struct spasm_list *cur = NULL;
    char buf[128];

    clear();
    box(mw, ACS_VLINE, ACS_HLINE);
    getmaxyx(stdscr, ymax, xmax);
    mvprintw(1, ((xmax/2)-7), "RegEx Menu");

    do {
	int i = 0;
	clear();
	box(mw, ACS_VLINE, ACS_HLINE);
	getmaxyx(stdscr, ymax, xmax);
	mvprintw(1, ((xmax/2)-7), "RegEx Menu");

	entries = spasm_get_re(dbname);

	cur = entries;
	while(cur != NULL) {
	    if(i % 2 == 0) mvprintw(3+i/2, 2, "%s", cur->entry);
	    else mvprintw(3+i/2, 40, "%s", cur->entry);
	    cur = cur->next;
	    i++;
	}
	cur = entries;
	while(cur != NULL) {
	    entries = cur->next;
	    free(cur);
	    cur = entries;
	}

	mvprintw(ymax-6, ((xmax/2)-11), "[A]dd RegEx Entry");
	mvprintw(ymax-5, ((xmax/2)-11), "[D]elete RegEx Entry");
	mvprintw(ymax-4, ((xmax/2)-11), "e[X]it to Main Menu");
	mvprintw(ymax-2, 1, "(%s) Ok: ", dbname);
	option = getch();

	if(option == 'a' || option == 'A') {
	    memset(buf, '\0', sizeof(buf));
	    move(ymax-6, 1);
	    clrtoeol();
	    move(ymax-5, 1);
	    clrtoeol();
	    move(ymax-4, 1);
	    clrtoeol();
	    move(ymax-3, 1);
	    clrtoeol();
	    move(ymax-2, 1);
	    clrtoeol();
	    mvprintw(ymax-2, 1, "Entry to add: ");
	    echo();
	    getnstr(buf, sizeof(buf));
	    noecho();
	    if(strlen(buf)) spasm_insert_re(dbname, buf);
	} else if(option == 'd' || option == 'D') {
	    memset(buf, '\0', sizeof(buf));
	    move(ymax-6, 1);
	    clrtoeol();
	    move(ymax-5, 1);
	    clrtoeol();
	    move(ymax-4, 1);
	    clrtoeol();
	    move(ymax-3, 1);
	    clrtoeol();
	    move(ymax-2, 1);
	    clrtoeol();
	    mvprintw(ymax-2, 1, "Entry to delete: ");
	    echo();
	    getnstr(buf, sizeof(buf));
	    noecho();
	    if(strlen(buf)) spasm_delete_re(dbname, buf);
	}
    } while(option != 'x' && option != 'X');
}

static void
wlMenu()
{
    int option = 0, current = 0;
    struct spasm_list *first = NULL;
    struct spasm_list *cur = NULL;
    char buf[128];

    do {
	first = spasm_get_wl(USERNAME);
	cur = first;
	doScreen("Whitelist Menu", first, current);
	mvprintw(ymax-2, 1, "(%s) [A]dd, [D]elete, [N]ext, e[X]it: ", USERNAME);
	option = getch();

	if(option == 'a' || option == 'A') {
	    memset(buf, '\0', sizeof(buf));
	    move(ymax-6, 1);
	    clrtoeol();
	    move(ymax-5, 1);
	    clrtoeol();
	    move(ymax-4, 1);
	    clrtoeol();
	    move(ymax-3, 1);
	    clrtoeol();
	    move(ymax-2, 1);
	    clrtoeol();
	    mvprintw(ymax-2, 1, "(%s) Entry to add: ", USERNAME);
	    echo();
	    getnstr(buf, sizeof(buf));
	    noecho();
	    if(strlen(buf)) spasm_insert_wl(USERNAME, buf, "1");
	} else if(option == 'd' || option == 'D') {
            memset(buf, '\0', sizeof(buf));
	    move(ymax-6, 1);
	    clrtoeol();
            move(ymax-5, 1);
            clrtoeol();
            move(ymax-4, 1);
            clrtoeol();
            move(ymax-3, 1);
            clrtoeol();
            move(ymax-2, 1);
            clrtoeol();
            mvprintw(ymax-2, 1, "(%s) Entry to delete: ", USERNAME);
            echo();
            getnstr(buf, sizeof(buf));
            noecho();
	    if(strlen(buf)) spasm_delete_wl(USERNAME, buf);
	} else if(option == 'n' || option == 'N') {
	    current = doScreen("Whitelist Menu", first, current);
	} else if(option != 'x' && option != 'X') {
	    beep();
	}
	clean_log_list(first);
    } while(option != 'x' && option != 'X');
}

static void
blMenu()
{
    int option = 0, current = 0;
    struct spasm_list *first = NULL;
    struct spasm_list *cur = NULL;
    char buf[128];

    do {
	first = spasm_get_bl(USERNAME);
	cur = first;
	doScreen("Blacklist Menu", first, current);
        mvprintw(ymax-2, 1, "(%s) [A]dd, [D]elete, [N]ext, e[X]it: ", USERNAME);
        option = getch();

        if(option == 'a' || option == 'A') {
            memset(buf, '\0', sizeof(buf));
            move(ymax-6, 1);
            clrtoeol();
            move(ymax-5, 1);
            clrtoeol();
            move(ymax-4, 1);
            clrtoeol();
            move(ymax-3, 1);
            clrtoeol();
            move(ymax-2, 1);
            clrtoeol();
            mvprintw(ymax-2, 1, "(%s) Entry to add: ", USERNAME);
            echo();
            getnstr(buf, sizeof(buf));
            noecho();
            if(strlen(buf)) spasm_insert_bl(USERNAME, buf, "1");
        } else if(option == 'd' || option == 'D') {
            memset(buf, '\0', sizeof(buf));
            move(ymax-6, 1);
            clrtoeol();
            move(ymax-5, 1);
            clrtoeol();
            move(ymax-4, 1);
            clrtoeol();
            move(ymax-3, 1);
            clrtoeol();
            move(ymax-2, 1);
            clrtoeol();
            mvprintw(ymax-2, 1, "(%s) Entry to delete: ", USERNAME);
            echo();
            getnstr(buf, sizeof(buf));
            noecho();
            if(strlen(buf)) spasm_delete_bl(USERNAME, buf);
        } else if(option == 'n' || option == 'N') {
	    current = doScreen("Blacklist Menu", first, current);
	} else if(option != 'x' && option != 'X') {
	    beep();
	}
	clean_log_list(first);
    } while(option != 'x' && option != 'X');
}

static void
gipMenu()
{
    int option = 0;
    struct spasm_list *entries = NULL;
    struct spasm_list *cur = NULL;
    char buf[128];

    clear();
    box(mw, ACS_VLINE, ACS_HLINE);
    getmaxyx(stdscr, ymax, xmax);
    mvprintw(1, ((xmax/2)-7), "Global IP Whitelist");

    do {
        int i = 0;
        clear();
        box(mw, ACS_VLINE, ACS_HLINE);
        getmaxyx(stdscr, ymax, xmax);
        mvprintw(1, ((xmax/2)-7), "Global IP Whitelist");

        entries = spasm_get_wl(spasm_config->globalipwl);

        cur = entries;
        while(cur != NULL) {
            if(i % 2 == 0) mvprintw(3+i/2, 2, "%s", cur->entry);
            else mvprintw(3+i/2, 40, "%s", cur->entry);
            cur = cur->next;
            i++;
        }
        cur = entries;
        while(cur != NULL) {
            entries = cur->next;
            free(cur);
            cur = entries;
        }
        mvprintw(ymax-6, ((xmax/2)-11), "[A]dd IP");
        mvprintw(ymax-5, ((xmax/2)-11), "[D]elete IP");
        mvprintw(ymax-4, ((xmax/2)-11), "e[X]it to Main Menu");
        mvprintw(ymax-2, 1, "(%s) Ok: ", USERNAME);
        option = getch();

        if(option == 'a' || option == 'A') {
            memset(buf, '\0', sizeof(buf));
            move(ymax-6, 1);
            clrtoeol();
            move(ymax-5, 1);
            clrtoeol();
            move(ymax-4, 1);
            clrtoeol();
            move(ymax-3, 1);
            clrtoeol();
            move(ymax-2, 1);
            clrtoeol();
            mvprintw(ymax-2, 1, "(%s) Entry to add: ", USERNAME);
            echo();
            getnstr(buf, sizeof(buf));
            noecho();
            if(strlen(buf)) spasm_insert_wl(spasm_config->globalipwl, buf, "1");
        }
        else if(option == 'd' || option == 'D') {
            memset(buf, '\0', sizeof(buf));
            move(ymax-6, 1);
            clrtoeol();
            move(ymax-5, 1);
            clrtoeol();
            move(ymax-4, 1);
            clrtoeol();
            move(ymax-3, 1);
            clrtoeol();
            move(ymax-2, 1);
            clrtoeol();
            mvprintw(ymax-2, 1, "(%s) Entry to delete: ", USERNAME);
            echo();
            getnstr(buf, sizeof(buf));
            noecho();
            if(strlen(buf)) spasm_delete_wl(spasm_config->globalipwl, buf);
        }
    } while(option != 'x' && option != 'X');
}

static void
filterMenu()
{
    char               buf[128], *answer;
    int                done, halfwidth, i, option, x, val, start = 0, line;
    const SPASM_FILTER *pfltr;

    /* following array indexed by:
     *
     *   SPASM_OFF, SPASM_TAG, SPASM_QUARANTINE, [Unused], SPASM_REJECT
     */
    const static char dbvaluemap[] = { '-', 'T', 'Q', '-', 'R' };

    halfwidth = COLS / 2;

    done   = 0;

    do {
	clear();
	box(mw, ACS_VLINE, ACS_HLINE);
	getmaxyx(stdscr, ymax, xmax);

        /* let's assume some anti-spam filters are available :-) ....*/
	mvprintw(1, ((xmax/2)-6), "Filter Menu");
	for (pfltr = spasm_filters, i = 0; pfltr->db_name != NULL && i < start; i++, pfltr++);
	if(pfltr->db_name == NULL) {
	    pfltr = spasm_filters;
	    i = 0;
	}
        for (line = 3; pfltr->db_name != NULL && line<ymax-3; i++, pfltr++) {
		char log = '-';
                val = spasm_lookup_filter(USERNAME, pfltr->db_name);
		if(val & SPASM_LOG) {
		    log = 'L';
		    val ^= SPASM_LOG;
		}
                x   = (i % 2 == 0) ? 2 : halfwidth;
		if(strcmp("SF_SPAMASSASSIN", pfltr->db_name) == 0) {
		    /* Special case for SpamAssassin. Show REQUIRED_HITS */
		    int hits;
		    hits = spasm_lookup_filter(USERNAME, "SA_REQUIRED_HITS");
		    mvprintw(line, x, "%2d. [%c] [%c] %s (Hits: %d)", i + 1, log, dbvaluemap[val], pfltr->ui_name, hits);
		} else {
		    mvprintw(line, x, "%2d. [%c] [%c] %s", i + 1, log, dbvaluemap[val], pfltr->ui_name);
		}
		if(i % 2) line++;
        }
        move(ymax-2, 1);
        clrtoeol();
        mvvline(ymax - 2, xmax - 1, 0, 1);  /* repair border */
        mvprintw(ymax-2, 1, "(%s) Select filter, [N]ext, e[X]it, ? for help: ", USERNAME);
        crmode();
        echo();
        getnstr(buf, sizeof(buf));
        answer = skipwhite(buf);
        lcase(answer);
        noecho();
        switch (*answer)
        {
            case '?':
                show_simple_filter_menu_help(general_filter_help, 0);
                break;
            case 'x':
                done = 1;  /* goodbye */
                break;
            case 'h':        /* help for an individual filter */
                answer++;
                answer = skipwhite(answer);
                if (! isdigit(*answer))
                    beep();
                else {
                    option = atoi(answer);

                    /* displayed filter numbers are 1-based */
                    if (option == 0 || option > NUM_FILTERS)
                        beep();
                    else
                        show_complex_filter_menu_help(option - 1, ITEM_FILTER);
                }
                break;
            case '0': case '1':
            case '2': case '3':
            case '4': case '5':
            case '6': case '7':
            case '8': case '9':
                option = atoi(answer);

                /* displayed filter numbers are 1-based */
                if (option == 0 || option > NUM_FILTERS)
                    beep();
                else
                    set_filter_menu_value(option - 1, xmax, ymax);
                break;
	    case 'n':
		start = i;
		break;
            default:
                beep();
                break;
        }
    } while(! done);
}

static void
prefsMenu()
{
    char               buf[128], *answer;
    int                done, option, val, line;
    const SPASM_FILTER_OPTION *popt;

    done   = 0;

    do {
        clear();
        box(mw, ACS_VLINE, ACS_HLINE);
        getmaxyx(stdscr, ymax, xmax);

        /* let's assume some anti-spam filters are available :-) ....*/
        mvprintw(1, ((xmax/2)-8), "Preferences Menu");
        for (line = 3, popt = spasm_filter_opts; popt->db_name != NULL; line++, popt++) {
	    val = 0;
	    if(spasm_lookup_prefs(USERNAME, popt->db_name, buf)) val = atoi(buf);
            mvprintw(line, 2, "%2d. [%c] %s", line - 2, val?'+':'-', popt->ui_name);
        }
        move(ymax-2, 1);
        clrtoeol();
        mvvline(ymax - 2, xmax - 1, 0, 1);  /* repair border */
        mvprintw(ymax-2, 1, "(%s) Select option, e[X]it, ? for help: ", USERNAME);
        crmode();
        echo();
        getnstr(buf, sizeof(buf));
        answer = skipwhite(buf);
        lcase(answer);
        noecho();
        switch (*answer)
        {
            case '?':
                show_simple_filter_menu_help(general_prefs_help, 0);
                break;
            case 'x':
                done = 1;  /* goodbye */
                break;
            case 'h':        /* help for an individual filter */
                answer++;
                answer = skipwhite(answer);
                if (! isdigit(*answer))
                    beep();
                else {
                    option = atoi(answer);

                    /* displayed filter numbers are 1-based */
                    if (option == 0 || option > NUM_OPTS)
                        beep();
                    else
                        show_complex_filter_menu_help(option - 1, ITEM_OPT);
                }
                break;
            case '0': case '1':
            case '2': case '3':
            case '4': case '5':
            case '6': case '7':
            case '8': case '9':
                option = atoi(answer);

                /* displayed filter numbers are 1-based */
                if (option == 0 || option > NUM_OPTS)
                    beep();
                else
                    set_prefs_menu_value(option - 1, xmax, ymax);
                break;
            default:
                beep();
                break;
        }
    } while(! done);
}

#define CMDLINE_HELP 128   /* largest ASCII value + 1 */
#define CMDLINE_VRSN 129

/* some getopt cmdline option stuff */
static struct option long_opts[] =
{
    {"help",          0, 0, CMDLINE_HELP},
    {"list-filters",  0, 0, 'f'},
    {"list-settings", 0, 0, 's'},
    {"version",       0, 0, CMDLINE_VRSN},
    {NULL, 0, 0, 0}
};

static void
bad_cmdline_syntax(void)
{
    fprintf(stderr, "Try `%s --help' for more information\n", progname);
    exit(1);
}

static void
version(void)
{
    fprintf(stderr, "%s %s\n", progname, SPASM_VERSION);
    exit(0);
}

static void
cmdline_help(void)
{
    fprintf(stderr,
"usage: %s [-fs] [--list-settings] [--list-filters] [--help] [--version]\n\n",
           progname);

    fprintf(stderr,
"where:\n\n"
"-f, --list-filters        list all supported anti-spam filters and exit\n"
"-s, --list-settings       list the values of all options, parameters, and\n"
"                          enabled filters and exit\n"
"\n"
"--help                    display this message and exit\n"
"--version                 display program version number and exit\n");

    exit(0);
}

static void
list_settings(void)
{
    int                          i, hit, gap, spacer, val;
    const SPASM_FILTER           *pfltr;
    const SPASM_FILTER_OPTION    *popt;
    const SPASM_FILTER_PARAMETER *pparam;

    const static char *states[] = {"Disabled", "Tag", "Quarantine", "Reject"};

    /* ----------- spasm filter parameters ------------- */
    spacer = 19;
    fprintf(stdout, "parameter%*csetting\n", spacer, ' ');
    fprintf(stdout, "---------%*c-------\n", spacer, ' ');
    for (i = 0, pparam = spasm_filter_params; pparam->ui_name; pparam++, i++) {
	val = spasm_lookup_filter(USERNAME, pparam->db_name);
	gap = ((int) sizeof("parameter")) - 1 + spacer - ((int) strlen(pparam->ui_name));

	/* spasm only has one type of parameter so far (UINT). */
	fprintf(stdout, "%s%*c%7u\n", pparam->ui_name, (gap > 0) ? gap : 1, ' ', val);
    }
    if (i == 0) fputs("<none>\n", stdout);
    fputc('\n', stdout);  /* upspace */

    /* ----------- spasm filter options ------------- */
    spacer = 22;
    fputs("option\n------\n", stdout);
    for (i = 0, popt = spasm_filter_opts; popt->db_name; popt++, i++) {
	const char *optstate;

	val = spasm_lookup_filter(USERNAME, popt->db_name);
	gap = ((int) sizeof("option")) - 1 + spacer - ((int) strlen(popt->ui_name));
	if (val == SPASM_OFF) optstate = states[val];
	else optstate = "Enabled";
	fprintf(stdout, "%s%*c%s\n", popt->ui_name, (gap > 0) ? gap : 1, ' ', optstate);
    }
    if (i == 0) fputs("<none>\n", stdout);
    fputc('\n', stdout);  /* upspace */

    /* ----------- spasm filters ------------- */
    spacer = 22;
    fputs("filter\n------\n", stdout);
    for (hit = 0, pfltr = spasm_filters; pfltr->db_name; pfltr++) {
	val = spasm_lookup_filter(USERNAME, pfltr->db_name);
	if (val != SPASM_OFF) {
	    hit = 1;
	    gap = ((int) sizeof("filter")) - 1 + spacer - ((int) strlen(pfltr->ui_name));
	    fprintf(stdout, "%s%*c%s\n", pfltr->ui_name, (gap > 0) ? gap : 1, ' ', states[val]);
	}
    }
    if (! hit) fputs("<none enabled>\n", stdout);
    exit(0);
}

static void
list_all_filters(void)
{
    int                i, spacer = 12, gap;
    const SPASM_FILTER *plst;

    fprintf(stdout, "database name%*cuser interface name\n", spacer, ' ');
    fprintf(stdout, "-------------%*c-------------------\n", spacer, ' ');
    for (i = 0, plst = spasm_filters; plst->db_name; plst++, i++) {
	gap = ((int) sizeof("database name")) - 1 + spacer - ((int) strlen(plst->db_name));
	fprintf(stdout, "%s%*c%s\n", plst->db_name, (gap > 0) ? gap : 1, ' ', plst->ui_name);
    }
    if (i == 0) fputs("<none>\n", stdout);
    exit(0);
}

int
main(int argc, char **argv)
{
    int option, wants_settings_listing = 0;
    uid_t uid;
    struct passwd *pw;
    const char *args = "fs";
    char *slash = strrchr(argv[0], '/');

    progname = (slash) ? slash + 1 : argv[0];

    /* command line processing first */
    for (;;)
    {
	int rc;

	rc = getopt_long(argc, argv, args, long_opts, NULL);
	if (rc == -1)
	    break;

	switch (rc)
	{
	case 'f':
	    list_all_filters();
	    /* NOT REACHED */
	    break;
	case 's':
	    wants_settings_listing = 1;
	    break;
	case CMDLINE_HELP:
	    cmdline_help();
	    /* NOT REACHED */
	    break;
	case CMDLINE_VRSN:
	    version();
	    /* NOT REACHED */
	    break;
	case '?':
	    bad_cmdline_syntax();
	    /* NOT REACHED */
	    break;
	}
	if (wants_settings_listing) {
	    /* can't do this without spasm configuration init */

	    break;
	}
    }
    if (optind < argc) {
	/* trailing garbage on the command line */

	fputs("too many command line arguments\n", stderr);
	bad_cmdline_syntax();
	/* NOT REACHED */
    }

    signal(SIGINT, cleanup);
    signal(SIGTERM, cleanup);

    uid = getuid();
    pw = getpwuid(uid);
    if(pw == NULL) {
	printf("Error looking up uid %d: %m\n", uid);
	cleanup(0);
    }
    if((spasm_config = spasm_cfg()) == NULL) {                                                  
        exit(1);                                                                                
    }
    if(strcmp(pw->pw_name, spasm_config->spasmuser) == 0) uid = 0;
    if((USERNAME = calloc(strlen(pw->pw_name)+strlen(spasm_config->defaultdomain)+2, 1)) == NULL) {
	printf("Error allocating memory for username: %m\n");
	cleanup(0);
    }
    sprintf(USERNAME, "%s@%s", pw->pw_name, spasm_config->defaultdomain);

    /*
     * Honor a cmdline switch that required initialization that's now
     * complete.
     */
    if (wants_settings_listing) {
	list_settings();
	/* NOT REACHED */
    }

    mw = initscr();
    keypad(stdscr, TRUE);
    nonl();
    cbreak();
    noecho();

    do {
	clear();
	box(mw, ACS_VLINE, ACS_HLINE);
	getmaxyx(stdscr, ymax, xmax);
	mvprintw(1, ((xmax/2)-4), "Main Menu");
	mvprintw(3, ((xmax/2)-7), "[F]ilter Menu");
	mvprintw(4, ((xmax/2)-7), "[B]lacklist Menu");
	mvprintw(5, ((xmax/2)-7), "[W]hitelist Menu");
	mvprintw(6, ((xmax/2)-7), "[R]ejection Logs");
	mvprintw(7, ((xmax/2)-7), "[V]iew Quarantine");
	mvprintw(8, ((xmax/2)-7), "[P]references Menu");
	if(uid == 0) mvprintw(9, ((xmax/2)-7), "[S]pamtrap Menu");
	if(uid == 0) mvprintw(10, ((xmax/2)-7), "[G]lobal IP Whitelist");
	if(uid == 0) mvprintw(11, ((xmax/2)-7), "[C]hange Username");
	mvprintw(13, ((xmax/2)-7), "e[X]it spasm");
	mvprintw(ymax-2, 1, "(%s) Ok: ", USERNAME);
	option = getch();

	if(option == 'f' || option == 'F') {
	    filterMenu();
	}
	else if(option == 'w' || option == 'W') {
	    wlMenu();
	}
	else if(option == 'b' || option == 'B') {
	    blMenu();
	}
	else if(option == 'r' || option == 'R') {
	    rlMenu();
	}
	else if(option == 'v' || option == 'V') {
	    qMenu();
	}
	else if(option == 'p' || option == 'P') {
	    prefsMenu();
	}
	else if(uid == 0 && (option == 'c' || option == 'C')) {
	    changeUsername();
	}
	else if(uid == 0 && (option == 's' || option == 'S')) {
	    reMenu(spasm_config->spamtrapdb);
	}
	else if(uid == 0 && (option == 'g' || option == 'G')) {
	    gipMenu();
	}
    } while(option != 'x' && option != 'X');

    cleanup(0);
    exit(0); /* Never gets here */
}
