/*
**    spasm - The spasm anti-spam milter
**    Copyright (C) 2002-2005 Sean Hanson <aradia@theasylum.org>
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pthread.h>
#include <pwd.h>
#include <resolv.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <sys/signal.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sysexits.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <libmilter/mfapi.h>

#include "spasm.h"
#include "getopt-portable.h"
#include "regex-portable.h"

#define SPASM_DAEMON
#include "filters.h"

#ifdef WITH_CLAMAV
  #include <clamav.h>
#endif /* WITH_CLAMAV */

_syscall0(pid_t,gettid);

/* Filter status structure */
struct filterStatus {
    char filter[FILTERNAME_MAX+1];	/* Filter name */
    int status;				/* Status of message, with SPASM_LOG option */
    int setting;			/* User's filter setting, with SPASM_LOG option */
};

/* Filter cache structure */
struct filterCache {
    const char *av;			/* Cached AV result */
};

/* Recipient structure */
struct spasmRecipient {
    char rcpt[SMTP_PATH_MAX+1];		/* Envelope recipient */
    char username[SMTP_LOCALPART_MAX+1];/* Canonical local part */
    char domain[SMTP_DOMAIN_MAX+1];	/* Canonical domain part */
    char quarantinefile[FILENAME_MAX+1];/* Quarantine file */
    char tmpqfile[FILENAME_MAX+1];	/* Temporary quarantine file */
    FILE *quarantinefp;			/* Quarantine file pointer */
    char cname[SMTP_PATH_MAX+1];	/* Canonical username for location of quarantine, logs, filters, prefs, wl, bl */
    int status;				/* Stores current message status, with SPASM_LOG option */
    int content_filtering;		/* Does user want any content filtering? */
    int logging;			/* Global logging preference */
    int subject_tagging;		/* Subject tagging preference */
    int use_domain_log;			/* Whether to use domain part only for log and quarantine */
    int delrcpt;			/* mlfi_header and mlfi_body rejection */
    int wl;				/* mlfi_header and mlfi_body whitelisting */
    struct filterStatus fstatus[NUM_FILTERS];	/* Stores filter settings and statuses */
    struct spasmRecipient *next;	/* Pointer to next recipient */
};

/* Private data structure */
struct mlfiPriv {
    char hostname[SMTP_DOMAIN_MAX+1];	/* Resolved hostname of connecting mail server - Connection Based */
    char *hostdomain;			/* ptr into hostname at SLD - Connection Based */
    char helohost[SMTP_DOMAIN_MAX+1];	/* HELO argument */
    char *helodomain;			/* ptr into helohost at SLD */
    char mailfrom[SMTP_PATH_MAX+1];	/* Envelope sender - Message Based */
    char *mailfromhost;			/* ptr into mailfrom at FQDN - Message Based */
    char *mailfromdomain;		/* ptr into mailfrom at SLD - Message Based */
    unsigned long int ipaddress;	/* IP address of connecting mail server - Connection Based */
    char ipstring[16];			/* Dotted quad IP address of connecting mail server - Connection Based */
    int unknownusers;			/* Current number of unknown users */
    char subject[1024];			/* Subject line - Message Based */
    struct filterCache fltrcache;	/* Filter cache - Message Based */
    struct spasmRecipient *rcpts;	/* List of recipients - Message Based */
    int num_rcpts;			/* Total number of recipients - Message Based */
};

#define MLFIPRIV ((struct mlfiPriv *) smfi_getpriv(ctx))

/* Various random thingies */
#define MAX_SA_SIZE 250000
/*#define DOMAIN_NAME_REGEX "([^.]*[.][^.]*)[.]*$"*/
/*#define CIDR_REGEX "^[[:digit:]]{1,3}[.][[:digit:]]{1,3}[.][[:digit:]]{1,3}[.][[:digit:]]{1,3}[/][[:digit:]]{1,2}$"*/
#define IP_REGEX "([[:digit:]]{1,3}[.][[:digit:]]{1,3}[.][[:digit:]]{1,3}[.][[:digit:]]{1,3})"
/*#define IPA_REGEX "^[[:digit:]]{1,3}[.][[:digit:]]{1,3}[.][[:digit:]]{1,3}[.][[:digit:]]{1,3}$"*/
#define EXE_REGEX "\n\r?\n\r?TVqQAAMAAAAEAAAATVqQAAMAAAAEAAAA"
#define SRS0_REGEX "^SRS0[=+-]{1}[^=]+=[^=]+=([^=]+)=([^@]+)@"
#define SA_RESPONSE_MAX 4096
#define RETERR(err) do { errno = (err); return (-1); } while (0)
/*static regex_t re_sld;*/
/*static regex_t re_cidr;*/
static regex_t re_ip;
/*static regex_t re_ipa;*/
static regex_t re_exe;
static regex_t re_srs0;
static char *progname;
static char local_hostname[SMTP_DOMAIN_MAX];

static const SPASM_FILTER *pwl_only_fltr;

static const char *DNS_ERRNO[] = { "SUCCESS", "NXDOMAIN", "TEMPFAIL", "PERMFAIL", "NODATA" };

static const char *DAYOFWEEK[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
static const char *MONTH[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

struct spasmConfigStruct *spasm_config;

time_t verifymx_lastexpire;
time_t greylist_lastexpire;

#ifdef WITH_CLAMAV
  struct cl_node *clamav_root = NULL;
  time_t clamav_lastload;
  int clamav_accept = 1;
  int clamav_num = 0;
  pthread_mutex_t mutexLock_clamav_accept = PTHREAD_MUTEX_INITIALIZER;
  pthread_cond_t cond_clamav_accept = PTHREAD_COND_INITIALIZER;
  pthread_mutex_t mutexLock_clamav_num = PTHREAD_MUTEX_INITIALIZER;
  pthread_cond_t cond_clamav_num = PTHREAD_COND_INITIALIZER;
#endif /* WITH_CLAMAV */

pthread_mutex_t mutexLock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t mutexLock_globalipbl = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t mutexLock_greylist = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t mutexLock_verifymx = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t mutexLock_stats = PTHREAD_MUTEX_INITIALIZER;

void
spasm_error(int syslevel, char *errstr, ...)
{
    va_list arglist;
    char spasm_errstr[1024];

    va_start(arglist, errstr);
    vsnprintf(spasm_errstr, sizeof(spasm_errstr), errstr, arglist);
    va_end(arglist);

    syslog(syslevel, "%s", spasm_errstr);
}

/* SIGQUIT */
static void
cleanup()
{
    int status;
    /*regfree(&re_sld);*/
    /*regfree(&re_cidr);*/
    regfree(&re_ip);
    /*regfree(&re_ipa);*/
    regfree(&re_exe);
    regfree(&re_srs0);
    syslog(LOG_INFO, "(cleanup) Shutting down (%d)", gettid());
    closelog();
    unlink(spasm_config->pidfile);
    while(wait3(&status,WNOHANG,(struct rusage*) 0) >= 0);
#ifdef WITH_CLAMAV
    cl_free(clamav_root);
#endif /* WITH_CLAMAV */
    free_spasm_cfg(spasm_config);
    exit(0);
}

/* SIGSEGV */
static void
sigSEGV()
{
    /*regfree(&re_sld);*/
    /*regfree(&re_cidr);*/
    regfree(&re_ip);
    /*regfree(&re_ipa);*/
    regfree(&re_exe);
    regfree(&re_srs0);
    syslog(LOG_ERR, "(sigSEGV) Caught SIGSEGV (%d)", gettid());
    closelog();
    abort();
    exit(0); /* Never gets here */
}

/* Reload config file */
static void
reloadConfig()
{
    struct spasmConfigStruct *newconf;
    struct spasmConfigStruct *oldconf;

    /* We load the config file into a tmp struct then swap pointers,
       to hopefully avoid any potential race conditions (ie, a thread
       running with a partially-filled spasm_config struct) */

    syslog(LOG_INFO, "(reloadConfig) Reloading configuration");

    /* First, we load the new configuration into a separate struct */
    if((newconf = spasm_cfg()) == NULL) return;

    /* Next, we save the ptr to the old configuration */
    oldconf = spasm_config;

    /* Now, point spasm_config to the new config */
    spasm_config = newconf;

    /* Finally, clean up the memory used by the old config */
    free_spasm_cfg(oldconf);
}

/* setsection(), ns_skiprr(), and ns_initparse(), ns_parserr() from BIND/glibc */
static void
setsection(ns_msg *msg, ns_sect sect) {
    msg->_sect = sect;
    if (sect == ns_s_max) {
        msg->_rrnum = -1;
        msg->_ptr = NULL;
    } else {
        msg->_rrnum = 0;
        msg->_ptr = msg->_sections[(int)sect];
    }
}

int
ns_skiprr(const u_char *ptr, const u_char *eom, ns_sect section, int count)
{
    const u_char *optr = ptr;
    for ((void)NULL; count > 0; count--) {
        int b, rdlength;

        b = dn_skipname(ptr, eom);
        if (b < 0) RETERR(EMSGSIZE);
        ptr += b/*Name*/ + NS_INT16SZ/*Type*/ + NS_INT16SZ/*Class*/;
        if (section != ns_s_qd) {
            if (ptr + NS_INT32SZ + NS_INT16SZ > eom) RETERR(EMSGSIZE);
            ptr += NS_INT32SZ/*TTL*/;
            NS_GET16(rdlength, ptr);
            ptr += rdlength/*RData*/;
        }
    }
    if (ptr > eom) RETERR(EMSGSIZE);
    return (ptr - optr);
}

int
ns_initparse(const u_char *msg, int msglen, ns_msg *handle)
{   
    const u_char *eom = msg + msglen;
    int i;
    
    memset(handle, 0x5e, sizeof *handle);
    handle->_msg = msg;
    handle->_eom = eom;
    if (msg + NS_INT16SZ > eom) RETERR(EMSGSIZE);
    NS_GET16(handle->_id, msg); 
    if (msg + NS_INT16SZ > eom) RETERR(EMSGSIZE);
    NS_GET16(handle->_flags, msg); 
    for (i = 0; i < ns_s_max; i++) {
        if (msg + NS_INT16SZ > eom) RETERR(EMSGSIZE);
        NS_GET16(handle->_counts[i], msg);
    }
    for (i = 0; i < ns_s_max; i++)
        if (handle->_counts[i] == 0) handle->_sections[i] = NULL;
        else {
            int b = ns_skiprr(msg, eom, (ns_sect)i, handle->_counts[i]);
            if (b < 0) {
                return (-1);
            }
            handle->_sections[i] = msg;
            msg += b;
        }
    if (msg != eom) RETERR(EMSGSIZE);
    setsection(handle, ns_s_max);
    return (0);
}

int
ns_parserr(ns_msg *handle, ns_sect section, int rrnum, ns_rr *rr)
{       
    int b;
    int tmp;
        
    /* Make section right. */
    if ((tmp = section) < 0 || section >= ns_s_max) RETERR(ENODEV);
    if (section != handle->_sect) setsection(handle, section);
    
    /* Make rrnum right. */
    if (rrnum == -1) rrnum = handle->_rrnum;
    if (rrnum < 0 || rrnum >= handle->_counts[(int)section]) RETERR(ENODEV);
    if (rrnum < handle->_rrnum) setsection(handle, section);
    if (rrnum > handle->_rrnum) {
        b = ns_skiprr(handle->_ptr, handle->_eom, section, rrnum - handle->_rrnum);
        if (b < 0) return (-1);
        handle->_ptr += b;
        handle->_rrnum = rrnum;
    }
                       
    /* Do the parse. */                                                            
    b = dn_expand(handle->_msg, handle->_eom, handle->_ptr, rr->name, NS_MAXDNAME);
    if (b < 0) return (-1);
    handle->_ptr += b; 
    if (handle->_ptr + NS_INT16SZ + NS_INT16SZ > handle->_eom) RETERR(EMSGSIZE);
    NS_GET16(rr->type, handle->_ptr);
    NS_GET16(rr->rr_class, handle->_ptr);
    if (section == ns_s_qd) {
        rr->ttl = 0; 
        rr->rdlength = 0;
        rr->rdata = NULL;
    } else {
        if (handle->_ptr + NS_INT32SZ + NS_INT16SZ > handle->_eom) RETERR(EMSGSIZE);
        NS_GET32(rr->ttl, handle->_ptr);     
        NS_GET16(rr->rdlength, handle->_ptr);                            
        if (handle->_ptr + rr->rdlength > handle->_eom) RETERR(EMSGSIZE);
        rr->rdata = handle->_ptr;                                                                                                                      
        handle->_ptr += rr->rdlength;
    }
    if (++handle->_rrnum > handle->_counts[(int)section]) setsection(handle, (ns_sect)((int)section + 1));

    /* All done. */
    return (0);
}
/* End BIND routines */

/* lowercase() - set a string to lowercase letters for case-insensitive matching */
static void
lowercase(char *buf)
{
    if (buf) {
        while (*buf) {
            if (isupper(*buf))
                *buf = tolower(*buf);
            buf++;
        }
    }
}

/*
 * create global pointer(s) to filter data structures that require O(1)
 * access time.
 */
static void
cache_filter_ptrs(void)
{
    const SPASM_FILTER *p;

    p = spasm_filters;
    while (p->db_name) {
	if (p->adhoc_id == SPASM_WL_ONLY) {
	    pwl_only_fltr = p;
	    break;
	}
	p++;
    }
}

/* getDateTime() - get timestamp in RFC2822 date-time format */
void
getDateTime(char *datetime, size_t size)
{
    time_t now;
    struct tm lt;
    char plusminus;
    int offset;

    now = time(NULL);
    if(localtime_r(&now, &lt) == NULL) return;
    offset = (timezone / 36) + (timezone % 3600); /* Divide by 36, not 3600, to get military-style hours; ie, 800 is 8 hours */
    if(offset < 1200) { plusminus = '-'; }
    else { plusminus = '+'; offset = 2400 - offset; }
    snprintf(datetime, size, "%s, %d %s %d %02d:%02d:%02d %c%04d",
      DAYOFWEEK[lt.tm_wday], lt.tm_mday, MONTH[lt.tm_mon], lt.tm_year+1900, lt.tm_hour, lt.tm_min, lt.tm_sec, plusminus, offset);
    return;
}

/* getSLD() - get the SLD.TLD of a FQDN */
char*
getSLD(char *hostname)
{
    char *ptr = hostname;
    char *ptr2;
    int finaldot = 0;
    char *rv = NULL;

    if(hostname[strlen(hostname)-1] == '.') {
        hostname[strlen(hostname)-1] = '\0';
        finaldot = 1;
    }

    if((ptr = strrchr(hostname, '.')) != NULL) {
        /* We've found our first dot (.tld) */
        *ptr = '\0';
        if((ptr2 = strrchr(hostname, '.')) != NULL) {
            /* We've found our second dot (.sld.tld) */
            ptr2++;
            rv = ptr2;
        } else {
            /* It's already SLD.TLD */
            rv = hostname;
        }
        *ptr = '.';
    }

    if(finaldot) hostname[strlen(hostname)] = '.';

    return rv;
}

/* isIP() - determine whether a string is an IP address */
int
isIP(char *entry)
{
    char *ptr;
    char *ptr2;

    if(strlen(entry) > 15) return 0;

    ptr2 = entry;
    ptr = ptr2;

    /* Check first byte */
    while(isdigit(*ptr)) ptr++;
    if(*ptr != '.') return 0;   /* We should be at the first dot */
    *ptr = '\0';
    if(atoi(ptr2) > 255) { *ptr = '.'; return 0; }
    *ptr = '.';
    ptr++; 
    ptr2 = ptr;

    /* Check second byte */
    while(isdigit(*ptr)) ptr++;
    if(*ptr != '.') return 0;   /* We should be at the second dot */
    *ptr = '\0';
    if(atoi(ptr2) > 255) { *ptr = '.'; return 0; }
    *ptr = '.';
    ptr++;
    ptr2 = ptr;

    /* Check third byte */
    while(isdigit(*ptr)) ptr++;
    if(*ptr != '.') return 0;   /* We should be at the third dot */
    *ptr = '\0';
    if(atoi(ptr2) > 255) { *ptr = '.'; return 0; }
    *ptr = '.';
    ptr++;
    ptr2 = ptr;

    /* Check fourth byte */
    while(isdigit(*ptr)) ptr++;
    if(*ptr != '\0') return 0;  /* We should be at the end of the string */
    if(atoi(ptr2) > 255) return 0;

    /* We have four numbers, each less than 255, separated by dots */
    return 1;

/*
    if(regexec(&re_ipa, entry, 0, NULL, 0) == 0) return 1;
    return 0;
*/
}

/* isCIDR() - determine whether a string is a CIDR entry */
int
isCIDR(char *entry)
{
    char *ptr;
    char *ptr2;

    if(strlen(entry) > 18) return 0;

    ptr2 = entry;
    ptr = ptr2;

    /* Check first byte */
    while(isdigit(*ptr)) ptr++;
    if(*ptr != '.') return 0;   /* We should be at the first dot */
    *ptr = '\0';
    if(atoi(ptr2) > 255) { *ptr = '.'; return 0; }
    *ptr = '.';
    ptr++;
    ptr2 = ptr;

    /* Check second byte */
    while(isdigit(*ptr)) ptr++;
    if(*ptr != '.') return 0;   /* We should be at the second dot */
    *ptr = '\0';
    if(atoi(ptr2) > 255) { *ptr = '.'; return 0; }
    *ptr = '.';
    ptr++;
    ptr2 = ptr;

    /* Check third byte */
    while(isdigit(*ptr)) ptr++;
    if(*ptr != '.') return 0;   /* We should be at the third dot */
    *ptr = '\0';
    if(atoi(ptr2) > 255) { *ptr = '.'; return 0; }
    *ptr = '.';
    ptr++;
    ptr2 = ptr;

    /* Check fourth byte */
    while(isdigit(*ptr)) ptr++;
    if(*ptr != '/') return 0;  /* We should be at the slash */
    *ptr = '\0';
    if(atoi(ptr2) > 255) { *ptr = '/'; return 0; }
    *ptr = '/';
    ptr++;
    ptr2 = ptr;

    /* Check prefix */
    while(isdigit(*ptr)) ptr++;
    if(*ptr != '\0') return 0;  /* We should be at the prefix */
    if(atoi(ptr2) > 32) return 0;

    /* We have four numbers, each <=255, separated by dots, with a prefix <=32 */
    return 1;

/*
    if(regexec(&re_cidr, entry, 0, NULL, 0) == 0) return 1;
    return 0;
*/
}

/* inCIDR() - check a host address against a CIDR block */
int
inCIDR(char *ipstring, char *cidrr)
{
    char cidr[20];
    char *chr;
    struct in_addr in;	/* cidr address */
    struct in_addr in2; /* host address */
    unsigned long hosts, ip1, ip2, ip3;
    
    if(ipstring == NULL || cidrr == NULL) return 1;

    memset(cidr, 0, sizeof(cidr));
    strncpy(cidr, cidrr, sizeof(cidr)-1);
    if((chr = strchr(cidr, '/')) == NULL) return 1;
    *chr = '\0';
    chr++;
    hosts = 1<<(32-(atoi(chr)));
    if(inet_aton(cidr, &in) == 0) return 1;
    else ip1 = ntohl(in.s_addr);
    if(inet_aton(ipstring, &in2) == 0) return 1;
    else ip2 = ntohl(in2.s_addr);
    ip3 = ip1 + hosts - 1;
    if(ip2 >= ip1 && ip2 <= ip3) return 1;
    return 0;
}

/* spasm_cidr_wlbl() - search whitelist/blacklist for CIDR matches */
int
spasm_cidr_wlbl(int listtype, char *dbname, char *ipaddress) {
    int rv = 0;

    if(dbname == NULL || ipaddress == NULL) return 0;

#ifdef BERKELEY_DB
    if((listtype == SPASM_WHITELIST && spasm_config->whitelistdb == BERKELEY_DB) ||
       (listtype == SPASM_BLACKLIST && spasm_config->blacklistdb == BERKELEY_DB)) {
        DB *db;
        DBT dbk, dbv;
        DBC *dbcp;
        char dbfile[FILENAME_MAX+1];
	char data[SPASM_DBT_MAX+1];
        int dbfd, rc;

	if(listtype == SPASM_WHITELIST) SP_STRCAT(dbfile, spasm_config->whitelists, dbname);
	else if(listtype == SPASM_BLACKLIST) SP_STRCAT(dbfile, spasm_config->blacklists, dbname);

        memset(&dbk, 0, sizeof(DBT));
        memset(&dbv, 0, sizeof(DBT));

        if((rc = db_create(&db, NULL, 0)) != 0) {
            spasm_error(LOG_ERR, "(spasm_cidr_wlbl) Error creating db: %s", db_strerror(rc));
            return 0;
        }
        if((rc = sp_db_open(db, NULL, dbfile, NULL, DB_HASH, DB_RDONLY, 0)) != 0) {
            if(rc != 2) spasm_error(LOG_ERR, "(spasm_cidr_wlbl) Error opening database (%d): %s", rc, db_strerror(rc));
            db->close(db, 0);
            return 0;
        }
        if((rc = db->fd(db, &dbfd)) != 0) {
            spasm_error(LOG_ERR, "(spasm_cidr_wlbl) Error reading file descriptor for database: %s", db_strerror(rc));
            db->close(db, 0);
            return 0;
        }
        if((rc = db->cursor(db, NULL, &dbcp, 0)) != 0) {
            spasm_error(LOG_ERR, "(spasm_cidr_wlbl) Error retrieving db cursor: %s", db_strerror(rc));
            db->close(db, 0);
            return 0;
        }
        while((dbcp->c_get(dbcp, &dbk, &dbv, DB_NEXT) == 0) && rv != 1) {
	    if(dbk.size > SPASM_DBT_MAX) continue;
	    memset(data, 0, sizeof(data));
	    strncpy(data, dbk.data, dbk.size);
            if(isCIDR(data) && inCIDR(ipaddress, data)) rv = 1;
        }                                                                                                                                            
        dbcp->c_close(dbcp);
        db->close(db, 0);
	return rv;
    }
#endif

#ifdef MYSQL_DB
    if((listtype == SPASM_WHITELIST && spasm_config->whitelistdb == MYSQL_DB) ||
       (listtype == SPASM_BLACKLIST && spasm_config->blacklistdb == MYSQL_DB)) {
        struct spasm_list *first, *cur;

	if(listtype == SPASM_WHITELIST) first = spasm_get_wl(dbname);
	else if(listtype == SPASM_BLACKLIST) first = spasm_get_bl(dbname);

        cur = first;
        while(cur != NULL) {
            if(!rv && isCIDR(cur->entry) && inCIDR(ipaddress, cur->entry)) rv = 1;
            first = cur->next;
            free(cur);
            cur = first;
        }
	return rv;
    }
#endif

    return rv;
}

/* search_regex() - Perform an individual regex */
int         
search_regex(char *pattern, char *key)
{       
    regex_t re; 
        
    if(pattern == NULL || key == NULL) return 0;
    if(regcomp(&re, pattern, REG_EXTENDED|REG_ICASE|REG_NOSUB) != 0) {
        syslog(LOG_ERR, "(search_regex) Error compiling regex \"%s\": %m (%d)", pattern, gettid());
        return 0;
    }
    if(regexec(&re, key, 0, NULL, 0) == 0) {
        regfree(&re);
        return 1;
    }
    regfree(&re);
    return 0;                                                                                                                                          
}                                                                                                                                                      

/* spasm_regex_wlbl() - search whitelist/blacklist for regex matches */
int         
spasm_regex_wlbl(int listtype, char *dbname, char *mailfrom, char *hostname)
{       
    int rv = 0;

    if(dbname == NULL || mailfrom == NULL || hostname == NULL) return 0;

#ifdef BERKELEY_DB
    if((listtype == SPASM_WHITELIST && spasm_config->whitelistdb == BERKELEY_DB) ||
       (listtype == SPASM_BLACKLIST && spasm_config->blacklistdb == BERKELEY_DB)) {
        DB *db;
        DBT dbk, dbv;
        DBC *dbcp;
        char dbfile[FILENAME_MAX+1];
        char data[SPASM_DBT_MAX+1];
        int dbfd, rc;
        char *entry;

	if(listtype == SPASM_WHITELIST) SP_STRCAT(dbfile, spasm_config->whitelists, dbname);
	else if(listtype == SPASM_BLACKLIST) SP_STRCAT(dbfile, spasm_config->blacklists, dbname);

        memset(&dbk, 0, sizeof(DBT));
        memset(&dbv, 0, sizeof(DBT));

        if((rc = db_create(&db, NULL, 0)) != 0) {
            spasm_error(LOG_ERR, "(spasm_regex_wlbl) Error creating db: %s", db_strerror(rc));
            return 0;
        }
        if((rc = sp_db_open(db, NULL, dbfile, NULL, DB_HASH, DB_RDONLY, 0)) != 0) {
            if(rc != 2) spasm_error(LOG_ERR, "(spasm_regex_wlbl) Error opening database (%d): %s", rc, db_strerror(rc));
            db->close(db, 0);
            return 0;
        }
        if((rc = db->fd(db, &dbfd)) != 0) {
            spasm_error(LOG_ERR, "(spasm_regex_wlbl) Error reading file descriptor for database: %s", db_strerror(rc));
            db->close(db, 0);
            return 0;
        }
        if((rc = db->cursor(db, NULL, &dbcp, 0)) != 0) {
            spasm_error(LOG_ERR, "(spasm_regex_wlbl) Error retrieving db cursor: %s", db_strerror(rc));
            db->close(db, 0);
            return 0;
        }
        while((dbcp->c_get(dbcp, &dbk, &dbv, DB_NEXT) == 0) && rv != 1) {
            if(dbk.size > SPASM_DBT_MAX) continue;
            memset(data, 0, sizeof(data));
            strncpy(data, dbk.data, dbk.size);
            if(!rv && data[0] == '/' && data[strlen(data)-1] == '/' && strlen(data) > 2) {
                entry = data+1;
                data[strlen(data)-1] = '\0';
                syslog(LOG_INFO, "(spasm_regex_wlbl) Searching /%s/ (%d)", entry, gettid());
                if(search_regex(entry, mailfrom) || search_regex(entry, hostname)) rv = 1;
            }
        
        }
        dbcp->c_close(dbcp);
        db->close(db, 0);
	return rv;
    }
#endif

#ifdef MYSQL_DB
    if((listtype == SPASM_WHITELIST && spasm_config->whitelistdb == MYSQL_DB) ||
       (listtype == SPASM_BLACKLIST && spasm_config->blacklistdb == MYSQL_DB)) {
        struct spasm_list *first, *cur;
        char *entry;

	if(listtype == SPASM_WHITELIST) first = spasm_get_wl(dbname);
	else if(listtype == SPASM_BLACKLIST) first = spasm_get_bl(dbname);

        cur = first;
        while(cur != NULL) {
            if(!rv && cur->entry[0] == '/' && cur->entry[strlen(cur->entry)-1] == '/' && strlen(cur->entry) > 2) {
                entry = cur->entry+1;
                cur->entry[strlen(cur->entry)-1] = '\0';
                syslog(LOG_INFO, "(spasm_regex_wlbl) Searching /%s/ (%d)", entry, gettid());
                if(search_regex(entry, mailfrom) || search_regex(entry, hostname)) rv = 1;
            }
            first = cur->next;
            free(cur);
            cur = first;
        }
	return rv;
    }
#endif

    return rv;
}           

/* add_header() - add header[s] to a quarantine file */
void
add_header(char *qfile, char *header)
{
    int fd;
    FILE *ifp, *ofp;
    char fname[FILENAME_MAX+1];
    int i;
    char fbuf[1024];

    if(header == NULL || qfile == NULL) return;
    snprintf(fname, sizeof(fname), "%s", qfile);
    for(i=1;i<7;i++) *(fname+strlen(qfile)-i) = 'X';
    if((ifp = fopen(qfile, "r")) < 0) {
	syslog(LOG_ERR, "(add_header) Error opening qfile: %m (%d)", gettid());
	return;
    }
    if((fd = mkstemp(fname)) < 0 ) {
	syslog(LOG_ERR, "(add_header) Error opening temp file: %m (%d)", gettid());
	fclose(ifp);
	return;
    }
    chmod(fname, 0644);
    if((ofp = fdopen(fd, "a")) == NULL) {
	syslog(LOG_ERR, "(add_header) Error opening temp file stream: %m (%d)", gettid());
	fclose(ifp);
	close(fd);
    }
    while(fgets(fbuf, sizeof(fbuf), ifp)) {
	if(strcmp(fbuf, "\r\n") == 0) break;
	fprintf(ofp, "%s", fbuf);
    }
    fprintf(ofp, "%s\r\n", header);
    while(fgets(fbuf, sizeof(fbuf), ifp)) {
	fprintf(ofp, "%s", fbuf);
    }
    fclose(ifp);
    fclose(ofp);
    if(rename(fname, qfile) != 0) syslog(LOG_ERR, "(add_header) Error replacing qfile: %m (%d)", gettid());
    return;
}

/* add_filter_status() - add a filter status entry */
void
add_filter_status(struct spasmRecipient *cur, char *filter, int status)
{
    int i, j, oldstatus;

    if(cur == NULL || filter == NULL || status == 0) return;

    for(i=0; i<NUM_FILTERS; i++)
	if(strcmp(cur->fstatus[i].filter, filter) == 0) break;

    cur->fstatus[i].status = status;

    /*
    ** cur->status is now doing double duty:
    ** It saves the current status (TAG, QUARANTINE, REJECT)
    ** as well as the current logging status as bitwise-OR.
    **
    ** The logic is implemented to cut down on the amount of
    ** logging. If a message is filtered with a REJECT status,
    ** then that filter's log setting defines whether a log
    ** entry gets written. If a message is filtered with a TAG
    ** or QUARANTINE setting, then a log entry will only be
    ** written if ALL of the filters which caught the message
    ** have logging set.
    */

    /* First, check for a current REJECT status. This is a safety
    ** measure, since a REJECTed message should never go through
    ** any further filter checks */
    if(cur->status & SPASM_REJECT) return;

    /* Next, if the new status is REJECT, copy it verbatim */
    if(status & SPASM_REJECT) {
	cur->status = status;
	return;
    }

    /* Finally, do the TAG/QUARANTINE logic */

    /* First, save the non-logging rejection statuses to i and j */
    i = (cur->status & SPASM_LOG) ? cur->status ^ SPASM_LOG : cur->status;
    j = (status & SPASM_LOG) ? status ^ SPASM_LOG : status; 

    /* Next, save the current status to oldstatus */
    oldstatus = cur->status;

    /* Finally, update the status by comparing the non-logging statuses */
    if(i < j) {
	/* We're updating the status. Clear logging bit if the old
	   status did not have logging set, but the new one does.
	   However, if the old status was SPASM_OFF, don't clear
	   the logging bit. */
	cur->status = status;
	if(oldstatus && !(oldstatus & SPASM_LOG))
	    if(status & SPASM_LOG) cur->status ^= SPASM_LOG;
    } else {
	/* We're not updating the status. However, if the new status
	   does not have logging set, clear the bit if the old one did */
	if(!(status & SPASM_LOG))
	    if(cur->status & SPASM_LOG) cur->status ^= SPASM_LOG;
    }
}

/* filter_lookup() - lookup single filter entry */
int
filter_lookup(char *username, char *domain, char *entry)
{
    int rv = 0;
    char email[SMTP_PATH_MAX+1];

    if(spasm_config->debug) syslog(LOG_DEBUG, "(filter_lookup) Starting (%s@%s -> %s)... (%d)", username, domain, entry, gettid());
    snprintf(email, sizeof(email), "%s@%s", username, domain);
    rv = spasm_lookup_filter(email, entry);
    if(spasm_config->debug) syslog(LOG_DEBUG, "(filter_lookup) Finished (rv=%d) (%d)", rv, gettid());
    return rv;
}

/* add_to_blacklist() - add an entry to a blacklist database */
void
add_to_blacklist(char *blfile, char *key)
{
    if(blfile == NULL || key == NULL) return;

    syslog(LOG_INFO, "(add_to_blacklist) Adding %s to %s (%d)", key, blfile, gettid());
    spasm_insert_bl(blfile, key, "1");
    return;
}

/* log_rejection() - Log rejected spam in per-user logfiles */
void
log_rejection(SMFICTX *ctx, struct spasmRecipient *cur)
{
    time_t timep;
    char timestr[35];
    char logentry[70+(2*SMTP_PATH_MAX)+(2*SMTP_DOMAIN_MAX)+REASON_MAX];
    char reason[REASON_MAX];
    struct timeval tv;
    int i;

    if(spasm_config->stats != NULL) {
	pthread_mutex_lock(&mutexLock_stats);
	spasm_update_stats("rejected");
	pthread_mutex_unlock(&mutexLock_stats);
    }

    if(spasm_config->syslogrejections) {
	syslog(LOG_INFO, "Rejecting message for %s from %s", cur->rcpt, MLFIPRIV->mailfrom);
    }

    if(cur->logging == SPASM_OFF || !(cur->status & SPASM_LOG)) return;

    if(cur->cname[0] == '\0') {
	syslog(LOG_ERR, "(log_rejection) Error: cur->cname is NULL (%d)", gettid());
	return;
    }

    gettimeofday(&tv, NULL);

    memset(reason, 0, sizeof(reason));
    /* Let's create the filter string in order of status */
    for(i=0; i<NUM_FILTERS; i++) {                     
        if(cur->fstatus[i].status & SPASM_REJECT) {
            strcat(reason, cur->fstatus[i].filter);                                                                                                   
            strcat(reason, "|");                                                                                                                      
        }
    }                                        
    for(i=0; i<NUM_FILTERS; i++) {
        if(cur->fstatus[i].status & SPASM_QUARANTINE) {
            strcat(reason, cur->fstatus[i].filter);
            strcat(reason, "|");
        }
    }
    for(i=0; i<NUM_FILTERS; i++) {
        if(cur->fstatus[i].status & SPASM_TAG) {
            strcat(reason, cur->fstatus[i].filter);
            strcat(reason, "|");
        }
    }

    timep = time(NULL);
    memset(timestr, '\0', sizeof(timestr));
    snprintf(timestr, sizeof(timestr), "%s", (char *)ctime((time_t*)&timep));
    /* Get rid of ctime()'s newline */
    timestr[strlen(timestr)-1] = '\0';
    snprintf(logentry, sizeof(logentry), "%010li%010li:%s:%s:%s:%s:%s:%s:%s:%d", tv.tv_sec, tv.tv_usec, timestr, cur->rcpt, MLFIPRIV->mailfrom, MLFIPRIV->hostname, MLFIPRIV->ipstring, MLFIPRIV->helohost, reason, (cur->status&SPASM_LOG)?cur->status^SPASM_LOG:cur->status);

    spasm_insert_log(cur->cname, logentry);

    return;
}

/* userwl_lookup() - Lookup a whitelist entry */
int
userwl_lookup(SMFICTX *ctx, char *username, char *domain)
{
    char ipstring2[16];
    char email[SMTP_PATH_MAX+1];
    char dombuf[SMTP_DOMAIN_MAX+1];
    char *buf;
    int rv = 0;

    if(spasm_config->debug) syslog(LOG_DEBUG, "(userwl_lookup) Starting (%d)", gettid());

    snprintf(email, sizeof(email), "%s@%s", username, domain);
    snprintf(ipstring2, sizeof(ipstring2), "%s", MLFIPRIV->ipstring);

    /* Lookup mailfrom */
    if(spasm_lookup_wl(email, MLFIPRIV->mailfrom)) rv = 1;
    /* Lookup mailfrom hostname */
    else if ((buf = strchr(MLFIPRIV->mailfrom, '@')) != NULL && spasm_lookup_wl(email, buf)) rv = 1;
    /* Lookup hostname */
    else if(spasm_lookup_wl(email, MLFIPRIV->hostname)) rv = 1;
    /* Lookup ip address */
    else if(spasm_lookup_wl(email, ipstring2)) rv = 1;
    else {
	/* Check all subdomains in hostname */
	buf = MLFIPRIV->hostname;
        while((buf = strchr(buf, '.')) && rv == 0) {
	    buf++;
	    if(spasm_lookup_wl(email, buf)) rv = 1;
	}
	/* Check all pseudo-classful network blocks in ip address */
	while((buf = strrchr(ipstring2, '.')) && rv == 0) {
	    *buf = '\0';
	    if(spasm_lookup_wl(email, ipstring2)) rv = 1;
	}
	/* Lookup all subdomains of mailfrom hostname in @-notation */
	if(rv == 0 && (buf = strchr(MLFIPRIV->mailfrom, '@')) != NULL) {
	    while((buf = strchr(buf, '.')) && rv == 0) {
		buf++;
		snprintf(dombuf, sizeof(dombuf), "@%s", buf);
		if(spasm_lookup_wl(email, dombuf)) rv = 1;
	    }
	}
    }

    /* Check CIDR blocks */
    if(rv == 0) {
	if(spasm_cidr_wlbl(SPASM_WHITELIST, email, MLFIPRIV->ipstring)) rv = 1;
	else if(spasm_cidr_wlbl(SPASM_WHITELIST, domain, MLFIPRIV->ipstring)) rv = 1;
    }

    /* Check regex entries */
    if(rv == 0 && spasm_regex_wlbl(SPASM_WHITELIST, email, MLFIPRIV->mailfrom, MLFIPRIV->hostname)) rv = 1;
    else if(rv == 0 && spasm_regex_wlbl(SPASM_WHITELIST, domain, MLFIPRIV->mailfrom, MLFIPRIV->hostname)) rv = 1;

    if(spasm_config->debug) syslog(LOG_DEBUG, "(userwl_lookup) Finished (%d)", gettid());
    return rv;
}

/* user_lookup() - verifies a user for unknownusers count */
int
user_lookup(char *username, char *domain)
{
    /* Return true if no verifications are selected */
    if(!spasm_config->verifypasswd && !spasm_config->verifyaliases &&
	!spasm_config->verifyspasm && !spasm_config->verifycustom)
    	return 1;

    /* First, check local users */
    if(strcasecmp(domain, spasm_config->defaultdomain) == 0) {
	/* passwd file */
	if(spasm_config->verifypasswd) {
	    if(getpwnam(username) != NULL) return 1;
	}
	/* sendmail aliases file */
	if(spasm_config->verifyaliases) {
            FILE *fp;
            char buf[256];
	    char *ptr;

            if((fp = fopen(spasm_config->aliasesfile, "r")) == NULL) {
		syslog(LOG_ERR, "(user_lookup) Error opening %s: %m (%d)", spasm_config->aliasesfile, gettid());
		return 1;
	    }
            while(fgets(buf, sizeof(buf), fp) != NULL) {
		if((ptr = strchr(buf, ':')) != NULL) {
		    *ptr = '\0';
		    if(strcasecmp(buf, username) == 0) {
	                fclose(fp);
	                return 1;
		    }
                }
            }
            fclose(fp);
        }
    }

    /* Next, does user exist in spasm? */
    if(spasm_config->verifyspasm) {
	if(verify_spasm_user(username, domain)) return 1;
    }

    /* Next, do custom lookup */
    if(spasm_config->verifycustom) {
	if(verify_custom(username, domain)) return 1;
    }

    /* We couldn't find the user. Sorry! */
    return 0;
}

/* dnsbl_lookup() - Lookup a hostname, return 1 if found, 0 if !found */
int
dnsbl_lookup(char *hostname)
{
    struct addrinfo *hstaddr;
    int res;
    int rc = 0;

    if(spasm_config->debug) syslog(LOG_DEBUG, "(dnsbl_lookup) Starting (%d)", gettid());
    if(hostname == NULL) return 0;

    res = getaddrinfo(hostname, NULL, NULL, &hstaddr);
    if(res == 0) {
	freeaddrinfo(hstaddr);
	rc = 1;
	if(spasm_config->debug) syslog(LOG_DEBUG, "(dnsbl_lookup) Found %s (%d)", hostname, gettid());
    }

    if(spasm_config->debug) syslog(LOG_DEBUG, "(dnsbl_lookup) Finished (%d)", gettid());
    return rc;
}

/* search_regex_spamtraps() = Run each entry in spamtrap database against arguments */
int
search_regex_spamtraps(char *key)
{
    struct spasm_list *entries = NULL;
    struct spasm_list *cur = NULL;
    int rv = 0;

    if(spasm_config->spamtrapdb == NULL) return 0;

    entries = spasm_get_re(spasm_config->spamtrapdb);
    cur = entries;
    while(cur != NULL) {
	if(search_regex(cur->entry, key) == 1) {
	    rv = 1;
	    break;
	}
	cur = cur->next;
    }
    cur = entries;
    while(cur != NULL) {
	entries = cur->next;
	free(cur);
	cur = entries;
    }
    return rv;
}

/* filter_local() - Lookup entries in local db4 hash */
int
filter_local(char *dbname, char *mailfrom, char *mailfromhost, char *hostname, char *ipstring)
{
    char *buf;
    char ipstring2[16];
    char hostbuf[SMTP_DOMAIN_MAX+2];
    int rv = 0;

    if(dbname == NULL) return 0;

    if(spasm_config->debug) syslog(LOG_DEBUG, "(filter_local) Starting (%d)", gettid());

    snprintf(ipstring2, sizeof(ipstring2), "%s", ipstring);
    snprintf(hostbuf, sizeof(hostbuf), "@%s", hostname);

    /* Lookup mailfrom address */
    if(spasm_lookup_bl(dbname, mailfrom)) rv = 1;

    /* Lookup mailfrom address's domain */
    else if(spasm_lookup_bl(dbname, mailfromhost)) rv = 1;

    /* Lookup mailfrom address's domain using @domain as a precaution */
    else if((mailfromhost != NULL) && spasm_lookup_bl(dbname, mailfromhost-1)) rv = 1;

    /* Lookup server hostname */
    else if(spasm_lookup_bl(dbname, hostname)) rv = 1;

    /* Lookup server hostname using @domain as a precaution */
    else if(spasm_lookup_bl(dbname, hostbuf)) rv = 1;

    /* Lookup server ip address */
    else if(spasm_lookup_bl(dbname, ipstring2)) rv = 1;
    
    /* Lookup all subdomains of server hostname */
    if(hostname != NULL) {
        buf = hostname;
        while((buf = strchr(buf, '.')) && rv != 1) {
	    buf++;
	    if(spasm_lookup_bl(dbname, buf)) rv = 1;
        }
    }

    /* Lookup all subdomains of mailfrom hostname */
    if(mailfromhost != NULL) {
	buf = mailfromhost;
	while((buf = strchr(buf, '.')) && rv != 1) {
	    buf++;
	    if(spasm_lookup_bl(dbname, buf)) rv = 1;
	}
    }

    /* Lookup pseudo-classful network blocks in ip address */
    while((buf = strrchr(ipstring2, '.')) && rv != 1) {
	*buf = '\0';
	if(spasm_lookup_bl(dbname, ipstring2)) rv = 1;
    }

    /* Check CIDR blocks */
    if(rv == 0 && spasm_cidr_wlbl(SPASM_BLACKLIST, dbname, ipstring)) rv = 1;

    /* Check regex entries */
    if(rv == 0 && spasm_regex_wlbl(SPASM_BLACKLIST, dbname, mailfrom, hostname)) rv = 1;

    if(spasm_config->debug) syslog(LOG_DEBUG, "(filter_local) Finished (%d)", gettid());
    return rv;
}

static int
test_personal_bl(SMFICTX *ctx, struct spasmRecipient *cur)
{
    char email[SMTP_PATH_MAX+1];
    int rc;

    snprintf(email, sizeof(email), "%s@%s", cur->username, cur->domain);
    rc = (filter_local(email, MLFIPRIV->mailfrom, MLFIPRIV->mailfromhost, MLFIPRIV->hostname, MLFIPRIV->ipstring) ||
            filter_local(cur->domain, MLFIPRIV->mailfrom, MLFIPRIV->mailfromhost, MLFIPRIV->hostname, MLFIPRIV->ipstring));
    return rc;
}

static int
test_local_bl(SMFICTX *ctx)
{
    return (filter_local(spasm_config->localbl,
		    MLFIPRIV->mailfrom,
		    MLFIPRIV->mailfromhost,
		    MLFIPRIV->hostname, MLFIPRIV->ipstring));
}

/* filter_rDNS() - Determine if ip address was resolved */
int
filter_rDNS(char *hostname)
{
    if(hostname == NULL) return 0;
    if(*hostname == '[') return 1;
    return 0;
}

/* filter_rDNS_APNIC() - Determine if APNIC ip address was resolved */
int
filter_rDNS_APNIC(char *ipstring, char *hostname)
{
    if(ipstring == NULL || hostname == NULL) return 0;

    /* Current APNIC Space */
    if((strncmp(ipstring, "58.", 3) == 0) ||
       (strncmp(ipstring, "59.", 3) == 0) ||
       (strncmp(ipstring, "60.", 3) == 0) ||
       (strncmp(ipstring, "61.", 3) == 0) ||
       (strncmp(ipstring, "121.", 4) == 0) ||
       (strncmp(ipstring, "122.", 4) == 0) ||
       (strncmp(ipstring, "123.", 4) == 0) ||
       (strncmp(ipstring, "124.", 4) == 0) ||
       (strncmp(ipstring, "125.", 4) == 0) ||
       (strncmp(ipstring, "126.", 4) == 0) ||
       (strncmp(ipstring, "202.", 4) == 0) ||
       (strncmp(ipstring, "203.", 4) == 0) ||
       (strncmp(ipstring, "210.", 4) == 0) ||
       (strncmp(ipstring, "211.", 4) == 0) ||
       (strncmp(ipstring, "218.", 4) == 0) ||
       (strncmp(ipstring, "219.", 4) == 0) ||
       (strncmp(ipstring, "220.", 4) == 0) ||
       (strncmp(ipstring, "221.", 4) == 0) ||
       (strncmp(ipstring, "222.", 4) == 0)) {
        if(*hostname == '[') return 1;
    }
    return 0;
}

/* filter_rDNS_RIPE() - Determine if RIPE ip address was resolved */
int
filter_rDNS_RIPE(char *ipstring, char *hostname)
{
    if(ipstring == NULL || hostname == NULL) return 0;

    /* Current RIPE Space */
    if((strncmp(ipstring, "62.", 3) == 0) ||
       (strncmp(ipstring, "80.", 3) == 0) ||
       (strncmp(ipstring, "81.", 3) == 0) ||
       (strncmp(ipstring, "82.", 3) == 0) ||
       (strncmp(ipstring, "83.", 3) == 0) ||
       (strncmp(ipstring, "84.", 3) == 0) ||
       (strncmp(ipstring, "85.", 3) == 0) ||
       (strncmp(ipstring, "86.", 3) == 0) ||
       (strncmp(ipstring, "87.", 3) == 0) ||
       (strncmp(ipstring, "88.", 3) == 0) ||
       (strncmp(ipstring, "89.", 3) == 0) ||
       (strncmp(ipstring, "90.", 3) == 0) ||
       (strncmp(ipstring, "91.", 3) == 0) ||
       (strncmp(ipstring, "193.", 4) == 0) ||
       (strncmp(ipstring, "194.", 4) == 0) ||
       (strncmp(ipstring, "195.", 4) == 0) ||
       (strncmp(ipstring, "212.", 4) == 0) ||
       (strncmp(ipstring, "213.", 4) == 0) ||
       (strncmp(ipstring, "217.", 4) == 0)) {
	if(*hostname == '[') return 1;
    }
    return 0;
}

/* filter_HELO() - Reject if no HELO/EHLO given, or argument != resolved hostname */
int
filter_HELO(char *helodomain, char *hostdomain)
{
    if(hostdomain == NULL) return 0;
    if(helodomain == NULL) return 1;
    if(strcasecmp(helodomain, hostdomain) != 0) return 1;
    return 0;
}

/* filter_MAIL_HOST() - Determine if domain of mailfrom == domain of hostname */
int
filter_MAIL_HOST(char *mailfromdomain, char *hostdomain)
{
    if(mailfromdomain == NULL || hostdomain == NULL) return 0;
    if(strcasecmp(mailfromdomain, hostdomain) == 0) return 0;
    else return 1;
}

/* filter_AV() - Scan message for viruses */
const char*
filter_AV(char *qfile)
{
    if(qfile == NULL) return NULL;

#ifdef WITH_CLAMAV
    if(spasm_config->clamav) {
	int status, rv;
	const char *virname;
	struct cl_limits limits;
	unsigned long int size=0;
	struct stat fstat;

	if(clamav_root == NULL) return NULL;

        /* Reload database if necessary */
        pthread_mutex_lock(&mutexLock);
        rv = stat(cl_retdbdir(), &fstat);
        if(rv == 0 && fstat.st_mtime > clamav_lastload) {
            int virnum = 0;
	    memset(&limits, 0, sizeof(limits));
	    /* Stop new virus scans */
	    pthread_mutex_lock(&mutexLock_clamav_accept);
	    clamav_accept = 0;
	    pthread_mutex_unlock(&mutexLock_clamav_accept);

	    /* Wait for current scans to stop */
	    pthread_mutex_lock(&mutexLock_clamav_num);
	    while(clamav_num > 0) pthread_cond_wait(&cond_clamav_num, &mutexLock_clamav_num);
	    pthread_mutex_unlock(&mutexLock_clamav_num);

	    /* Start reloading database */
            clamav_lastload = fstat.st_mtime;
            cl_free(clamav_root);
            clamav_root = NULL;
            if((rv = cl_loaddbdir(cl_retdbdir(), &clamav_root, &virnum))) {
	        /* Restart virus scans */
	        pthread_mutex_lock(&mutexLock_clamav_accept);
	        clamav_accept = 1;
	        pthread_cond_broadcast(&cond_clamav_accept);
	        pthread_mutex_unlock(&mutexLock_clamav_accept);
                syslog(LOG_ERR, "(filter_AV) Error reloading database: %d (%d)", rv, gettid());
                pthread_mutex_unlock(&mutexLock);
                return NULL;
            }
            if(clamav_root == NULL) {
                syslog(LOG_ERR, "(filter_AV) Error reloading database (%d)", gettid());
            }
            if((rv = cl_build(clamav_root)) != 0) {
                /* Restart virus scans */
                pthread_mutex_lock(&mutexLock_clamav_accept);
                clamav_accept = 1;
                pthread_cond_broadcast(&cond_clamav_accept);
                pthread_mutex_unlock(&mutexLock_clamav_accept);
                syslog(LOG_ERR, "(filter_AV) Error reinitialising database: %d (%d)", rv, gettid());
	        pthread_mutex_unlock(&mutexLock);
                return NULL;
            }
	    /* Restart virus scans */
            pthread_mutex_lock(&mutexLock_clamav_accept);
            clamav_accept = 1;
            pthread_cond_broadcast(&cond_clamav_accept);
            pthread_mutex_unlock(&mutexLock_clamav_accept);
            syslog(LOG_NOTICE, "(filter_AV) Database reloaded (%d entries) (%d)", virnum, gettid());
        } else if(rv != 0) {
            syslog(LOG_ERR, "(filter_AV) Error checking database status: %d (%d)", rv, gettid());
        }
        pthread_mutex_unlock(&mutexLock);

        /* Determine whether we can start a new scan */
        pthread_mutex_lock(&mutexLock_clamav_accept);
        while(!clamav_accept) pthread_cond_wait(&cond_clamav_accept, &mutexLock_clamav_accept);
        pthread_mutex_unlock(&mutexLock_clamav_accept);

        pthread_mutex_lock(&mutexLock_clamav_num);
        clamav_num++;
        pthread_mutex_unlock(&mutexLock_clamav_num);
        if((status = cl_scanfile(qfile, &virname, &size, clamav_root, &limits, CL_SCAN_STDOPT)) != CL_VIRUS) {
	    /* Message is clean */
	    if(status != CL_CLEAN) syslog(LOG_ERR, "(filter_AV) Error scanning %s: %d (%d)", qfile, status, gettid());
	    virname = NULL;
        }
        pthread_mutex_lock(&mutexLock_clamav_num);
        clamav_num--;
        if(clamav_num == 0) pthread_cond_broadcast(&cond_clamav_num);
        pthread_mutex_unlock(&mutexLock_clamav_num);
        return virname;
    }
#endif /* WITH_CLAMAV */

    return NULL;
}

/* filter_SA() - Check message against SpamAssassin */
int
filter_SA(char *qfile, int required_hits)
{
    int rv = 0;
    int fd, sock;
    char fbuf[1024];
    char response[SA_RESPONSE_MAX];
    size_t size, len;
    char *sa_report;
    char *chr;
    char header[32767];
    struct sockaddr_in sin;
    float sa_score;
    struct stat qfstat;

    if(qfile == NULL || required_hits == 0) {
        return 0;
    }
    if((fd = open(qfile, O_RDONLY)) < 0) {
        syslog(LOG_ERR, "(filter_SA) Error opening qfile: %m (%d)", gettid());
        return 0;
    }
    fstat(fd, &qfstat);
    if(qfstat.st_size > MAX_SA_SIZE) {
	close(fd);
	return 0;
    }
    if((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	syslog(LOG_ERR, "(filter_SA) Error creating socket: %m (%d)", gettid());
	close(fd);
	return 0;
    }
    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_port = spasm_config->saport;
    sin.sin_addr = *(spasm_config->sahost);
    if(connect(sock, (struct sockaddr *)&sin, sizeof(struct sockaddr)) == -1) {
	syslog(LOG_ERR, "(filter_SA) Error connecting to spamd: %m (%d)", gettid());
	close(fd);
	close(sock);
	return 0;
    }
    size = snprintf(fbuf, sizeof(fbuf), "SYMBOLS SPAMC/1.3\r\nUser: %s\r\n\r\n", spasm_config->spasmuser);
    send(sock, fbuf, size, 0);
    while((size = read(fd, fbuf, sizeof(fbuf))) > 0) {
        send(sock, fbuf, size, 0);
    }
    close(fd);
    shutdown(sock, 1);
    memset(fbuf, 0, sizeof(fbuf));
    len=1; /* initialise with length of required \0 terminator */
    memset(response, 0, sizeof(response));
    while((size = read(sock, fbuf, sizeof(fbuf))) > 0) {
	len += size;
	if(len < SA_RESPONSE_MAX) strncat(response, fbuf, len);
    }
    close(sock);
    if(len == 1) {
	syslog(LOG_ERR, "(filter_SA) Error reading response from spamd (%d)", gettid());
	return 0;
    }
    sa_report = strstr(response, "Spam:");
    sscanf(sa_report, "Spam: %*s ; %f / %*f", &sa_score);
    sa_report = strstr(response, "\r\n\r\n");
    sa_report += 4;
    if(sa_score >= required_hits) {
	snprintf(fbuf, sizeof(fbuf), "X-spasm-SA: %.2f ; ", sa_score);
	len = strlen(fbuf) + 3;
	memset(header, 0, sizeof(header));
	strncpy(header, fbuf, sizeof(header)-1);
	while(1) {
	    char *chr2 = NULL;
	    if((chr = strchr(sa_report, ',')) == NULL) break;
	    *chr = '\0';
	    while(strlen(sa_report) < 50) {
		chr2 = chr;
		*chr = ',';
		chr++;
		if((chr = strchr(chr, ',')) != NULL) {
		    *chr = '\0';
		} else {
		    chr2 = NULL;
		    break;
		}
	    }
	    if(chr2 != NULL) {
		*chr = ',';
		*chr2 = '\0';
	    }
	    len += (strlen(sa_report) + 4);
	    strncat(header, "\r\n\t", sizeof(header)-strlen(header)-1);
	    strncat(header, sa_report, sizeof(header)-strlen(header)-1);
	    if(chr2 == NULL) break;
	    sa_report = chr2 + 1;
	}
	add_header(qfile, header);
	rv = 1;
    }
    return rv;
}

/* filter_DSPAM() - Check message against DSPAM */
int
filter_DSPAM(char *qfile)
{
    int pfd1[2], pfd2[2];
    int fd;
    int pid;
    char fbuf[1024], fbuf2[1024];
    int status;
    size_t size;
    int rv = 0;
        
    if(qfile == NULL) {
        return 0;
    }
    if(pipe(pfd1) < 0) {
        syslog(LOG_ERR, "(filter_DSPAM) Error creating DSPAM pipe pfd1: %m (%d)", gettid());
        return 0;
    }
    if(pipe(pfd2) < 0) {
        syslog(LOG_ERR, "(filter_DSPAM) Error creating DSPAM pipe pfd2: %m (%d)", gettid());
        close(pfd1[0]);
        close(pfd1[1]); 
        return 0;
    }
    if((pid = fork()) < 0) {
        syslog(LOG_ERR, "(filter_DSPAM) Error forking DSPAM process: %m (%d)", gettid());
        close(pfd1[0]);
        close(pfd1[1]);
        close(pfd2[0]);
        close(pfd2[1]);
        return 0; 
    }       
    if(pid == 0) { /* Child process */
        extern int errno;
        char *args[5];

        args[0] = "dspam";
	args[1] = "--classify";
	args[2] = "--user";
	args[3] = spasm_config->spasmuser;
	args[4] = NULL;

        dup2(pfd1[0], 0);
        dup2(pfd2[1], 1);
        close(pfd1[1]);
        close(pfd2[0]);
        execv(spasm_config->dspam, args);
        /* Shouldn't get here */
        syslog(LOG_ERR, "(filter_DSPAM) Error executing %s (%d): %m (%d)", spasm_config->dspam, errno, gettid());
        exit(128);
    }
    close(pfd1[0]);
    close(pfd2[1]);

    if((fd = open(qfile, O_RDONLY)) < 0) {
        syslog(LOG_ERR, "(filter_SA) Error opening qfile: %m (%d)", gettid());
	close(pfd1[1]);
	close(pfd2[0]);
        return 0;
    }           
    while((size = read(fd, fbuf, sizeof(fbuf))) > 0) {
        write(pfd1[1], fbuf, size);
    }   
    close(fd);
    close(pfd1[1]);
    memset(fbuf, 0, sizeof(fbuf));
    memset(fbuf2, 0, sizeof(fbuf2));
    read(pfd2[0], fbuf, sizeof(fbuf));
    close(pfd2[0]);
    if(waitpid(pid, &status, 0) == -1) {
        syslog(LOG_ERR, "(filter_DSPAM) Error waiting for child: %m (%d)", gettid());
    }
    else if(WIFEXITED(status) == 0) {
        syslog(LOG_ERR, "(filter_DSPAM) Exited non-normally: (%d) (%d)", status, gettid());
    }
    else if(WEXITSTATUS(status) != 0) {
	syslog(LOG_ERR, "(filter_DSPAM) DSPAM returned an error (%d) (%d)", status, gettid());
    } else {
	char result[32];
	float prob=0.0, conf=0.0;

	memset(result, 0, sizeof(result));
	sscanf(fbuf, "X-DSPAM-Result: spasm; result=\"%[^\"]\"; probability=%f; confidence=%f ", result, &prob, &conf);
	snprintf(fbuf2, sizeof(fbuf2), "X-spasm-DSPAM: result=\"%s\"; probability=%01.4f; confidence=%02.2f\r\n", result, prob, conf);
	if(strcmp("Spam", result) == 0) rv = 1;
	syslog(LOG_NOTICE, "(filter_DSPAM) %s (%d)", fbuf2, gettid());
	add_header(qfile, fbuf2);
    }
    return rv;
}

/* filter_EXE() - Check body for executable attachment */
int
filter_EXE(char *body)
{
    if(regexec(&re_exe, body, 0, NULL, 0) == 0) return 1; /* Found */
    else return 0; /* Not found */
}

/* deliverQuarantine() - Deliver a quarantine file */
int
deliverQuarantine(char *from, char *rcpt, char *qfile, char *subject)
{
    int pfd[2];
    int pid;
    char fbuf[1024];
    int status;
    FILE *fptr;
    int headers = 1;

    if(from == NULL || rcpt == NULL || qfile == NULL) {
	return -1;
    }
    if((fptr = fopen(qfile, "r")) == NULL) {
	syslog(LOG_ERR, "(deliverQuarantine) Error opening qfile (%s): %m (%d)", qfile, gettid());
	return -1;
    }
    if(pipe(pfd) < 0) {
        syslog(LOG_ERR, "(deliverQuarantine) Error creating mail pipe: %m (%d)", gettid());
	fclose(fptr);
        return -1;
    }
    if((pid = fork()) < 0) {
        syslog(LOG_ERR, "(deliverQuarantine) Error forking mail process: %m (%d)", gettid());
	fclose(fptr);
        close(pfd[0]);
        close(pfd[1]);
        return -1;
    }
    if(pid == 0) { /* Child process */
        extern int errno;
        dup2(pfd[0], 0);
        close(pfd[1]);
	fclose(fptr);
        execl(spasm_config->sendmail, "sendmail", "-f", from, "--", rcpt, 0);
        /* Shouldn't get here */
        syslog(LOG_ERR, "(deliverQuarantine) Error executing sendmail (%d): %m (%d)", errno, gettid());
        exit(128);
    }
    close(pfd[0]);
    while(fgets(fbuf, sizeof(fbuf), fptr) != NULL) {
        if(strcmp(fbuf, "\r\n") == 0) headers = 0;
        if(headers == 1 && subject != NULL && strncasecmp(fbuf, "subject:", 8) == 0) {
            snprintf(fbuf, sizeof(fbuf), "Subject: **** X-spasm **** %s\r\n", subject);
        }
	write(pfd[1], fbuf, strlen(fbuf));
    }
    fclose(fptr);
    close(pfd[1]);
    if(waitpid(pid, &status, 0) == -1) {
	syslog(LOG_ERR, "(deliverQuarantine) Error waiting for child: %m (%d)", gettid());
    }
    else if(WIFEXITED(status) == 0) {
	syslog(LOG_ERR, "(deliverQuarantine) Exited non-normally: (%d) (%d)", status, gettid());
    }
    else if(WEXITSTATUS(status) != 0) {
        syslog(LOG_ERR, "(deliverQuarantine) Error delivering quarantine file %s (%d): %s (%d)", qfile, WEXITSTATUS(status), strerror(WEXITSTATUS(status)), gettid());
        return -1;
    }
    return 0;
}

void
moveQfile(char *src, char *dst, char *subject)
{
    FILE *sf, *df;
    char fbuf[1024];
    int headers = 1;
    
    if(src == NULL || dst == NULL) return;

    if((sf = fopen(src, "r")) == NULL) {
        syslog(LOG_ERR, "(moveQfile) Error opening %s for reading: %m (%d)", src, gettid());
        return;
    }
    if((df = fopen(dst, "w")) == NULL) {
        syslog(LOG_ERR, "(moveQfile) Error opening %s for writing: %m (%d)", src, gettid());
        fclose(sf);
        return;
    }
    while(fgets(fbuf, sizeof(fbuf), sf)) {
        if(strcmp(fbuf, "\r\n") == 0) headers = 0;
        if(headers == 1 && subject != NULL && strncasecmp(fbuf, "subject:", 8) == 0) {
            snprintf(fbuf, sizeof(fbuf), "Subject: **** X-spasm **** %s\r\n", subject);
        }
        fputs(fbuf, df);
    }
    if(ferror(sf)) syslog(LOG_ERR, "(moveQfile) Error reading %s: %d (%d)", src, ferror(sf), gettid());
    if(ferror(df)) syslog(LOG_ERR, "(moveQfile) Error writing %s: %d (%d)", dst, ferror(df), gettid());
    fclose(sf);
    if(fclose(df) != 0) syslog(LOG_ERR, "(moveQfile) Error closing %s: %m (%d)", dst, gettid());
    else if(unlink(src) != 0) syslog(LOG_ERR, "(moveQfile) Error unlinking %s: %m (%d)", src, gettid());
    return;
}

/* Expire Greylist entries */
void
expire_gl_entries(void)
{
    DB *db;
    DBC *dbcp;
    DBT dbk, dbv;
    int dbfd;
    struct flock lockinfo;
    char exptime[SPASM_DBT_MAX+1];
    int rc;
    int count=0;
    char *chr;

    memset(&dbk, '\0', sizeof(DBT));
    memset(&dbv, '\0', sizeof(DBT));
    memset(&lockinfo, '\0', sizeof(lockinfo));
    lockinfo.l_type = F_WRLCK;

    if((rc = db_create(&db, NULL, 0)) != 0) return;
    if((rc = sp_db_open(db, NULL, spasm_config->greylist, NULL, DB_HASH, DB_CREATE, 0644)) != 0) {
        db->close(db, 0);
        return;
    }
    if((rc = db->fd(db, &dbfd)) != 0) {
        db->close(db, 0);
        return;
    }
    if(fcntl(dbfd, F_SETLKW, &lockinfo) == -1) {
        db->close(db, 0);
        return;
    }
    if(db->cursor(db, NULL, &dbcp, 0) != 0) {
        syslog(LOG_ERR, "(expire_gl_entries) Error obtaining database cursor");
        db->close(db, 0);
        return;
    }
    while(dbcp->c_get(dbcp, &dbk, &dbv, DB_NEXT) == 0) {
        memset(exptime, '\0', sizeof(exptime));
        strncpy(exptime, dbv.data, dbv.size);
        if((chr = strchr(exptime, ':')) == NULL) {
            if(time(NULL) > atol(exptime)) {
                count++;
                if(dbcp->c_del(dbcp, 0) != 0) {
                    syslog(LOG_ERR, "(expire_gl_entries) Error removing database entry: %m");
                }
            }
        } else {
            *chr = '\0';
            if(time(NULL) > atol(exptime)) {
                count++;
                if(dbcp->c_del(dbcp, 0) != 0) {
                    syslog(LOG_ERR, "(expire_gl_entries) Error removing database entry: %m");
                }
            }
        }
    }
    dbcp->c_close(dbcp);
    db->close(db, 0);
    syslog(LOG_NOTICE, "(expire_gl_entries) Expired %d entries", count);
    return;
}

int
greylist_lookup(SMFICTX *ctx, char *server)
{
    int rc = 0;
    time_t curtime;
    time_t exptime;
    time_t delaytime;
    char result[SPASM_DBT_MAX+1];
    char *chr;
    char times[128];
    int forcedelay = 0;
    int res;

    /* Expire entries if necessary */
    pthread_mutex_lock(&mutexLock_greylist);
    if(greylist_lastexpire + 900 < time(NULL)) {
        expire_gl_entries();
        greylist_lastexpire = time(NULL);
    }
    pthread_mutex_unlock(&mutexLock_greylist);

    if(server == NULL) return 0;

    memset(times, 0, sizeof(times));
    curtime = time(NULL);

    memset(result, 0, sizeof(result));
    res = spasm_lookup_gl(server, result);

    if(spasm_config->greylistforcebl) forcedelay = test_local_bl(ctx);

    if(res == 0) {			/* No entry */
	delaytime = curtime+spasm_config->greylistdelay;
	exptime = curtime+spasm_config->greylistexpire;
	sprintf(times, "%ld:%ld", exptime, delaytime);
	pthread_mutex_lock(&mutexLock_greylist);
	spasm_insert_gl(server, times);
	pthread_mutex_unlock(&mutexLock_greylist);
	rc = 1;
    } else {					/* Entry found */
	if((chr = strchr(result, ':')) == NULL) { /* No delay time */
	    exptime = atol(result);
	    if(curtime > exptime) {		/* Entry has expired, start over */
		exptime = curtime+spasm_config->greylistexpire;
		delaytime = curtime+spasm_config->greylistdelay;
		sprintf(times, "%ld:%ld", exptime, delaytime);
		pthread_mutex_lock(&mutexLock_greylist);
		spasm_insert_gl(server, times);
		pthread_mutex_unlock(&mutexLock_greylist);
		rc = 1;
	    } else {				/* Entry is valid, accept message */
		if(forcedelay) exptime = curtime;
		else exptime = curtime + spasm_config->greylistupdate;
		sprintf(times, "%ld", exptime);
		pthread_mutex_lock(&mutexLock_greylist);
		spasm_insert_gl(server, times);
		pthread_mutex_unlock(&mutexLock_greylist);
		rc = 0;
	    }
	} else {				/* Check delay time */
	    *chr = '\0';
	    chr++;
	    delaytime = atol(chr);
	    exptime = atol(result);
	    if(curtime > exptime) {		/* Entry has expired, start over */
		exptime = curtime+spasm_config->greylistexpire;
		delaytime = curtime+spasm_config->greylistdelay;
		sprintf(times, "%ld:%ld", exptime, delaytime);
		pthread_mutex_lock(&mutexLock_greylist);
		spasm_insert_gl(server, times);
		pthread_mutex_unlock(&mutexLock_greylist);
		rc = 1;
	    } else {
		if(curtime < delaytime) {	/* Delay hasn't passed */
		    rc = 1;
		} else {			/* Delay has passed, accept message */
		    if(forcedelay) exptime = curtime;
		    else exptime = curtime+spasm_config->greylistupdate;
		    sprintf(times, "%ld", exptime);
		    pthread_mutex_lock(&mutexLock_greylist);
		    spasm_insert_gl(server, times);
		    pthread_mutex_unlock(&mutexLock_greylist);
		    rc = 0;
		}
	    }
	}
    }
    return rc;
}

/* common things to do when a filter triggers a rejected message */
static int
filter_reject_ops(SMFICTX *ctx,
                  struct spasmRecipient *cur,
                  const SPASM_FILTER *pfltr)
{   
    char reject_msg[256];
    
    log_rejection(ctx, cur);
    
    if (pfltr->adhoc_id == SPASM_HELO) {
        /* create custom rejection message ... */
        
        snprintf(reject_msg,
                 sizeof(reject_msg),
                 "Rejected: You are not %s",
                 MLFIPRIV->helohost);
    } else {
        /* 
         * generate rejection message from smtp_rmsg, which can include up
         * to include up to 3 instances of "%s", which are substibuted with
         * the name of the triggered dnsbl.
         */
        
        snprintf(reject_msg,
                 sizeof(reject_msg),
                 pfltr->smtp_rmsg,
                 pfltr->dnsbl,
                 pfltr->dnsbl,
                 pfltr->dnsbl);
    }
    
    return (smfi_setreply(ctx, pfltr->smtp_rc, pfltr->smtp_xrc, reject_msg));
}

int
getUserPrefs(char *username, char *domain, char *key, char *value)
{
    char email[SMTP_PATH_MAX+1];
    int rc = 0;

    snprintf(email, sizeof(email), "%s@%s", username, domain);
    if(spasm_config->debug) syslog(LOG_DEBUG, "(getUserPrefs) Starting (%s -> %s)... (%d)", email, key, gettid());
    rc = spasm_lookup_prefs(email, key, value);
    return rc;
}

int
getFilterSetting(struct filterStatus *fstatus, char *filter)
{
    int i;

    if(fstatus == NULL || filter == NULL) return 0;

    for(i=0; i<NUM_FILTERS; i++) {
	if(strcmp(fstatus[i].filter, filter) == 0)
	    return fstatus[i].setting;
    }
    return 0;
}

/* Compare filter setting with current status, return filter setting */
int
runFilter(struct spasmRecipient *cur, char *filter)
{
    int filtersetting;
    int filtersetting_nolog;
    int curstatus_nolog;

    if(cur == NULL || filter == NULL) return 0;

    filtersetting = getFilterSetting(cur->fstatus, filter);
    filtersetting_nolog = filtersetting;
    curstatus_nolog = cur->status;

    /* Compare without SPASM_LOG settings */
    if(filtersetting_nolog & SPASM_LOG) filtersetting_nolog ^= SPASM_LOG;
    if(curstatus_nolog & SPASM_LOG) curstatus_nolog ^= SPASM_LOG;

    /* Run if filter setting is greater than current status */
    if(filtersetting_nolog > curstatus_nolog) return filtersetting;

    /* Otherwise, do not run */
    return 0;
}

/* Returns -1 on error, 0 on success, 1 on failure */
int
connectHost(unsigned long int ipaddress, char *hostname, int *num_attempts)
{
    struct sockaddr_in sin;
    char buf[256];
    int s, rc;
    char ipstring[16];

    if(ipaddress == 0) {
        /* Lookup IP Address */
        struct addrinfo *hstaddr;

        if(hostname == NULL) return -1;
        if(getaddrinfo(hostname, NULL, NULL, &hstaddr) != 0) return -1;
        ipaddress = ((struct sockaddr_in*)hstaddr->ai_addr)->sin_addr.s_addr;
        freeaddrinfo(hstaddr);
    }

    (*num_attempts)++;
    if(spasm_config->verifymxmaxattempts && (*num_attempts > spasm_config->verifymxmaxattempts)) {
        syslog(LOG_INFO, "(connectHost) MAX_ATTEMPTS reached: %s", hostname);
        return -1;
    }

    snprintf(ipstring, sizeof(ipstring), "%u.%u.%u.%u", NIPQUAD(ipaddress));
    memset(buf, 0, sizeof(buf));
    sin.sin_addr.s_addr = ipaddress;
    sin.sin_port = htons(25);
    sin.sin_family = AF_INET;

    /* Create socket */
    if((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
        syslog(LOG_ERR, "(connectHost) Error creating socket: %m (%d)", gettid());
        return -1;
    }
    /* Set non-blocking */
    if(fcntl(s, F_SETFL, O_NONBLOCK) != 0) {
        syslog(LOG_ERR, "(connectHost) Error setting socket non-blocking: %m (%d)", gettid());
	close(s);
        return -1;
    }
    /* Attempt connection */
    rc = connect(s, (struct sockaddr*)&sin, sizeof(sin));
    if(rc < 0 && errno == EINPROGRESS) {
        /* Connection is in progress */
        int rc2;
        struct pollfd pfd;

        pfd.fd = s;
        pfd.events = POLLIN;

        rc2 = poll(&pfd, 1, 45000); /* 45 second timeout */
        if(rc2 == 0) {
            /*syslog(LOG_ERR, "(connectHost) Connection timed out to %s (%s) (%d)", ipstring, hostname, gettid());*/
	    close(s);
            return 1;
        }
        if(rc2 < 0) {
            syslog(LOG_ERR, "(connectHost) Error polling fd for %s (%s): %m (%d)", ipstring, hostname, gettid());
	    close(s);
            return -1;
        }
    } else if(rc < 0 && errno == ECONNREFUSED) {
        /*syslog(LOG_ERR, "(connectHost) Connection refused by %s (%d)", ipstring, gettid());*/
	close(s);
        return 1;
    } else if(rc < 0) {
        syslog(LOG_ERR, "(connectHost) Connection error (%d): %m (%d)", rc, gettid());
	close(s);
        return -1;
    }
    if(read(s, buf, sizeof(buf)-1) > 0) {
        if(strncmp(buf, "220", 3) != 0 && strncmp(buf, "421", 3) != 0) {
            syslog(LOG_ERR, "(connectHost) Banner from %s (%s): %s (%d)", ipstring, hostname, buf, gettid());
            snprintf(buf, sizeof(buf), "QUIT\n");
            write(s, buf, strlen(buf));
            close(s);
            return 1;
        } else {
            snprintf(buf, sizeof(buf), "QUIT\n");
            write(s, buf, strlen(buf));
            close(s);
            return 0;
        }
    } else {
        if(errno != ECONNREFUSED)
            syslog(LOG_ERR, "(connectHost) Read error %d: %m (%s) (%d)", errno, ipstring, gettid());
        close(s);
        return 1;
    }   
    /* Shouldn't get here */
    close(s);
    return -1;
}       

/* Returns -1 on error, 0 on success, 1 on failure */
int         
getMX(char *hostname, int *num_attempts)
{       
    struct __res_state res;
    ns_msg msg;
    char buf[NS_MAXMSG];
    int msg_size, rc, rrnum;
    char mxhost[NS_MAXDNAME];
        
    if(hostname == NULL) return -1;
    res_ninit(&res);
    if((msg_size = res_nquery(&res, hostname, ns_c_in, ns_t_mx, buf, sizeof(buf))) <= 0) {
        rc = res.res_h_errno;
        res_nclose(&res);
        if(rc != NO_DATA)
            syslog(LOG_ERR, "(getMX) DNS Error: %s (%s) (%d)", DNS_ERRNO[rc], hostname, gettid());
        return -1;
    }
    if((rc = ns_initparse(buf, msg_size, &msg)) < 0) {
        syslog(LOG_ERR, "(getMX) ns_initparse failed (%d)", gettid());
        res_nclose(&res);
        return -1;
    }
    rc = -1;
    for(rrnum=0; rrnum<ns_msg_count(msg, ns_s_an) && rc != 0; rrnum++) {
        ns_rr nsrr;
        if(ns_parserr(&msg, ns_s_an, rrnum, &nsrr) < 0) {
            syslog(LOG_ERR, "(getMX) ns_parserr failed (%d)", gettid());
            continue;
        }
        if(ns_rr_rdlen(nsrr) < NS_INT16SZ) {
            syslog(LOG_ERR, "(getMX) rdlen < NS_INT16SZ (%d)", gettid());
            continue;
        }
        if(ns_rr_type(nsrr) == ns_t_cname) {
            dn_expand(ns_msg_base(msg), ns_msg_base(msg)+ns_msg_size(msg), ns_rr_rdata(nsrr), mxhost, sizeof(mxhost));
            rc = connectHost(0, mxhost, num_attempts);
        } else if(ns_rr_type(nsrr) == ns_t_mx) {
            u_int p;
            NS_GET16(p, (u_char*)ns_rr_rdata(nsrr));
            dn_expand(ns_msg_base(msg), ns_msg_base(msg)+ns_msg_size(msg), ns_rr_rdata(nsrr), mxhost, sizeof(mxhost));
            rc = connectHost(0, mxhost, num_attempts);
        } else {
            syslog(LOG_ERR, "(getMX) Unknown rrtype for %s: %d (%d)", hostname, ns_rr_type(nsrr), gettid());
        }
    }
    res_nclose(&res);
    return rc;
}

/* Expire MX Verification cache entries */
void
expire_mx_entries(void)
{
    DB *db;
    DBC *dbcp;
    DBT dbk, dbv;
    int dbfd;
    struct flock lockinfo;
    char exptime[13];
    int rc;
    int count=0;

    memset(&dbk, '\0', sizeof(DBT));
    memset(&dbv, '\0', sizeof(DBT));
    memset(&lockinfo, '\0', sizeof(lockinfo));
    lockinfo.l_type = F_WRLCK;

    if((rc = db_create(&db, NULL, 0)) != 0) return;
    if((rc = sp_db_open(db, NULL, spasm_config->verifymx, NULL, DB_HASH, DB_CREATE, 0644)) != 0) {
        db->close(db, 0);
        return;
    }
    if((rc = db->fd(db, &dbfd)) != 0) {
        db->close(db, 0);
        return;
    }
    if(fcntl(dbfd, F_SETLKW, &lockinfo) == -1) {
        db->close(db, 0);
        return;
    }
    if(db->cursor(db, NULL, &dbcp, 0) != 0) {
        syslog(LOG_ERR, "(expire_mx_entries) Error obtaining database cursor");
        db->close(db, 0);
        return;
    }
    while(dbcp->c_get(dbcp, &dbk, &dbv, DB_NEXT) == 0) {
        memset(exptime, '\0', sizeof(exptime));
        strncpy(exptime, dbv.data, dbv.size);
        if(strtoul(exptime+2, (char **) NULL, 10) < time(NULL)) {
            count++;
            if(dbcp->c_del(dbcp, 0) != 0) {
                syslog(LOG_ERR, "(expire_mx_entries) Error removing database entry: %m");
            }
        }
    }
    dbcp->c_close(dbcp);
    db->close(db, 0);
    syslog(LOG_NOTICE, "(expire_mx_entries) Expired %d entries", count);
    return;
}

/* Returns -1 on error, 0 on success, 1 on failure */
int 
verifyMX(char *hostname)
{
    int rc = -1;
    int num_attempts = 0;
        
    if(hostname == NULL) return -1;
    
    /* Expire entries if necessary */ 
    pthread_mutex_lock(&mutexLock_verifymx);
    if(verifymx_lastexpire + spasm_config->verifymxexpireinterval < time(NULL)) {
        expire_mx_entries();
        verifymx_lastexpire = time(NULL);
    }   
    pthread_mutex_unlock(&mutexLock_verifymx);
       
    if((rc = getMX(hostname, &num_attempts)) != 0) {
        rc = connectHost(0, hostname, &num_attempts);
    }   
    return rc;
}   

sfsistat
mlfi_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
    struct mlfiPriv *priv;
    struct stat filestat;
    char ipstring[16];
    char *buf;

    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_connect) Starting (thread %d)", gettid());

    /* If mtime of config file has changed, reload */
    if(stat(SPASM_CONF, &filestat) == 0) {
	if(filestat.st_mtime > spasm_config->mtime) reloadConfig();
    } else syslog(LOG_ERR, "(mlfi_connect) Unable to stat config file: %m");

    /* Check for long hostname and take safest action */
    if(strlen(hostname) > SMTP_DOMAIN_MAX) {
	syslog(LOG_WARNING, "(mlfi_connect) Hostname exceed maximum length: %s (%d)", hostname, gettid());
	return SMFIS_CONTINUE;
    }

    /* Don't scan messages from localhost */
    memset(ipstring, 0, sizeof(ipstring));
    snprintf(ipstring, sizeof(ipstring), "%u.%u.%u.%u", NIPQUAD(((struct sockaddr_in*)hostaddr)->sin_addr.s_addr));
    if(inCIDR(ipstring, "127.0.0.1/32")) return SMFIS_ACCEPT;

    /* Don't scan messages from global IP whitelist */
    if(spasm_cidr_wlbl(SPASM_WHITELIST, spasm_config->globalipwl, ipstring)) return SMFIS_ACCEPT;

    if((priv = calloc(1, sizeof(*priv))) == NULL) {
	syslog(LOG_ERR, "(mlfi_connect) Error allocating memory for priv: %m (%d)", gettid());
	return SMFIS_CONTINUE;
    }
    strncpy(priv->hostname, hostname, sizeof(priv->hostname)-1);
    if(strlen(hostname) > sizeof(priv->hostname)-1)
	syslog(LOG_NOTICE, "(mlfi_connect) Warning: hostname is too long for array: %s (%d)", hostname, gettid());
    lowercase(priv->hostname);
    priv->ipaddress = ((struct sockaddr_in *) hostaddr)->sin_addr.s_addr;
    snprintf(priv->ipstring, sizeof(priv->ipstring), "%u.%u.%u.%u", NIPQUAD(priv->ipaddress));
    if(*priv->hostname != '[') priv->hostdomain = getSLD(priv->hostname);

    smfi_setpriv(ctx, priv);

    /* Check global IP blacklist */
    if(spasm_lookup_bl(spasm_config->globalipbl, ipstring)) {
	syslog(LOG_INFO, "(mlfi_connect) Rejected %s (%s) by global blacklist (%d)", priv->ipstring, priv->hostname, gettid());
        if(spasm_config->stats != NULL) {
            pthread_mutex_lock(&mutexLock_stats);
            spasm_update_stats("global_ip_bl");
            pthread_mutex_unlock(&mutexLock_stats);
        }
	smfi_setreply(ctx, "550", "5.1.1", "Blacklisted by system");
	return SMFIS_REJECT;
    }

    /* Lookup pseudo-classful network blocks in ip address */
    while((buf = strrchr(ipstring, '.'))) {
	*buf = '\0';
	if(spasm_lookup_bl(spasm_config->globalipbl, ipstring)) {
	    syslog(LOG_INFO, "(mlfi_connect) Rejected %s (%s) by global blacklist (thread %d)", priv->ipstring, priv->hostname, gettid());
	    if(spasm_config->stats != NULL) {
		pthread_mutex_lock(&mutexLock_stats);
		spasm_update_stats("global_ip_bl");
		pthread_mutex_unlock(&mutexLock_stats);
	    }
	    smfi_setreply(ctx, "550", "5.1.1", "Blacklisted by system");
	    return SMFIS_REJECT;
	}
    }

    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_connect) Finished (%d)", gettid());

    return SMFIS_CONTINUE;
}

sfsistat
mlfi_helo(SMFICTX *ctx, char *helohost)
{
    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_helo) Starting (%d)", gettid());
    if(MLFIPRIV == NULL) return SMFIS_CONTINUE;
    if(helohost == NULL) {
	syslog(LOG_INFO, "(mlfi_helo) No HELO issued (%d)", gettid());
	return SMFIS_CONTINUE;
    }

    /* Check for a HELO argument that's in the Global Whitelist. It seems to be
    ** common now for spammers to give a HELO of the destination's MX IP address.
    ** The MX server, however, should be a part of the Global Whitelist, and 
    ** therefore no server that gets to this point should ever give such a HELO,
    ** even ignoring the fact that a HELO argument should be a hostname, not an
    ** IP address. */
    if(isIP(helohost) && spasm_cidr_wlbl(SPASM_WHITELIST, spasm_config->globalipwl, helohost)) {
	syslog(LOG_NOTICE, "(mlfi_helo) Global WL HELO (%s) from %s (%s) (%d)", helohost, MLFIPRIV->ipstring, MLFIPRIV->hostname, gettid());
        if(spasm_config->stats != NULL) {
            pthread_mutex_lock(&mutexLock_stats);
            spasm_update_stats("global_wl_helo");
            pthread_mutex_unlock(&mutexLock_stats);
        }
	pthread_mutex_lock(&mutexLock_globalipbl);
	add_to_blacklist(spasm_config->globalipbl, MLFIPRIV->ipstring);
	pthread_mutex_unlock(&mutexLock_globalipbl);
	smfi_setreply(ctx, "550", "5.1.1", "Fraudulent HELO argument");
	return SMFIS_REJECT;
    }

    /* Check for long hostname and take safest action */
    if(strlen(helohost) > SMTP_DOMAIN_MAX) {
	syslog(LOG_WARNING, "(mlfi_helo) Hostname exceeds maximum length: %s (%d)", helohost, gettid());
	return SMFIS_CONTINUE;
    }

    strncpy(MLFIPRIV->helohost, helohost, sizeof(MLFIPRIV->helohost)-1);
    if(strlen(helohost) > sizeof(MLFIPRIV->helohost)-1)
	syslog(LOG_NOTICE, "(mlfi_helo) Warning: HELO is too long for array: %s (%d)", helohost, gettid());
    lowercase(MLFIPRIV->helohost);
    MLFIPRIV->helodomain = getSLD(MLFIPRIV->helohost);

    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_helo) Finished (%d)", gettid());
    return SMFIS_CONTINUE;
}

sfsistat
mlfi_envfrom(SMFICTX *ctx, char **argv)
{
    char *chr;
    regmatch_t pmatch[3];
    char envfrom[SMTP_PATH_MAX+1];
    char *auth_authen;
    char *auth_type;

    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_envfrom) Starting (%d)", gettid());
    if(MLFIPRIV == NULL) return SMFIS_CONTINUE;

    /* Check for long reverse-path and take safest action */
    if(strlen(*argv) > SMTP_PATH_MAX) {
	syslog(LOG_WARNING, "(mlfi_envfrom) Reverse-Path exceeds maximum length: %s (%d)", *argv, gettid());
	return SMFIS_CONTINUE;
    }

    auth_authen = smfi_getsymval(ctx, "{auth_authen}");
    auth_type = smfi_getsymval(ctx, "{auth_type}");
    if(auth_authen == NULL || auth_type == NULL) {
        /* Check greylist */
        if(spasm_config->greylist != NULL) {
            if(greylist_lookup(ctx, MLFIPRIV->ipstring)) {
                if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_connect) Greylist delayed %s (%s) (%d)", MLFIPRIV->ipstring, MLFIPRIV->hostname, gettid());
                if(spasm_config->stats != NULL) {
                    pthread_mutex_lock(&mutexLock_stats);
                    spasm_update_stats("greylist");
                    pthread_mutex_unlock(&mutexLock_stats);
                }
                smfi_setreply(ctx, "451", "4.7.1", "Greylisted: Please try again later");
                return SMFIS_TEMPFAIL;
            } else {
                syslog(LOG_INFO, "(mlfi_connect) Connection from %s (%s) (thread %d)", MLFIPRIV->ipstring, MLFIPRIV->hostname, gettid());
                if(spasm_config->stats != NULL) {
                    pthread_mutex_lock(&mutexLock_stats);
                    spasm_update_stats("connect");
                    pthread_mutex_unlock(&mutexLock_stats);
                }
            }
        } else {
            syslog(LOG_INFO, "(mlfi_connect) Connection from %s (%s) (thread %d)", MLFIPRIV->ipstring, MLFIPRIV->hostname, gettid());
            if(spasm_config->stats != NULL) {
                pthread_mutex_lock(&mutexLock_stats);
                spasm_update_stats("connect");
                pthread_mutex_unlock(&mutexLock_stats);
            }
	}
    } else {
	syslog(LOG_NOTICE, "(mlfi_envfrom) Authenticated user: %s using %s (%d)", auth_authen, auth_type, gettid());
    }

    memset(envfrom, 0, sizeof(envfrom));
    if(strcmp(*argv, "<>") == 0) strncpy(envfrom, "<>", sizeof(envfrom)-1);
    else if(**argv == '<') strncpy(envfrom, (*argv)+1, sizeof(envfrom)-1);
    else strncpy(envfrom, *argv, sizeof(envfrom)-1);

    if(strcmp(*argv, "<>") != 0) {
        chr = strchr(envfrom, '>');
        if(chr != NULL) *chr = '\0';
    }
    lowercase(envfrom);

    /* SRS0 Check */
    if(regexec(&re_srs0, envfrom, 3, pmatch, 0) == 0 && pmatch[1].rm_so != -1 && pmatch[2].rm_so != -1 &&
        pmatch[1].rm_eo-pmatch[1].rm_so+1 < SMTP_PATH_MAX && pmatch[2].rm_eo-pmatch[2].rm_so+1 < SMTP_PATH_MAX) {
        char envuser[SMTP_PATH_MAX+1];
        char envdom[SMTP_PATH_MAX+1];
        snprintf(envdom, pmatch[1].rm_eo-pmatch[1].rm_so+1, "%s", envfrom+pmatch[1].rm_so); 
        snprintf(envuser, pmatch[2].rm_eo-pmatch[2].rm_so+1, "%s", envfrom+pmatch[2].rm_so);
        snprintf(MLFIPRIV->mailfrom, sizeof(MLFIPRIV->mailfrom), "%s@%s", envuser, envdom);
    } else {                                                              
        strncpy(MLFIPRIV->mailfrom, envfrom, sizeof(MLFIPRIV->mailfrom)-1);
    }

    chr = strchr(MLFIPRIV->mailfrom, '@');
    if(chr != NULL) {
	chr++;
	MLFIPRIV->mailfromhost = chr;
	MLFIPRIV->mailfromdomain = getSLD(MLFIPRIV->mailfromhost);
    }

    /* MX Verification Callback */
    if(spasm_config->verifymx != NULL) {
	int rc = 0;
	char result[SPASM_DBT_MAX+1];
	char exptime[13];

	memset(result, 0, sizeof(result));
	pthread_mutex_lock(&mutexLock_verifymx);
	rc = spasm_lookup_mx(MLFIPRIV->mailfromhost, result);
	pthread_mutex_unlock(&mutexLock_verifymx);
	if(rc == 1) {
	    if(result[0] == '0') rc = 0;
	    else rc = 1;
	} else {
	    /* Not found in cache */
	    if((rc = verifyMX(MLFIPRIV->mailfromhost)) == 1) {
		sprintf(exptime, "1-%lu", time(NULL)+spasm_config->verifymxexpire);
		pthread_mutex_lock(&mutexLock_verifymx);
		spasm_insert_mx(MLFIPRIV->mailfromhost, exptime);
		pthread_mutex_unlock(&mutexLock_verifymx);
	    } else {
		sprintf(exptime, "0-%lu", time(NULL)+spasm_config->verifymxexpire);
		pthread_mutex_lock(&mutexLock_verifymx);
		spasm_insert_mx(MLFIPRIV->mailfromhost, exptime);
		pthread_mutex_unlock(&mutexLock_verifymx);
	    }
	}
	if(rc == 1) {
	    syslog(LOG_ERR, "Rejected %s (%s): Failed MX Verification for %s (%d)", MLFIPRIV->ipstring, MLFIPRIV->hostname, MLFIPRIV->mailfromhost, gettid());
	    if(spasm_config->stats != NULL) {
		pthread_mutex_lock(&mutexLock_stats);
		spasm_update_stats("verifymx");
		pthread_mutex_unlock(&mutexLock_stats);
	    }
	    smfi_setreply(ctx, "421", "4.0.0", "Can't connect to a valid MX for your hostname");
	    return SMFIS_TEMPFAIL;
	}
    }

    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_envfrom) Finished (%d)", gettid());
    return SMFIS_CONTINUE;
}

static sfsistat
mlfi_envrcpt(SMFICTX *ctx, char **argv)
{
    char *fqdn;
    char *chr;
    struct spasmRecipient *cur = NULL, *tmpcur;
    char buf[256];
    int result;
    int i;
    int fd = -1;
    int domlogex = 0;
    int founduser = 0;
    const SPASM_FILTER *pfltr;
    char rcpt[SMTP_PATH_MAX+1];
    char username[SMTP_LOCALPART_MAX+1];
    char domain[SMTP_DOMAIN_MAX+1];
    char alias[SMTP_PATH_MAX+1];
    struct spasm_list *entries;
    struct spasm_list *first;

    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_envrcpt) Starting (%d)", gettid());
    if(MLFIPRIV == NULL) return SMFIS_CONTINUE;

    MLFIPRIV->num_rcpts++;

    /* Check for long forward-path and take safest action */
    if(strlen(*argv) > SMTP_PATH_MAX) {
	syslog(LOG_WARNING, "(mlfi_envrcpt) Forward-Path exceeds maximum length: %s (%d)", *argv, gettid());
	return SMFIS_CONTINUE;
    }

    syslog(LOG_NOTICE, "(mlfi_envrcpt) Starting %s from %s (%d)", *argv, MLFIPRIV->hostname, gettid());
    if(spasm_config->stats != NULL) {
	pthread_mutex_lock(&mutexLock_stats);
	spasm_update_stats("rcpt");
	pthread_mutex_unlock(&mutexLock_stats);
    }

    /* Add to MAX_UNKNOWN_USERS blacklist if threshold has been reached */
    if((spasm_config->unknownusersthreshold) && MLFIPRIV->unknownusers > spasm_config->unknownusersthreshold) {
	add_to_blacklist(MAX_UNKNOWN_USERS_BL, MLFIPRIV->ipstring);
    }

    /* No duplicate recipients */
    cur = MLFIPRIV->rcpts;
    while(cur != NULL) {
	if(strcmp(cur->rcpt, *argv) == 0) {
	    syslog(LOG_NOTICE, "(mlfi_envrcpt) Duplicate recipient from %s (%s)", MLFIPRIV->ipstring, MLFIPRIV->hostname);
	    smfi_setreply(ctx, "550", "5.1.1", "Already specified that recipient");
	    return SMFIS_REJECT;
	}
	cur = cur->next;
    }

    /* Prepare recipient info */
    memset(alias, '\0', sizeof(alias));
    snprintf(rcpt, sizeof(rcpt), "%s", *argv);
    if(**argv == '<') snprintf(username, sizeof(username), "%s", ((*argv)+1));
    else snprintf(username, sizeof(username), "%s", (*argv));
    lowercase(username);
    chr = strchr(username, '>');
    if(chr != NULL) *chr = '\0';
    fqdn = strchr(username, '@');
    if(fqdn != NULL && username[strlen(username)-1] == '.') username[strlen(username)-1] = '\0';
    if(fqdn == NULL) {
	char email[SMTP_PATH_MAX+1];
	if(strlen(spasm_config->defaultdomain) > SMTP_DOMAIN_MAX) {
	    syslog(LOG_WARNING, "(mlfi_envrcpt) spasm_config->defaultdomain exceeds maximum length: %s (%d)", spasm_config->defaultdomain, gettid());
	    return SMFIS_CONTINUE;
	}
	snprintf(email, sizeof(email), "%s@%s", username, spasm_config->defaultdomain);
	if(spasm_verify_filters(email) == 1) founduser = 1;
	if(getUserPrefs(username, spasm_config->defaultdomain, "alias", alias) && (spasm_verify_filters(alias) == 1)) founduser = 1;
	snprintf(domain, sizeof(domain), "%s", spasm_config->defaultdomain);
    } else {
	/* To enable administrators of subdomains to support user@host.domain.com instead of just
	   user@domain.com, we first find the envelope rcpt domain. We test each subdomain with
	   user@ to find the user's canonical domain name. If no user configuration is found, we
	   then test each subdomain on its own to find the first matching global domain options,
	   and that becomes the user's canonical domain name. If still no match is found, we fall
	   back to the fqdn of the specified envelope rcpt. Whew! */

	char *subdomain;

	*fqdn = '\0';
	fqdn++;

	subdomain = fqdn;
	while(subdomain != NULL) {
	    char email[SMTP_PATH_MAX+1];
	    snprintf(email, sizeof(email), "%s@%s", username, subdomain);
	    /* Aliases get first priority */
	    memset(alias, 0, sizeof(alias));
            if(getUserPrefs(username, subdomain, "alias", alias) && (spasm_verify_filters(alias) == 1)) {
                founduser = 1;
                if(strlen(subdomain) > SMTP_DOMAIN_MAX) {
                    syslog(LOG_WARNING, "(mlfi_envrcpt) Domain exceeds maximum length: %s (%d)", subdomain, gettid());
                    return SMFIS_CONTINUE;
                }
                snprintf(domain, sizeof(domain), "%s", subdomain);
                break;
            }
	    if(spasm_verify_filters(email) == 1) {
		/* Eureka! */
		founduser = 1;
		if(strlen(subdomain) > SMTP_DOMAIN_MAX) {
		    syslog(LOG_WARNING, "(mlfi_envrcpt) Domain exceeds maximum length: %s (%d)", subdomain, gettid());
		    return SMFIS_CONTINUE;
		}
		snprintf(domain, sizeof(domain), "%s", subdomain);
		break;
	    }
	    if((subdomain = strchr(subdomain, '.')) != NULL) subdomain++;
	}
	if(founduser == 0) {
	    /* Ok, we haven't yet found it... */
	    subdomain = fqdn;
	    while(subdomain != NULL) {
		if(spasm_verify_filters(subdomain) == 1) {
		    /* Eureka! */
		    founduser = 1;
		    if(strlen(subdomain) > SMTP_DOMAIN_MAX) {
			syslog(LOG_WARNING, "(mlfi_envrcpt) Domain exceeds maximum length: %s (%d)", subdomain, gettid());
			return SMFIS_CONTINUE;
		    }
		    snprintf(domain, sizeof(domain), "%s", subdomain);
		    break;
		}
		if((subdomain = strchr(subdomain, '.')) != NULL) subdomain++;
	    }
	}
	if(!founduser) {
	    /* Dammit. Guess we'll just use the fqdn for spamtrap and verification, then continue the milter */
	    snprintf(domain, sizeof(domain), "%s", fqdn);
	}
    }

    /* Is recipient a spamtrap? */
    if(search_regex_spamtraps(rcpt) != 0) {
	/* Sanity Checks */
	if(strcmp(MLFIPRIV->mailfrom, "<>") != 0) {
	    add_to_blacklist(spasm_config->spamtrapbl, MLFIPRIV->ipstring);
	}
        smfi_setreply(ctx, "550", "5.7.1", "Rejected");
        return SMFIS_REJECT;
    }

    /* Verify User */
    if(user_lookup(username, domain) == 0) {
	if(spasm_config->debug) syslog(LOG_WARNING, "Unknown user from %s (%s - %s) (%d)", MLFIPRIV->mailfrom, MLFIPRIV->hostname, MLFIPRIV->ipstring, gettid());
	MLFIPRIV->unknownusers++;
	smfi_setreply(ctx, "550", "5.1.1", "User unknown");
	return SMFIS_REJECT;
    }

    /* If we didn't find user filters, no need to test any filters... */
    if(founduser == 0) return SMFIS_CONTINUE;

    /* If we found an alias, rewrite username and domain */
    if(*alias != '\0') {
	char *ptr;
	if((ptr = strchr(alias, '@')) != NULL) {
	    *ptr = '\0';
	    ptr++;
	    snprintf(username, sizeof(username), "%s", alias);
	    snprintf(domain, sizeof(domain), "%s", ptr);
	}
    }

    /* Check whitelist prior to creating rcpt struct */
    if(userwl_lookup(ctx, username, domain)) return SMFIS_CONTINUE;

    /* Grab filter settings -- we'll just reuse the alias array */
    snprintf(alias, sizeof(alias), "%s@%s", username, domain);
    if((entries = spasm_get_filters(alias)) == NULL) return SMFIS_CONTINUE;
    first = entries;

    /* Prepare new rcpt struct */
    if((tmpcur = calloc(sizeof(struct spasmRecipient), 1)) == NULL) {
        syslog(LOG_ERR, "(mlfi_envrcpt) Error allocating memory for MLFIPRIV->rcpts: %m (%d)", gettid());
        return SMFIS_CONTINUE;
    }
    if(MLFIPRIV->rcpts == NULL)
        MLFIPRIV->rcpts = tmpcur;
    else {
        tmpcur->next = MLFIPRIV->rcpts;
        MLFIPRIV->rcpts = tmpcur;
    }
    cur = tmpcur;
    strcpy(cur->rcpt, rcpt);
    strcpy(cur->username, username);
    strcpy(cur->domain, domain);

    /* Check for domain logging exceptions */
    if((spasm_config->domainlogging) && (spasm_config->domlogexceptions != NULL)) {
	char *aptr, *bptr;
	aptr = spasm_config->domlogexceptions;
	bptr = spasm_config->domlogexceptions;
	while(aptr != NULL) {
	    if((bptr = strchr(aptr, ',')) != NULL) {
		*bptr = '\0';
		if(strcasecmp(cur->domain, aptr) == 0) domlogex = 1;
		*bptr = ',';
		bptr++;
		aptr = bptr;
	    } else {
		/* No more domlogexceptions */
		if(strcasecmp(cur->domain, aptr) == 0) domlogex = 1;
		aptr = bptr = NULL;
	    }
	}
    }
    if(spasm_config->debug && domlogex) syslog(LOG_DEBUG, "(mlfi_envrcpt) Found domain logging exception for %s (%d)", cur->domain, gettid());

    /* Check for domain logging */
    if(!domlogex &&
	((spasm_config->domainlogging == 1 && (strcasecmp(cur->domain, spasm_config->defaultdomain) != 0)) ||
	  getUserPrefs(cur->username, cur->domain, "use_domain_logging", NULL))) {
	snprintf(cur->cname, sizeof(cur->cname), "%s", cur->domain);
    } else {
        snprintf(cur->cname, sizeof(cur->cname), "%s@%s", cur->username, cur->domain);
    }
    lowercase(cur->cname);

    /* Get logging and subject tagging options */
    memset(buf, 0, sizeof(buf));
    if(getUserPrefs(cur->username, cur->domain, "use_logging", buf)) cur->logging = atoi(buf);
    memset(buf, 0, sizeof(buf));
    if(getUserPrefs(cur->username, cur->domain, "use_subject_tagging", buf)) cur->subject_tagging = atoi(buf);

    /* Preload filterStatus struct */
    i = 0;
    while(first != NULL) {
	pfltr = spasm_filters;
	while(pfltr->db_name) {
	    if(strcmp(pfltr->db_name, entries->entry) == 0) {
		strncpy(cur->fstatus[i].filter, pfltr->db_name, sizeof(cur->fstatus[i].filter)-1);
		cur->fstatus[i].setting = atoi(entries->strval);
		if(pfltr->filter_type == FTYPE_CONTENT && cur->fstatus[i].setting) cur->content_filtering = 1;
		i++;
		break;
	    }
	    pfltr++;
	}
	entries = first->next;
	free(first);
	first = entries;
    }

    /*
     * Begin processing the various anti-spam filters.  Start with
     * "whitelist only" because if the recipient has enabled this filter,
     * s/he doesn't want e-mail from any source other than what's been
     * listed on his/her personal whitelist.
     */
    if(pwl_only_fltr) {
	if((result = getFilterSetting(cur->fstatus, pwl_only_fltr->db_name))) {
	/*if((result = filter_lookup(cur->username, cur->domain, pwl_only_fltr->db_name))) {*/
	    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_envrcpt) %s (%d)", pwl_only_fltr->db_name, gettid());
	    add_filter_status(cur, pwl_only_fltr->db_name, result);
	    if(result & SPASM_REJECT) { /* Reject outright */
		(void) filter_reject_ops(ctx, cur, pwl_only_fltr);
		return SMFIS_REJECT;
	    }
	}
    }

    /* and now...the remainder of the anti-spam filters */
    pfltr = spasm_filters;
    while (pfltr->db_name) {
	switch (pfltr->adhoc_id)
	{
	    case SPASM_WL_ONLY:      /* did that */
		break;
	    case SPASM_SPAMASSASSIN: /* do that later */
	    case SPASM_VIRUS_SCAN:   /* ditto         */
	    case SPASM_DSPAM:        /* ditto         */
	    case SPASM_EXE_ATTACH:   /* ditto         */
		break;
	    default:
		/* do we want to run the filter? */

		  if((result = runFilter(cur, pfltr->db_name))) {
		/*if((result = filter_lookup(cur->username, cur->domain, pfltr->db_name))) {*/
		    /* Yes, as a matter of fact... */

		    int fltrd = 0;

		    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_envrcpt) %s (%d)", pfltr->db_name, gettid());
		    switch(pfltr->adhoc_id)
		    {
			case 0:    /* filter uses non-adhoc processing */
                            if (pfltr->filter_type == FTYPE_IP) {
                                snprintf(buf,
                                         sizeof(buf),
                                         "%u.%u.%u.%u.%s",
                                         RNIPQUAD(MLFIPRIV->ipaddress),
                                         pfltr->dnsbl);
                            } else { /* filter_type == FTYPE_HOST */
                                snprintf(buf,
                                         sizeof(buf),
                                         "%s.%s",
                                         MLFIPRIV->hostname,
                                         pfltr->dnsbl);
                            }
                            fltrd = dnsbl_lookup(buf);
			    break;
			case SPASM_HELO:
                            fltrd = filter_HELO(MLFIPRIV->helodomain, MLFIPRIV->hostdomain);
			    break;
			case SPASM_RDNS:
                            fltrd = filter_rDNS(MLFIPRIV->hostname);
			    break;
			case SPASM_RDNS_APNIC:
			    fltrd = filter_rDNS_APNIC(MLFIPRIV->ipstring, MLFIPRIV->hostname);
			    break;
			case SPASM_RDNS_RIPE:
			    fltrd = filter_rDNS_RIPE(MLFIPRIV->ipstring, MLFIPRIV->hostname);
			    break;
			case SPASM_SPAMTRAP:
                            fltrd = spasm_lookup_bl(spasm_config->spamtrapbl, MLFIPRIV->ipstring);
			    break;
			case SPASM_MAIL_HOST:
                            fltrd = filter_MAIL_HOST(MLFIPRIV->mailfromdomain, MLFIPRIV->hostdomain);
			    break;
			case SPASM_LOCALFILTER:
			    fltrd = test_local_bl(ctx);
			    break;
			case SPASM_MAX_UNK_USERS:
                            fltrd = spasm_lookup_bl(MAX_UNKNOWN_USERS_BL, MLFIPRIV->ipstring);
			    break;
			case SPASM_PERSONAL_BL:
			    fltrd = test_personal_bl(ctx, cur);
			    break;
			default:
			    syslog(LOG_ERR, "(mlfi_envrcpt) encountered unknown filter id %d (%s) (%d)",
				   pfltr->adhoc_id,
				   pfltr->db_name,
				   gettid());
			    break;
		    }
		    if (fltrd) {
			add_filter_status(cur, pfltr->db_name, result);
			if (result & SPASM_REJECT) {
			    (void) filter_reject_ops(ctx, cur, pfltr);
			    return SMFIS_REJECT;
			}
		    }
		}
		break;
	}
	pfltr++;
    }

    /* If content filter are selected, or the status is TAG or QUARANTINE, create qfile */
    if(cur->content_filtering || cur->status & SPASM_TAG || cur->status & SPASM_QUARANTINE) {
	char qfile[FILENAME_MAX+1];
	struct timeval tv;
	char *chr;

	SP_STRCAT(qfile, spasm_config->quarantinedir, cur->cname);
        mkdir(qfile, 0755);
	gettimeofday(&tv, NULL);
        snprintf(cur->tmpqfile, sizeof(cur->tmpqfile), "%s/%s%010li%010li.XXXXXX", spasm_config->tempdir, cur->cname, tv.tv_sec, tv.tv_usec);
        if((fd = mkstemp(cur->tmpqfile)) < 0 || (cur->quarantinefp = fdopen(fd, "a")) == NULL) {
            syslog(LOG_ERR, "(mlfi_envrcpt) Error opening quarantine file %s: %m (%d)", cur->tmpqfile, gettid());
        } else {
	    char datetime[32];
	    memset(datetime, 0, sizeof(datetime));
	    getDateTime(datetime, sizeof(datetime));
            chmod(cur->tmpqfile, 0644);
            chr = strrchr(cur->tmpqfile, '.');
            chr++;
            snprintf(cur->quarantinefile, sizeof(cur->quarantinefile), "%s/%s/%010li%010li.%s", spasm_config->quarantinedir, cur->cname, tv.tv_sec, tv.tv_usec, chr);
	    /* Write new Received: line */
	    fprintf(cur->quarantinefp, "Received: from %s (%s [%s])\r\n", MLFIPRIV->helohost, MLFIPRIV->hostname, MLFIPRIV->ipstring);
	    fprintf(cur->quarantinefp, "\tby %s (spasm) for %s;\r\n", local_hostname, cur->rcpt);
	    fprintf(cur->quarantinefp, "\t%s\r\n", datetime);
	    if(fflush(cur->quarantinefp) != 0) syslog(LOG_ERR, "(mlfi_envrcpt) Error flushing %s: %m (%d)", cur->tmpqfile, gettid());
        }
    }

    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_envrcpt) Finished %s@%s (%d)", cur->username, cur->domain, gettid());

    return SMFIS_CONTINUE;
}

sfsistat
mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
    struct spasmRecipient *cur;
    char email[SMTP_PATH_MAX+1];
    int result;

    if(MLFIPRIV == NULL) return SMFIS_CONTINUE;
    cur = MLFIPRIV->rcpts;

    /* Save subject line for later tagging */
    if(strcasecmp(headerf, "subject") == 0) {
	snprintf(MLFIPRIV->subject, sizeof(MLFIPRIV->subject), "%s", headerv);
    }

    while(cur != NULL) {
	/* If the message has already been rejected, don't bother with any of this stuff */
	if(cur->status & SPASM_REJECT) { cur = cur->next; continue; }

	/*
	** If we're here, then we're already accepting the body of the message.
	** Let's go ahead and test the From: header against the user's wl/bl.
	**
	** This needs to be fixed, to only check the email address portion of the header =/
	*/
	if((strcasecmp(headerf, "from")) == 0) {
	    snprintf(email, sizeof(email), "%s@%s", cur->username, cur->domain);
	    if(spasm_lookup_wl(email, headerv)) {
		cur->status = SPASM_OFF;
		cur->wl = 1;
	    } else if((result = runFilter(cur, "SF_PERSONAL_BLACKLIST")) && spasm_lookup_bl(email, headerv)) {
		add_filter_status(cur, "SF_PERSONAL_BLACKLIST", result);
		if(result & SPASM_REJECT) cur->delrcpt = 1;
	    }
	}

        /* If we're doing content filtering, or we have a TAG or QUARANTINE,
           print the header to the quarantine file */
	if((cur->content_filtering || cur->status) && (cur->quarantinefp != NULL) && (cur->wl == 0)) {
	    if(fprintf(cur->quarantinefp, "%s: %s\r\n", headerf, headerv) < 0)
		syslog(LOG_WARNING, "(mlfi_header) fprintf returned an error: %m");
	}
        cur = cur->next;
    }
    return SMFIS_CONTINUE;
}

sfsistat
mlfi_eoh(SMFICTX *ctx)
{
    struct spasmRecipient *cur;

    if(MLFIPRIV == NULL) return SMFIS_CONTINUE;
    cur = MLFIPRIV->rcpts;
    while(cur != NULL) {
	if(!(cur->status & SPASM_REJECT) && (cur->content_filtering || cur->status) && (cur->quarantinefp != NULL) && (cur->wl == 0)) {
	    time_t timep;
	    char times[35];
	    int i;
	    for(i=0; i<NUM_FILTERS; i++) {
		if(cur->fstatus[i].status) fprintf(cur->quarantinefp, "X-spasm: %s\r\n", cur->fstatus[i].filter);
	    }
	    timep = time(NULL);
	    memset(times, '\0', sizeof(times));
	    sprintf(times, "%s", (char *)ctime((time_t*)&timep));
	    /* Get rid of ctime()'s newline */
	    times[strlen(times)-1] = '\0';
	    fprintf(cur->quarantinefp, "X-spasm-rcpt: %s@%s\r\n", cur->username, cur->domain);
	    fprintf(cur->quarantinefp, "X-spasm-sender: %s\r\n", MLFIPRIV->mailfrom);
	    fprintf(cur->quarantinefp, "X-spasm-sender-IP: %s\r\n", MLFIPRIV->ipstring);
	    fprintf(cur->quarantinefp, "X-spasm-sender-host: %s\r\n", MLFIPRIV->hostname);
	    fprintf(cur->quarantinefp, "X-spasm-date: %s\r\n", times);
	    fprintf(cur->quarantinefp, "\r\n");
	}
	cur = cur->next;
    }

    return SMFIS_CONTINUE;
}

sfsistat
mlfi_body(SMFICTX *ctx, u_char *bodyp, size_t bodylen)
{
    struct spasmRecipient *cur;
    int result;

    if(MLFIPRIV == NULL) return SMFIS_CONTINUE;
    cur = MLFIPRIV->rcpts;
    while(cur != NULL) {
	if(!(cur->status & SPASM_REJECT) && (cur->content_filtering || cur->status) && (cur->quarantinefp != NULL) && (cur->wl == 0)) {
	    if((result = runFilter(cur, "SF_EXE_ATTACH"))) {
		if(filter_EXE(bodyp)) {
		    add_filter_status(cur, "SF_EXE_ATTACH", result);
		    if(result & SPASM_REJECT) cur->delrcpt = 1;
		} else {
		    fwrite(bodyp, bodylen, 1, cur->quarantinefp);
		}
	    } else {
		fwrite(bodyp, bodylen, 1, cur->quarantinefp);
	    }
	}
	cur = cur->next;
    }

    return SMFIS_CONTINUE;
}

sfsistat
mlfi_eom(SMFICTX *ctx)
{
    struct mlfiPriv *priv = MLFIPRIV;
    struct spasmRecipient *cur;
    int result;
    int numrcpts = 0;
    char email[SMTP_PATH_MAX+1];

    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_eom) Starting (%d)", gettid());
    memset(email, 0, sizeof(email));
    if(priv != NULL) {
        cur = MLFIPRIV->rcpts;
        while(cur != NULL) {
	    if(cur->quarantinefp != NULL) {
		/* Flush and close quarantine file */
		fflush(cur->quarantinefp);
		fclose(cur->quarantinefp);
		cur->quarantinefp = NULL;
	    }
	    if(cur->delrcpt) smfi_delrcpt(ctx, cur->rcpt); /* Handle header or body rejection */
	    if(cur->tmpqfile[0] != '\0') {
		/* Perform virus scan if selected, regardless of runFilter() result, to get virus background colour in CGI */
		if(!(cur->status & SPASM_REJECT) && cur->wl == 0 && (result = getFilterSetting(cur->fstatus, "SF_VIRUS_SCAN"))) {
		    const char *virname;
		    if((virname = MLFIPRIV->fltrcache.av) != NULL || (virname = filter_AV(cur->tmpqfile)) != NULL) {
			char fbuf[1024];
			if(MLFIPRIV->fltrcache.av == NULL) MLFIPRIV->fltrcache.av = virname;
		        syslog(LOG_NOTICE, "(mlfi_eom) Virus found in message from %s: %s (%d)", MLFIPRIV->ipstring, virname, gettid());
			if(spasm_config->stats != NULL) {
			    pthread_mutex_lock(&mutexLock_stats);
			    spasm_update_stats("virus");
			    pthread_mutex_unlock(&mutexLock_stats);
			}
			snprintf(fbuf, sizeof(fbuf), "X-spasm-Virus: %s\r\n", virname);
			add_header(cur->tmpqfile, fbuf);
		        add_filter_status(cur, "SF_VIRUS_SCAN", result);
		        if(result & SPASM_REJECT) smfi_delrcpt(ctx, cur->rcpt);
			/* else need to add Subject tagging */
	            }
	        }

	        /* Run SpamAssassin based on runFilter() */
	        if(!(cur->status & SPASM_REJECT) && spasm_config->sahost && cur->wl == 0 && (result = runFilter(cur, "SF_SPAMASSASSIN"))) {
		    int required_hits;
		    required_hits = filter_lookup(cur->username, cur->domain, SA_REQUIRED_HITS);
		    if(filter_SA(cur->tmpqfile, required_hits) == 1) {
		        add_filter_status(cur, "SF_SPAMASSASSIN", result);
		        if(result & SPASM_REJECT) smfi_delrcpt(ctx, cur->rcpt);
			/* else need to add Subject tagging */
		    }
	        }

		/* Run DSPAM based on runFilter() */
		if(!(cur->status & SPASM_REJECT) && spasm_config->dspam && cur->wl == 0 && (result = runFilter(cur, "SF_DSPAM"))) {
		    if(filter_DSPAM(cur->tmpqfile) == 1) {
			add_filter_status(cur, "SF_DSPAM", result);
			if(result & SPASM_REJECT) smfi_delrcpt(ctx, cur->rcpt);
			/* else need to add Subject tagging */
		    }
		}

	        /* Remove quarantined recipients and move file to permanent quarantine */
	        if(cur->status & SPASM_QUARANTINE) {
		    smfi_delrcpt(ctx, cur->rcpt);
		    if(cur->quarantinefile[0] != '\0') {
			moveQfile(cur->tmpqfile, cur->quarantinefile, cur->subject_tagging?priv->subject:NULL);
		    } else {
			syslog(LOG_ERR, "(mlfi_eom) Error moving %s: cur->quarantinefile is NULL (%d)", cur->tmpqfile, gettid());
		    }
		    memset(cur->tmpqfile, 0, sizeof(cur->tmpqfile));
	        }
	        /* Deliver to tag-only recipients and delete file */
	        if(cur->status & SPASM_TAG) {
                    if(deliverQuarantine(priv->mailfrom, cur->rcpt, cur->tmpqfile, cur->subject_tagging?priv->subject:NULL) != 0) {
                        syslog(LOG_ERR, "(mlfi_eom) Error delivering quarantine file %s (%d)", cur->tmpqfile, gettid());
                    } else {
                        smfi_delrcpt(ctx, cur->rcpt);
                        if((unlink(cur->tmpqfile)) != 0) {
                            syslog(LOG_ERR, "(mlfi_eom) Error unlinking quarantine file: %m (%d)", gettid());
                        }
			memset(cur->tmpqfile, 0, sizeof(cur->tmpqfile));
                    }
	        }
		/* Rejected or not filtered, so delete the file */
	        if(cur->status == SPASM_OFF || (cur->status & SPASM_REJECT)) {
                    if((unlink(cur->tmpqfile)) != 0) {
                        syslog(LOG_ERR, "(mlfi_eom) Error unlinking quarantine file: %m (%d)", gettid());
                    }
		    memset(cur->tmpqfile, 0, sizeof(cur->tmpqfile));
	        }
	        if(cur->status) log_rejection(ctx, cur);
            }
	    cur = cur->next;
	}

        /* End of message - Clear message based variables, keep private data
           and connection-related variables (hostname, ipaddress, ipstring) */
	memset(priv->mailfrom, 0, sizeof(priv->mailfrom));
	priv->mailfromhost = NULL;
	priv->mailfromdomain = NULL;
	memset(priv->subject, 0, sizeof(priv->subject));
	numrcpts = priv->num_rcpts;
	priv->num_rcpts = 0;
	cur = priv->rcpts;
	while(cur != NULL) {
	    snprintf(email, sizeof(email), "%s@%s", cur->username, cur->domain);
	    if(cur->quarantinefp != NULL) { fflush(cur->quarantinefp); fclose(cur->quarantinefp); }
            if(cur->tmpqfile[0] != '\0') {
                if(unlink(cur->tmpqfile) != 0) syslog(LOG_ERR, "(mlfi_eom) Error unlinking quarantine file: %m (%d)", gettid());
		memset(cur->tmpqfile, 0, sizeof(cur->tmpqfile));
            }
	    priv->rcpts = cur->next;
	    free(cur);
	    cur = priv->rcpts;
	}
	priv->rcpts = NULL;
    }
    /* Add tracking headers for any non-filtered recipients */
    smfi_addheader(ctx, "X-spasm-sender-IP", MLFIPRIV->ipstring);
    smfi_addheader(ctx, "X-spasm-sender-host", MLFIPRIV->hostname);
    if(numrcpts == 1) {
	smfi_addheader(ctx, "X-spasm-rcpt", email);
    }
    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_eom) Finished (%d)", gettid());
    return SMFIS_CONTINUE;
}

sfsistat
mlfi_abort(SMFICTX *ctx)
{
    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_abort) Starting (%d)", gettid());

    if(MLFIPRIV != NULL) {
	struct spasmRecipient *cur, *first;
	cur = MLFIPRIV->rcpts;
	while(cur != NULL) {
	    if(cur->quarantinefp != NULL) { fclose(cur->quarantinefp); cur->quarantinefp = NULL; }
	    if(cur->tmpqfile[0] != '\0') {
		if((unlink(cur->tmpqfile)) != 0) {
		    syslog(LOG_ERR, "(mlfi_abort) Error unlinking quarantine file: %m (%d)", gettid());
		}
		memset(cur->tmpqfile, 0, sizeof(cur->tmpqfile));
	    }
	    first = cur->next;
	    free(cur);
	    cur = first;
	}
	/* Clear message based variables */
	MLFIPRIV->mailfromhost = NULL;
	MLFIPRIV->mailfromdomain = NULL;
	memset(MLFIPRIV->mailfrom, 0, sizeof(MLFIPRIV->mailfrom));
	memset(MLFIPRIV->subject, 0, sizeof(MLFIPRIV->subject));
	MLFIPRIV->rcpts = NULL;
	MLFIPRIV->num_rcpts = 0;
    }

    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_abort) Finished (%d)", gettid());
    return SMFIS_CONTINUE;
}

sfsistat
mlfi_close(SMFICTX *ctx)
{
    struct mlfiPriv *priv = MLFIPRIV;
    struct spasmRecipient *cur;

/*
    if(priv != NULL && priv->ipstring != NULL) syslog(LOG_NOTICE, "(mlfi_close) Closing connection from %s (%d)", priv->ipstring, gettid());
    else syslog(LOG_NOTICE, "(mlfi_close) Closing connection (%d)", gettid());
*/

    /* Connection Closed - Free all allocated memory, set private data to NULL */
    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_close) Starting (%d)", gettid());
    if(priv != NULL) {
	cur = priv->rcpts;
	while(cur != NULL) {
	    if(cur->quarantinefp != NULL) { fflush(cur->quarantinefp); fclose(cur->quarantinefp); cur->quarantinefp = NULL; }
	    if(cur->tmpqfile[0] != '\0') {
		if((cur->status & SPASM_REJECT) || cur->status & SPASM_TAG || cur->status == SPASM_OFF) {
		    if((unlink(cur->tmpqfile)) != 0) {
		        syslog(LOG_ERR, "(mlfi_close) Error unlinking quarantine file: %m (%d)", gettid());
		    }
		}
	    }
	    priv->rcpts = cur->next;
	    free(cur);
	    cur = priv->rcpts;
	}
        free(priv);
    }
    smfi_setpriv(ctx, NULL);
    if(spasm_config->debug) syslog(LOG_DEBUG, "(mlfi_close) Finished (%d)", gettid());
    return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
    "spasm",
    SMFI_VERSION,
    SMFIF_ADDHDRS|SMFIF_DELRCPT,
    mlfi_connect,
    mlfi_helo,
    mlfi_envfrom,
    mlfi_envrcpt,
    mlfi_header,
    mlfi_eoh,
    mlfi_body,
    mlfi_eom,
    mlfi_abort,
    mlfi_close
};

static void
die(char *msg, int rc)
{
    char errstring[FILENAME_MAX + 128];

    snprintf(errstring, sizeof(errstring), "%s: %s", progname, msg);
    perror(errstring);
    exit(rc);
}

#define CMDLINE_HELP 128   /* largest ASCII value + 1 */
#define CMDLINE_VRSN 129

/* some getopt cmdline option stuff */
static struct option long_opts[] =
{
    {"help",         0, 0, CMDLINE_HELP},
    {"version",      0, 0, CMDLINE_VRSN},
    {NULL, 0, 0, 0}
};

static void
bad_cmdline_syntax(void)
{
    fprintf(stderr, "Try `%s --help' for more information\n", progname);
    exit(1);
}

static void
version(void)
{
    fprintf(stderr, "%s %s\n", progname, SPASM_VERSION);
    exit(0);
}

static void
cmdline_help(void)
{
    fprintf(stderr, "usage: %s [--help --version]\n\n", progname);
    fprintf(stderr,
"where:\n\n"
"--help                    display this message and exit\n"
"--version                 display program version number and exit\n");

    exit(0);
}

int
main(int argc, char **argv)
{
    int rc;
    const char *args = "";
    struct passwd *pw;
    char msg[128];
    char *slash;
#ifdef WITH_CLAMAV
    struct stat fstat;
#endif /* WITH_CLAMAV */
    int fd;
    int tries = 0;
    pid_t pid;
    int nap = 1;

    slash = strrchr(argv[0], '/');

    progname = (slash) ? slash + 1 : argv[0];

    for (;;)
    {
        rc = getopt_long(argc, argv, args, long_opts, NULL);
        if (rc == -1)
           break;
       switch (rc)
       {
           case CMDLINE_HELP:
               cmdline_help();
               break;
           case CMDLINE_VRSN:
               version();
               break;
           case '?':
               bad_cmdline_syntax();
               break;
       }
    }
    if (optind < argc) {
       /* trailing garbage on the command line */

       fprintf(stderr, "%s: too many command line arguments\n", progname);
       bad_cmdline_syntax();
       /* NOT REACHED */
    }

    {
	struct rlimit rlim;
	rlim.rlim_cur = rlim.rlim_max = RLIM_INFINITY;
	setrlimit(RLIMIT_CORE, &rlim);
    }

    /* Open syslog to catch any spasm_cfg errors */
    openlog("spasmd", LOG_PID, LOG_FACILITY);

    /* Read configuration file */
    if((spasm_config = spasm_cfg()) == NULL) {
        exit(1);
    }

    /* Get hostname */
    if(gethostname(local_hostname, sizeof(local_hostname)) != 0) {
	die("error getting local hostname", 1);
    }

#ifdef WITH_CLAMAV
    /* Open clamav database */
    if((rc = cl_loaddbdir(cl_retdbdir(), &clamav_root, NULL))) {
        /*printf("cl_loaddbdir: %s\n", cl_perror(ret));*/
	die("error loading ClamAV database", 1);
    }
    if((rc = cl_build(clamav_root))) {
        /*printf("Database initialization error: %s\n", cl_strerror(ret));*/
        cl_free(clamav_root); /* free the partial trie */
	die("error initialising ClamAV database", 1);
    }
    if(stat(cl_retdbdir(), &fstat) != 0)
	die("error getting status of ClamAV db directory", 1);
    clamav_lastload = fstat.st_mtime;
#endif /* WITH_CLAMAV */

    /* Touch pidfile before dropping privs */
    fd = open(spasm_config->pidfile, O_WRONLY|O_CREAT, 0600);
    if(fd < 0) syslog(LOG_WARNING, "Error opening pidfile: %m");
    else close(fd);
    
    /* Drop privs immediately */
    if((pw = getpwnam(spasm_config->spasmuser)) == NULL) {
	snprintf(msg, sizeof(msg), "error reading passwd entry for \"%s\"", spasm_config->spasmuser);
	die(msg, 1);
    }
    chown(spasm_config->pidfile, pw->pw_uid, pw->pw_gid);
    if((setgid(pw->pw_gid) != 0) || (setuid(pw->pw_uid) != 0)) {
	unlink(spasm_config->pidfile);
	die("error dropping privileges", 1);
    }

    /* Setup milter connection */
    if(smfi_setconn(spasm_config->socket) == MI_FAILURE) {
	unlink(spasm_config->pidfile);
	fprintf(stderr, "%s: mlfi_setconn() failed\n", progname);
	exit(EX_UNAVAILABLE);
    }

    /* Register the milter */
    if(smfi_register(smfilter) == MI_FAILURE) {
	unlink(spasm_config->pidfile);
	fprintf(stderr, "%s: smfi_register() failed\n", progname);
	exit(EX_UNAVAILABLE);
    }

    /* Set the timeout */
    smfi_settimeout(300);

    /* Compile our regex to return (SLD.TLD) of a FQDN */
/*
    if((rc = regcomp(&re_sld, DOMAIN_NAME_REGEX, REG_EXTENDED)) != 0) {
	unlink(spasm_config->pidfile);
	fprintf(stderr, "%s: error compiling regex \"%s\" (errcode: %d)\n",
		progname, DOMAIN_NAME_REGEX, rc);
	exit(EX_UNAVAILABLE);
    }
*/
    /* Compile our regex to check for a CIDR entry */
/*
    if((rc = regcomp(&re_cidr, CIDR_REGEX, REG_EXTENDED|REG_NOSUB)) != 0) {
	unlink(spasm_config->pidfile);
	fprintf(stderr, "%s: error compiling regex \"%s\" (errcode: %d)\n",
		progname, CIDR_REGEX, rc);
	exit(EX_UNAVAILABLE);
    }
*/
    /* Compile our regex to return an IP address */
    if((rc = regcomp(&re_ip, IP_REGEX, REG_EXTENDED)) != 0) {
        unlink(spasm_config->pidfile);
        fprintf(stderr, "%s: error compiling regex \"%s\" (errcode: %d)\n",
                progname, IP_REGEX, rc);
        exit(EX_UNAVAILABLE);
    }
    /* Compile our regex to check for an IP address */
/*
    if((rc = regcomp(&re_ipa, IPA_REGEX, REG_EXTENDED|REG_NOSUB)) != 0) {
        unlink(spasm_config->pidfile);
        fprintf(stderr, "%s: error compiling regex \"%s\" (errcode: %d)\n",
                progname, IPA_REGEX, rc);
        exit(EX_UNAVAILABLE);
    }
*/
    /* Compile our regex to check for an executable attachment */
    if((rc = regcomp(&re_exe, EXE_REGEX, REG_EXTENDED|REG_NOSUB)) != 0) {
	unlink(spasm_config->pidfile);
	fprintf(stderr, "%s: error compiling regex \"%s\" (errcode: %d)\n",
		progname, EXE_REGEX, rc);
	exit(EX_UNAVAILABLE);
    }
    /* Compile our regex to check for a regex entry */
    if((rc = regcomp(&re_srs0, SRS0_REGEX, REG_EXTENDED|REG_ICASE)) != 0) {
	unlink(spasm_config->pidfile);
	fprintf(stderr, "%s: error compiling regex \"%s\" (errcode: %d)\n",
		progname, SRS0_REGEX, rc);
	exit(EX_UNAVAILABLE);
    }

    /* cache pointer(s) to "important" filter data structure(s) */
    cache_filter_ptrs();

    /* Set Signal Handlers */
    signal(SIGQUIT, cleanup);
    signal(SIGSEGV, sigSEGV);

/*return smfi_main();*/

    /* Fork daemon. Looping code adapted from vile's npopen.c */
    while ((pid = fork()) < 0) {
        if (++tries > 5) {
            unlink(spasm_config->pidfile);
            die("daemon fork() failed", EX_UNAVAILABLE);
        }   
        sleep(nap);
        nap <<= 1;
    }   
    /* We have forkage! */
    if (pid != 0) {
        /* parental unit returns */
        return 0;
    } else {                                                                                                                                            
        /* double-fork() to prevent regaining a controlling terminal */
        if(setsid() < 0) syslog(LOG_WARNING, "Error starting new session");
        tries = 0;
        nap = 1;
        while((pid = fork())< 0) {
            if(++tries > 5) {
                unlink(spasm_config->pidfile);
                die("daemon double-fork() failed", EX_UNAVAILABLE);
            }
            sleep(nap);
            nap <<= 1;
        }
        if(pid != 0) {
            /* parental unit returns */
            return 0;
        } else {
            /* child goes out to play -- let the spam rejection begin! */
            FILE *fp;
            if((fp = fopen(spasm_config->pidfile, "w")) != NULL) {
                fprintf(fp, "%d", getpid());
                fclose(fp);
            } else {
                syslog(LOG_WARNING, "Error opening pidfile for output: %m");
            }
            if(chdir(spasm_config->spasmroot) < 0) syslog(LOG_WARNING, "Error changing directory: %m");
            umask(0);
	    stdin = freopen("/dev/null", "r", stdin);
	    stdout = freopen("/dev/null", "w", stdout);
	    stderr = freopen("/dev/null", "w", stderr);
            syslog(LOG_INFO, "(main) Starting (%d)", gettid());
            return (smfi_main());
        }
    }
}
